
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.io.*;
import java.util.*;

/**
 * An object that represents the Java<sup>TM</sup> "class file" format.
 * <p>
 *   {@link #store(OutputStream)} generates Java<sup>TM</sup> bytecode
 *   which is suitable for being processed by a Java<sup>TM</sup> virtual
 *   machine.
 * </p>
 */
public class ClassFile {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/ClassFile.java,v 1.15 2002/05/25 20:59:03 arno Exp $";

    /**
     * Construct from parsed components.
     */
    public ClassFile(
        short    accessFlags,
        String   thisClassName,
        String   superclassFD,
        String[] interfaceNames
    ) {
        this.accessFlags   = accessFlags;
        this.thisClass     = this.addConstantClassInfo(
            Descriptor.fromClassName(thisClassName) // typeFD
        );
        this.thisClassName = thisClassName;
        this.superclass    = this.addConstantClassInfo(
            superclassFD // typeFD
        );
        this.interfaces    = new short[interfaceNames.length];
        for (int i = 0; i < interfaceNames.length; ++i) {
            this.interfaces[i] = addConstantClassInfo(
                Descriptor.fromClassName(interfaceNames[i]) // typeFD
            );
        }
    }

    /**
     * Add a "CONSTANT_Class_info" structure to the class file (see JVM
     * specification, section 4.4.1).
     */
    public short addConstantClassInfo(
        String typeFD
    ) {
        String s;
        if (Descriptor.isClassOrInterfaceReference(typeFD)) {
            s = Descriptor.toClassName(typeFD).replace('.', '/');
        } else
        if (Descriptor.isArrayReference(typeFD)) {
            s = typeFD;
        } else
        {
            throw new RuntimeException("\"" + Descriptor.toString(typeFD) + "\" is neither a class nor an array");
        }

        return addToConstantPool(new ConstantClassInfo(addConstantUtf8Info(s)));
    }

    /**
     * Add a "CONSTANT_Fieldref_info" structure to the class file (see JVM
     * specification, section 4.4.2).
     */
    public short addConstantFieldrefInfo(
        String classFD, // Descriptor (class or array)
        String fieldName,
        String fieldFD
    ) {
        return addToConstantPool(new ConstantFieldrefInfo(
            addConstantClassInfo(classFD),
            addConstantNameAndTypeInfo(fieldName, fieldFD)
        ));
    }

    /**
     * Add a "CONSTANT_Methodref_info" structure to the class file (see JVM
     * specification, section 4.4.2).
     */
    public short addConstantMethodrefInfo(
        String classFD,
        String methodName,
        String methodMD
    ) {
        return addToConstantPool(new ConstantMethodrefInfo(
            addConstantClassInfo(classFD),
            addConstantNameAndTypeInfo(methodName, methodMD)
        ));
    }

    /**
     * Add a "CONSTANT_InterfaceMethodref_info" structure to the class file
     * (see JVM specification, section 4.4.2).
     */
    private short addConstantInterfaceMethodrefInfo(
        short  classIndex,
        String name,
        String descriptor
    ) {
        return addToConstantPool(new ConstantInterfaceMethodrefInfo(
            classIndex,
            addConstantNameAndTypeInfo(name, descriptor)
        ));
    }
    private short addConstantInterfaceMethodrefInfo(
        String className, // Descriptor (class or array)
        String name,
        String descriptor
    ) {
        return addToConstantPool(new ConstantInterfaceMethodrefInfo(
            addConstantClassInfo(className),
            addConstantNameAndTypeInfo(name, descriptor)
        ));
    }

    /**
     * Add a "CONSTANT_String_info" structure to the class file (see JVM
     * specification, section 4.4.3).
     */
    public short addConstantStringInfo(String string) {
        return addToConstantPool(new ConstantStringInfo(addConstantUtf8Info(string)));
    }

    /**
     * Add a "CONSTANT_Integer_info" structure to the class file (see JVM
     * specification, section 4.4.4).
     */
    public short addConstantIntegerInfo(final int value) {
        return addToConstantPool(new ConstantIntegerInfo(value));
    }

    /**
     * Add a "CONSTANT_Float_info" structure to the class file (see JVM
     * specification, section 4.4.4).
     */
    public short addConstantFloatInfo(final float value) {
        return addToConstantPool(new ConstantFloatInfo(value));
    }

    /**
     * Add a "CONSTANT_Long_info" structure to the class file (see JVM
     * specification, section 4.4.5).
     */
    public short addConstantLongInfo(final long value) {
        return addToConstantPool(new ConstantLongInfo(value));
    }

    /**
     * Add a "CONSTANT_Double_info" structure to the class file (see JVM
     * specification, section 4.4.5).
     */
    public short addConstantDoubleInfo(final double value) {
        return addToConstantPool(new ConstantDoubleInfo(value));
    }

    /**
     * Add a "CONSTANT_NameAndType_info" structure to the class file (see JVM
     * specification, section 4.4.6).
     */
    private short addConstantNameAndTypeInfo(String name, String descriptor) {
        return addToConstantPool(new ConstantNameAndTypeInfo(
            addConstantUtf8Info(name),
            addConstantUtf8Info(descriptor)
        ));
    }

    /**
     * Add a "CONSTANT_Utf8_info" structure to the class file (see JVM
     * specification, section 4.4.7).
     */
    private short addConstantUtf8Info(final String s) {
        return addToConstantPool(new ConstantUtf8Info(s));
    }

    // Convenience method that adds a String, Integer, Float, Long or
    // Double ConstantInfo
    private short addConstantSIFLDInfo(Object cv) {
        if (cv instanceof String) {
            return addConstantStringInfo((String) cv);
        } else
        if (
            cv instanceof Byte    ||
            cv instanceof Short   ||
            cv instanceof Integer
        ) {
            return addConstantIntegerInfo(((Number) cv).intValue());
        } else
        if (cv instanceof Boolean) {
            return addConstantIntegerInfo(((Boolean) cv).booleanValue() ? 1 : 0);
        } else
        if (cv instanceof Character) {
            return addConstantIntegerInfo(((Character) cv).charValue());
        } else
        if (cv instanceof Float) {
            return addConstantFloatInfo(((Float) cv).floatValue());
        } else
        if (cv instanceof Long) {
            return addConstantLongInfo(((Long) cv).longValue());
        } else
        if (cv instanceof Double) {
            return addConstantDoubleInfo(((Double) cv).doubleValue());
        } else
        {
            throw new RuntimeException("Unexpected constant value type \"" + cv.getClass().getName() + "\"");
        }
    }

    private short addToConstantPool(ConstantPoolInfo cpi) {
        Short index = (Short) constantPoolMap.get(cpi);
        if (index != null) return index.shortValue();

        short res = (short) constantPool.size();
        constantPool.add(cpi);
        if (cpi.isWide()) constantPool.add(null);

        constantPoolMap.put(cpi, new Short(res));
        return res;
    }

    public void addFieldInfo(
        short  accessFlags,
        String fieldTypeFD,
        String fieldName,
        Object optionalConstantValue
    ) {
        this.fieldInfos.add(new FieldInfo(
            accessFlags,          // accessFlags
            fieldTypeFD,          // typeFD
            fieldName,            // name
            optionalConstantValue // optionalConstantValue
        ));
    }

    public MethodInfo addMethodInfo(
        short  accessFlags,
        String methodName,
        String methodMD
    ) {
        MethodInfo mi = new MethodInfo(
            accessFlags, // accessFlags
            methodName,  // name
            methodMD     // descriptor
        );
        this.methodInfos.add(mi);
        return mi;
    }

    /**
     * Write to an {@link OutputStream}, in "class file" format.
     */
    public void store(OutputStream os) throws IOException {
        DataOutputStream dos = new DataOutputStream(os);
        dos.writeInt(0xcafebabe);                       // magic
        dos.writeShort(3);                              // minor_version
        dos.writeShort(45);                             // major_version
        dos.writeShort(constantPool.size());            // constant_pool_count
        for (int i = 1; i < constantPool.size(); ++i) { // constant_pool
            ConstantPoolInfo cpi = (ConstantPoolInfo) constantPool.elementAt(i);
            if (cpi == null) continue; // (Double or Long CPI.)
            cpi.store(dos);
        }
        dos.writeShort(accessFlags);                    // access_flags
        dos.writeShort(thisClass);                      // this_class
        dos.writeShort(superclass);                     // super_class
        dos.writeShort(interfaces.length);              // interfaces_count
        for (int i = 0; i < interfaces.length; ++i) {   // interfaces
            dos.writeShort(interfaces[i]);
        }
        dos.writeShort(fieldInfos.size());              // fields_count
        for (int i = 0; i < fieldInfos.size(); ++i) {   // fields
            ((FieldInfo) fieldInfos.elementAt(i)).store(dos);
        }
        dos.writeShort(methodInfos.size());             // methods_count
        for (int i = 0; i < methodInfos.size(); ++i) {  // methods
            ((MethodInfo) methodInfos.elementAt(i)).store(dos);
        }
        dos.writeShort(attributes.size());              // attributes_count
        for (int i = 0; i < attributes.size(); ++i) {   // attributes
            ((AttributeInfo) attributes.elementAt(i)).store(dos);
        }
    }

    private final Vector    constantPool = new Vector(); // ConstantPoolInfo
    { constantPool.add(null); } // Add fake "0" index entry.
    public  /*final*/ short     accessFlags;
    private /*final*/ short     thisClass;
    private /*final*/ short     superclass;
    private /*final*/ short[]   interfaces;
    private final Vector    fieldInfos  = new Vector(); // FieldInfo
    private final Vector    methodInfos = new Vector(); // MethodInfo
    private final Vector    attributes  = new Vector(); // AttributeInfo

    // Convenience.
    private final HashMap   constantPoolMap = new HashMap(); // ConstantPoolInfo => Short
    public  /*final*/ String    thisClassName;

    private interface ConstantPoolInfo {
        abstract public void store(DataOutputStream dos) throws IOException;
        abstract public boolean isWide();
    }
    private static class ConstantClassInfo implements ConstantPoolInfo {
        public ConstantClassInfo(short nameIndex) { this.nameIndex = nameIndex; }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(7);
            dos.writeShort(nameIndex);
        }
        public boolean equals(Object o) {
            return o instanceof ConstantClassInfo && ((ConstantClassInfo) o).nameIndex == nameIndex;
        }
        public int hashCode() { return nameIndex; }
        private final short nameIndex;
    }
    private static class ConstantFieldrefInfo implements ConstantPoolInfo {
        public ConstantFieldrefInfo(short classIndex, short nameAndTypeIndex) {
            this.classIndex = classIndex;
            this.nameAndTypeIndex = nameAndTypeIndex;
        }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(9);
            dos.writeShort(classIndex);
            dos.writeShort(nameAndTypeIndex);
        }
        public boolean equals(Object o) {
            return (
                o instanceof ConstantFieldrefInfo &&
                ((ConstantFieldrefInfo) o).classIndex == classIndex &&
                ((ConstantFieldrefInfo) o).nameAndTypeIndex == nameAndTypeIndex
            );
        }
        public int hashCode() { return classIndex + (nameAndTypeIndex << 16); }
        private /*final*/ short classIndex;       // JAVAC 1.2.2 bug
        private /*final*/ short nameAndTypeIndex; // JAVAC 1.2.2 bug
    }
    private static class ConstantMethodrefInfo implements ConstantPoolInfo {
        public ConstantMethodrefInfo(short classIndex, short nameAndTypeIndex) {
            this.classIndex = classIndex;
            this.nameAndTypeIndex = nameAndTypeIndex;
        }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(10);
            dos.writeShort(classIndex);
            dos.writeShort(nameAndTypeIndex);
        }
        public boolean equals(Object o) {
            return (
                o instanceof ConstantMethodrefInfo &&
                ((ConstantMethodrefInfo) o).classIndex == classIndex &&
                ((ConstantMethodrefInfo) o).nameAndTypeIndex == nameAndTypeIndex
            );
        }
        public int hashCode() { return classIndex + (nameAndTypeIndex << 16); }
        private /*final*/ short classIndex;       // JAVAC 1.2.2 bug
        private /*final*/ short nameAndTypeIndex; // JAVAC 1.2.2 bug
    }
    private static class ConstantInterfaceMethodrefInfo implements ConstantPoolInfo {
        public ConstantInterfaceMethodrefInfo(short classIndex, short nameAndTypeIndex) {
            this.classIndex = classIndex;
            this.nameAndTypeIndex = nameAndTypeIndex;
        }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(11);
            dos.writeShort(classIndex);
            dos.writeShort(nameAndTypeIndex);
        }
        public boolean equals(Object o) {
            return (
                o instanceof ConstantInterfaceMethodrefInfo &&
                ((ConstantInterfaceMethodrefInfo) o).classIndex == classIndex &&
                ((ConstantInterfaceMethodrefInfo) o).nameAndTypeIndex == nameAndTypeIndex
            );
        }
        public int hashCode() { return classIndex + (nameAndTypeIndex << 16); }
        private final short classIndex;
        private final short nameAndTypeIndex;
    }
    private static class ConstantStringInfo implements ConstantPoolInfo {
        public ConstantStringInfo(short stringIndex) {
            this.stringIndex = stringIndex;
        }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(8);
            dos.writeShort(stringIndex);
        }
        public boolean equals(Object o) {
            return o instanceof ConstantStringInfo && ((ConstantStringInfo) o).stringIndex == stringIndex;
        }
        public int hashCode() { return stringIndex; }
        private final short stringIndex;
    }
    private static class ConstantIntegerInfo implements ConstantPoolInfo {
        public ConstantIntegerInfo(int value) {
            this.value = value;
        }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(3);
            dos.writeInt(value);
        }
        public boolean equals(Object o) {
            return o instanceof ConstantIntegerInfo && ((ConstantIntegerInfo) o).value == value;
        }
        public int hashCode() { return value; }
        private final int value;
    }
    private static class ConstantFloatInfo implements ConstantPoolInfo {
        public ConstantFloatInfo(float value) {
            this.value = value;
        }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(4);
            dos.writeFloat(value);
        }
        public boolean equals(Object o) {
            return o instanceof ConstantFloatInfo && ((ConstantFloatInfo) o).value == value;
        }
        public int hashCode() { return Float.floatToIntBits(value); }
        private final float value;
    }
    private static class ConstantLongInfo implements ConstantPoolInfo {
        public ConstantLongInfo(long value) {
            this.value = value;
        }
        public boolean isWide() { return true; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(5);
            dos.writeLong(value);
        }
        public boolean equals(Object o) {
            return o instanceof ConstantLongInfo && ((ConstantLongInfo) o).value == value;
        }
        public int hashCode() { return (int) value ^ (int) (value >> 32); }
        private final long value;
    }
    private static class ConstantDoubleInfo implements ConstantPoolInfo {
        public ConstantDoubleInfo(double value) {
            this.value = value;
        }
        public boolean isWide() { return true; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(6);
            dos.writeDouble(value);
        }
        public boolean equals(Object o) {
            return o instanceof ConstantDoubleInfo && ((ConstantDoubleInfo) o).value == value;
        }
        public int hashCode() {
            long bits = Double.doubleToLongBits(value);
            return (int) bits ^ (int) (bits >> 32);
        }
        private final double value;
    }
    private static class ConstantNameAndTypeInfo implements ConstantPoolInfo {
        public ConstantNameAndTypeInfo(short nameIndex, short descriptorIndex) {
            this.nameIndex       = nameIndex;
            this.descriptorIndex = descriptorIndex;
        }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(12);
            dos.writeShort(nameIndex);
            dos.writeShort(descriptorIndex);
        }
        public boolean equals(Object o) {
            return (
                o instanceof ConstantNameAndTypeInfo &&
                ((ConstantNameAndTypeInfo) o).nameIndex       == nameIndex &&
                ((ConstantNameAndTypeInfo) o).descriptorIndex == descriptorIndex
            );
        }
        public int hashCode() { return nameIndex + (descriptorIndex << 16); }
        private /*final*/ short nameIndex;       // JAVAC 1.2.2 bug
        private /*final*/ short descriptorIndex; // JAVAC 1.2.2 bug
    }
    private static class ConstantUtf8Info implements ConstantPoolInfo {
        public ConstantUtf8Info(String s) {
            if (s == null) throw new RuntimeException();
            this.s = s;
        }
        public boolean isWide() { return false; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeByte(1);
            dos.writeUTF(s);
        }
        public boolean equals(Object o) {
            return o instanceof ConstantUtf8Info && ((ConstantUtf8Info) o).s.equals(this.s);
        }
        public int hashCode() { return this.s.hashCode(); }
        private /*final*/ String s; // JAVAC 1.2.2 bug
    }

    /**
     * This class represents a "method_info" structure, as defined by the
     * JVM specification.
     */
    public class MethodInfo {

        /**
         * Initialize the "method_info" structure.
         */
        public MethodInfo(
            short     accessFlags,
            String    name,
            String    descriptor
        ) {
            this.accessFlags     = accessFlags;
            this.nameIndex       = ClassFile.this.addConstantUtf8Info(name);
            this.descriptorIndex = ClassFile.this.addConstantUtf8Info(descriptor);
        }

        /**
         * Write this object to a {@link DataOutputStream}, in the format
         * defined by the JVM specification.
         */
        public void store(DataOutputStream dos) throws IOException {
            dos.writeShort(accessFlags);       // access_flags
            dos.writeShort(nameIndex);         // name_index
            dos.writeShort(descriptorIndex);   // descriptor_index
            dos.writeShort(attributes.size()); // attributes_count
            for (int i = 0; i < attributes.size(); ++i) {
                ((AttributeInfo) attributes.elementAt(i)).store(dos);
            }
        }

        /**
         * Create the "Code" attribute, if it does not yet exist.
         */
        public CodeAttribute getCodeAttribute() {
            if (this.codeAttribute == null) {
                this.codeAttribute = ClassFile.this.new CodeAttribute();
                attributes.add(this.codeAttribute);
            }
            return this.codeAttribute;
        }

        private final short  accessFlags;
        private final short  nameIndex;
        private final short  descriptorIndex;
        private final Vector attributes = new Vector(); // AttributeInfo

        private CodeAttribute codeAttribute = null;
    }

    private class FieldInfo {
        public FieldInfo(
            short  accessFlags,
            String typeFD,
            String name,
            Object optionalConstantValue
        ) {
            this.accessFlags     = accessFlags;
            this.nameIndex       = ClassFile.this.addConstantUtf8Info(name);
            this.descriptorIndex = ClassFile.this.addConstantUtf8Info(typeFD);
            if (optionalConstantValue != null) {
                this.attributes.add(
                    new ConstantValueAttribute(optionalConstantValue)
                );
            }
        }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeShort(accessFlags);       // access_flags
            dos.writeShort(nameIndex);         // name_index
            dos.writeShort(descriptorIndex);   // descriptor_index
            dos.writeShort(attributes.size()); // attributes_count
            for (int i = 0; i < this.attributes.size(); ++i) {
                ((AttributeInfo) this.attributes.elementAt(i)).store(dos);
            }
        }

        private final short  accessFlags;
        private final short  nameIndex;
        private final short  descriptorIndex;
        private final Vector attributes = new Vector(); // AttributeInfo
    }

    private abstract class AttributeInfo {
        public AttributeInfo(String name) {
            this.nameIndex = ClassFile.this.addConstantUtf8Info(name);
        }
        public ClassFile getClassFile() { return ClassFile.this; }
        public void store(DataOutputStream dos) throws IOException {
            dos.writeShort(nameIndex);                 // attribute_name_index;
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            storeBody(new DataOutputStream(baos));
            dos.writeInt(baos.size());                 // attribute_length
            baos.writeTo(dos);                         // info
        }
        protected abstract void storeBody(DataOutputStream dos) throws IOException;

        private final short nameIndex;
    }

    private class ConstantValueAttribute extends AttributeInfo {
        ConstantValueAttribute(Object constantValue) {
            super("ConstantValue");
            this.constantValueIndex = ClassFile.this.addConstantSIFLDInfo(constantValue);
        }
        // Implement "AttributeInfo".
        protected void storeBody(DataOutputStream dos) throws IOException {
            dos.writeShort(constantValueIndex);
        }
        private final short constantValueIndex;
    }

    /**
     * A representation of a "Code" attribute. Supplies several "write()"
     * methods, similar to an {@link OutputStream}, plus {@Offset}s and
     *{@link Inserter}s, which allow for very flexible generation of bytecode.
     */
    public class CodeAttribute extends AttributeInfo {

        /**
         * Create an empty "Code" attribute.
         */
        public CodeAttribute() {
            super("Code");
        }

        /**
         * Allocate space for a local variable of the given size (1 or 2)
         * on the local variable array.
         *
         * As a side effect, the "max_locals" field of the "Code" attribute
         * is updated.
         *
         * The only way to deallocate local variables is to
         * {@#saveLocalVariables()} and later {@#restoreLocalVariables()}.
         */
        public short allocateLocalVariable(
            short size // 1 or 2
        ) {
            short res = this.localVariableArrayLength;
            this.localVariableArrayLength += size;
            if (this.localVariableArrayLength > this.maxLocals) {
                this.maxLocals = this.localVariableArrayLength;
            }

            return res;
        }

        /**
         * Remember the current size of the local variables array.
         */
        public void saveLocalVariables() {
            this.savedLocalVariableArrayLengths.push(new Short(this.localVariableArrayLength));
        }

        /**
         * Restore the previous size of the local variables array.
         */
        public void restoreLocalVariables() {
            this.localVariableArrayLength = ((Short) savedLocalVariableArrayLengths.pop()).shortValue();
        }

        private static final byte UNEXAMINED = -1;
        private static final byte INVALID_OFFSET = -2;

        protected void storeBody(DataOutputStream dos) throws IOException {
            dos.writeShort(maxStack);                     // max_stack
            dos.writeShort(maxLocals);                    // max_locals
            dos.writeInt(end.offset);                     // code_length
            dos.write(code, 0, end.offset);               // code
            dos.writeShort(exceptionTable.size());        // exception_table_length
            for (int i = 0; i < this.exceptionTable.size(); ++i) { // exception_table
                ExceptioN exception = (ExceptioN) exceptionTable.elementAt(i);
                dos.writeShort(exception.startPC.offset);
                dos.writeShort(exception.endPC.offset);
                dos.writeShort(exception.handlerPC.offset);
                dos.writeShort(exception.catchType);
            }
            dos.writeShort(attributes.size());            // attributes_count
            for (int i = 0; i < attributes.size(); ++i) { // attributes;
                ((AttributeInfo) attributes.elementAt(i)).store(dos);
            }
        }

        /**
         * Checks the code for consistency; updates the "maxStack" member.
         *
         * Notice: On inconsistencies, a "RuntimeException" is thrown (KLUDGE).
         */
        public void flowAnalysis() {
            byte[] stackSizes = new byte[this.end.offset];
            Arrays.fill(stackSizes, UNEXAMINED);

            // Analyze flow from offset zero.
            this.flowAnalysis(
                this.code,       // code
                this.end.offset, // codeSize
                0,               // offset
                0,               // stackSize
                stackSizes,      // stackSizes
                ClassFile.this.constantPool
            );

            // Analyze flow from exception handler entry points.
            int analyzedExceptionHandlers = 0;
            while (analyzedExceptionHandlers != this.exceptionTable.size()) {
                for (int i = 0; i < this.exceptionTable.size(); ++i) {
                    ExceptioN ex = (ExceptioN) this.exceptionTable.elementAt(i);
                    if (stackSizes[ex.startPC.offset] != UNEXAMINED) {
                        this.flowAnalysis(
                            this.code,                         // code
                            this.end.offset,                   // codeSize
                            ex.handlerPC.offset,               // offset
                            stackSizes[ex.startPC.offset] + 1, // stackSize
                            stackSizes,                        // stackSizes
                            ClassFile.this.constantPool        // constantPool
                        );
                        ++analyzedExceptionHandlers;
                    }
                }
            }

            // Check results and determine maximum stack size.
            this.maxStack = 0;
            for (int i = 0; i < stackSizes.length; ++i) {
                byte ss = stackSizes[i];
                if (ss == UNEXAMINED) throw new RuntimeException("Unexamined code at offset " + i);
                if (ss > this.maxStack) this.maxStack = ss;
            }
        }

        private void flowAnalysis(
            byte[] code,       // Bytecode
            int    codeSize,   // Size
            int    offset,     // Current PC
            int    stackSize,  // Stack size on entry
            byte[] stackSizes, // Stack sizes in code
            Vector constantPool
        ) {
            for (;;) {
//System.out.println("Offset = " + offset + ", stack size = " + stackSize);

                // Check current bytecode offset.
                if (offset < 0 || offset >= codeSize) throw new RuntimeException("Offset out of range");

                // Have we hit an area that has already been analyzed?
                byte css = stackSizes[offset];
                if (css == stackSize) return; // OK.
                if (css == INVALID_OFFSET) throw new RuntimeException("Invalid offset");
                if (css != UNEXAMINED) {
//                  return;
                    throw new RuntimeException("Operand stack inconsistent at offset " + offset + ": Previous size " + css + ", now " + stackSize);
                }
                stackSizes[offset] = (byte) stackSize;

                // Analyze current opcode.
                int opcode = 0xff & code[offset];
                int operandOffset = offset + 1;
                short props;
                if (opcode == Opcode.WIDE) {
                    opcode = 0xff & code[operandOffset++];
                    props = Opcode.WIDE_OPCODE_PROPERTIES[opcode];
                } else {
                    props = Opcode.OPCODE_PROPERTIES[opcode];
                }
                if (props == Opcode.INVALID_OPCODE) throw new RuntimeException("Invalid opcode");

                switch (props & Opcode.SD_MASK) {

                case Opcode.SD_M4:
                case Opcode.SD_M3:
                case Opcode.SD_M2:
                case Opcode.SD_M1:
                case Opcode.SD_P0:
                case Opcode.SD_P1:
                case Opcode.SD_P2:
                    stackSize += (props & Opcode.SD_MASK) - Opcode.SD_P0;
                    if (stackSize < 0) throw new RuntimeException("Operand stack underrun at offset " + offset);
                    break;

                case Opcode.SD_0:
                    stackSize = 0;
                    break;

                case Opcode.SD_1:
                    stackSize = 0;
                    break;

                case Opcode.SD_GETFIELD:
                    stackSize += determineFieldSize(
                        constantPool,
                        (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
                    ) - 1;
                    break;

                case Opcode.SD_GETSTATIC:
                    stackSize += determineFieldSize(
                        constantPool,
                        (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
                    );
                    break;

                case Opcode.SD_PUTFIELD:
                    stackSize -= determineFieldSize(
                        constantPool,
                        (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
                    ) + 1;
                    break;

                case Opcode.SD_PUTSTATIC:
                    stackSize -= determineFieldSize(
                        constantPool,
                        (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
                    );
                    break;

                case Opcode.SD_INVOKEVIRTUAL:
                    stackSize -= 1 + determineArgumentsSize(
                        constantPool,
                        (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
                    );
                    break;

                case Opcode.SD_INVOKESPECIAL:
                    stackSize -= 1 + determineArgumentsSize(
                        constantPool,
                        (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
                    );
                    break;

                case Opcode.SD_INVOKESTATIC:
                    stackSize -= determineArgumentsSize(
                        constantPool,
                        (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
                    );
                    break;

                case Opcode.SD_INVOKEINTERFACE:
                    stackSize -= 1 + determineArgumentsSize(
                        constantPool,
                        (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
                    );
                    break;

                case Opcode.SD_MULTIANEWARRAY:
                    stackSize -= code[operandOffset + 2] - 1;
                    break;

                default:
                    throw new RuntimeException("Invalid stack delta");
                }

                if (stackSize > Byte.MAX_VALUE) throw new RuntimeException("Operand stack overflow");

                switch (props & Opcode.OP1_MASK) {

                case 0:
                    ;
                    break;

                case Opcode.OP1_SB:
                case Opcode.OP1_UB:
                case Opcode.OP1_CP1:
                case Opcode.OP1_LV1:
                    ++operandOffset;
                    break;

                case Opcode.OP1_SS:
                case Opcode.OP1_CP2:
                case Opcode.OP1_LV2:
                    operandOffset += 2;
                    break;

                case Opcode.OP1_BO2:
                case Opcode.OP1_JSR:
//System.out.println("Offset = " + offset);
//System.out.println("Operand offset = " + operandOffset);
//System.out.println((int) code[operandOffset]);
//System.out.println((int) code[operandOffset + 1]);
                    flowAnalysis(
                        code, codeSize,
                        offset + (
                            ((       code[operandOffset++]) << 8) +
                            ((0xff & code[operandOffset++])     )
                        ),
                        stackSize + (
                            (props & Opcode.OP1_MASK) == Opcode.OP1_JSR ? 1 : 0
                        ),
                        stackSizes,
                        constantPool
                    );
                    break;

                case Opcode.OP1_BO4:
                    flowAnalysis(
                        code, codeSize,
                        offset + (
                            ((       code[operandOffset++]) << 24) +
                            ((0xff & code[operandOffset++]) << 16) +
                            ((0xff & code[operandOffset++]) <<  8) +
                            ((0xff & code[operandOffset++])      )
                        ),
                        stackSize, stackSizes,
                        constantPool
                    );
                    break;

                case Opcode.OP1_LOOKUPSWITCH:
                    while ((operandOffset & 3) != 0) ++operandOffset;
                    flowAnalysis(
                        code, codeSize,
                        offset + (
                            ((       code[operandOffset++]) << 24) +
                            ((0xff & code[operandOffset++]) << 16) +
                            ((0xff & code[operandOffset++]) <<  8) +
                            ((0xff & code[operandOffset++])      )
                        ),
                        stackSize, stackSizes,
                        constantPool
                    );
                    int npairs = (
                        ((       code[operandOffset++]) << 24) +
                        ((0xff & code[operandOffset++]) << 16) +
                        ((0xff & code[operandOffset++]) <<  8) +
                        ((0xff & code[operandOffset++])      )
                    );
                    for (int i = 0; i < npairs; ++i) {
                        operandOffset += 4;
                        flowAnalysis(
                            code, codeSize,
                            offset + (
                                ((       code[operandOffset++]) << 24) +
                                ((0xff & code[operandOffset++]) << 16) +
                                ((0xff & code[operandOffset++]) <<  8) +
                                ((0xff & code[operandOffset++])      )
                            ),
                            stackSize, stackSizes,
                            constantPool
                        );
                    }
                    break;

                case Opcode.OP1_TABLESWITCH:
                    while ((operandOffset & 3) != 0) ++operandOffset;
                    flowAnalysis(
                        code, codeSize,
                        offset + (
                            ((       code[operandOffset++]) << 24) +
                            ((0xff & code[operandOffset++]) << 16) +
                            ((0xff & code[operandOffset++]) <<  8) +
                            ((0xff & code[operandOffset++])      )
                        ),
                        stackSize, stackSizes,
                        constantPool
                    );
                    int low = (
                        ((       code[operandOffset++]) << 24) +
                        ((0xff & code[operandOffset++]) << 16) +
                        ((0xff & code[operandOffset++]) <<  8) +
                        ((0xff & code[operandOffset++])      )
                    );
                    int hi = (
                        ((       code[operandOffset++]) << 24) +
                        ((0xff & code[operandOffset++]) << 16) +
                        ((0xff & code[operandOffset++]) <<  8) +
                        ((0xff & code[operandOffset++])      )
                    );
                    for (int i = low; i <= hi; ++i) {
                        flowAnalysis(
                            code, codeSize,
                            offset + (
                                ((       code[operandOffset++]) << 24) +
                                ((0xff & code[operandOffset++]) << 16) +
                                ((0xff & code[operandOffset++]) <<  8) +
                                ((0xff & code[operandOffset++])      )
                            ),
                            stackSize, stackSizes,
                            constantPool
                        );
                    }
                    break;

                default:
                    throw new RuntimeException("Invalid OP1");
                }

                switch (props & Opcode.OP2_MASK) {

                case 0:
                    ;
                    break;

                case Opcode.OP2_SB:
                    ++operandOffset;
                    break;

                case Opcode.OP2_SS:
                    operandOffset += 2;
                    break;

                default:
                    throw new RuntimeException("Invalid OP2");
                }

                switch (props & Opcode.OP3_MASK) {

                case 0:
                    ;
                    break;

                case Opcode.OP3_SB:
                    ++operandOffset;
                    break;

                default:
                    throw new RuntimeException("Invalid OP3");
                }

                Arrays.fill(stackSizes, offset + 1, operandOffset, INVALID_OFFSET);

                if ((props & Opcode.NO_FALLTHROUGH) != 0) return;
                offset = operandOffset;
            }
        }

        public void fixUp() {
            for (Offset o = this.beginning; o != this.end; o = o.next) {
                if (o instanceof FixUp) ((FixUp) o).fixUp();
            }
        }

        public void relocate() {
            for (int i = 0; i < this.relocatables.size(); ++i) {
                ((Relocatable) this.relocatables.elementAt(i)).relocate();
            }
        }

        /**
         * Analyse the descriptor of the Fieldref and return its size.
         */
        private int determineFieldSize(Vector constantPool, int idx) {
            ConstantFieldrefInfo    fld = (ConstantFieldrefInfo)    constantPool.elementAt(idx);
            ConstantNameAndTypeInfo nat = (ConstantNameAndTypeInfo) constantPool.elementAt(fld.nameAndTypeIndex);
            ConstantUtf8Info        u8i = (ConstantUtf8Info)        constantPool.elementAt(nat.descriptorIndex);
            return Descriptor.size(u8i.s);
        }

        /**
         * Analyse the descriptor of the Methodref and return the sum of the
         * arguments' sizes minus the return value's size.
         */
        private int determineArgumentsSize(Vector constantPool, int idx) {
            ConstantMethodrefInfo   mth = (ConstantMethodrefInfo)   constantPool.elementAt(idx);
            ConstantNameAndTypeInfo nat = (ConstantNameAndTypeInfo) constantPool.elementAt(mth.nameAndTypeIndex);
            ConstantUtf8Info        u8i = (ConstantUtf8Info)        constantPool.elementAt(nat.descriptorIndex);
            String desc = u8i.s;

            if (desc.charAt(0) != '(') throw new RuntimeException("Method descriptor does not start with \"(\"");
            int i = 1;
            int res = 0;
            for (;;) {
                switch (desc.charAt(i++)) {
                case ')':
                    return res - Descriptor.size(desc.substring(i));
                case 'B': case 'C': case 'F': case 'I': case 'S': case 'Z':
                    res += 1;
                    break;
                case 'D': case 'J':
                    res += 2;
                    break;
                case '[':
                    res += 1;
                    while (desc.charAt(i++) == '[');
                    if ("BCFISZDJ".indexOf(desc.charAt(i)) != -1) { ++i; break; }
                    if (desc.charAt(i) != 'L') throw new RuntimeException("Invalid char aftfer \"[\"");
                    ++i;
                    while (desc.charAt(i++) != ';');
                    break;
                case 'L':
                    res += 1;
                    while (desc.charAt(i++) != ';');
                    break;
                default:
                    throw new RuntimeException("Invalid method descriptor");
                }
            }
        }

        private short maxStack  = 0;
        private short maxLocals = 0;

        public void write(byte[] b) {
            if (b.length == 0) return;
            int co = currentInserter.offset;
            if (end.offset + b.length <= code.length) {
                System.arraycopy(code, co, code, co + b.length, end.offset - co);
            } else {
                byte[] oldCode = code;
                code = new byte[code.length + SIZE_INCREMENT];
                System.arraycopy(oldCode, 0, code, 0, co);
                System.arraycopy(oldCode, co, code, co + b.length, end.offset - co);
            }
            System.arraycopy(b, 0, code, co, b.length);
            for (Offset o = currentInserter; o != null; o = o.next) o.offset += b.length;
        }
//      public void write(byte[] b, int off, int len)
        public void write(int b) { write(new byte[] { (byte) b }); }
//      public void writeBoolean(boolean v)
        public void writeByte(int v) { write(new byte[] { (byte) v }); }
//      public void writeBytes(String s)
        public void writeChar(int v) { write(new byte[] { (byte) (v >> 8), (byte) v }); }
//      public void writeChars(String s)
        public void writeDouble(double v) { writeLong(Double.doubleToLongBits(v)); }
        public void writeFloat(float v) { writeInt(Float.floatToIntBits(v)); }
        public void writeInt(int v) { write(new byte[] { (byte) (v >> 24), (byte) (v >> 16), (byte) (v >> 8), (byte) v }); }
        public void writeLong(long v) { write(new byte[] { (byte) (v >> 56), (byte) (v >> 48), (byte) (v >> 40), (byte) (v >> 32), (byte) (v >> 24), (byte) (v >> 16), (byte) (v >> 8), (byte) v }); }
        public void writeShort(int v) { write(new byte[] { (byte) (v >> 8), (byte) v }); }
//      public void writeUTF(String str)

        public void writeOpcode(int opcode) {
            writeByte(opcode);
        }

        public void writeBranch(int opcode, final Offset dst) {
            relocatables.add(new Branch(newOffset(), dst));
            writeOpcode(opcode);
            writeShort(Short.MIN_VALUE);
        }

        private class Branch extends Relocatable {
            public Branch(Offset source, Offset destination) {
                this.source = source;
                this.destination = destination;
            }
            public void relocate() {
                if (destination.offset == -1) throw new RuntimeException("Cannot relocate branch to unset destination offset");
                int offset = destination.offset - source.offset;
                if (offset > Short.MAX_VALUE || offset < Short.MIN_VALUE) throw new RuntimeException("Branch offset out of range");
                byte[] ba = new byte[] { (byte) (offset >> 8), (byte) offset };
                System.arraycopy(ba, 0, code, source.offset + 1, 2);
            }
            private final Offset source;
            private final Offset destination;
        }

        public void writeOffset(Offset src, final Offset dst) {
            relocatables.add(new OffsetBranch(this.newOffset(), src, dst));
            this.writeInt(Integer.MIN_VALUE);
        }

        private class OffsetBranch extends Relocatable {
            public OffsetBranch(Offset where, Offset source, Offset destination) {
                this.where       = where;
                this.source      = source;
                this.destination = destination;
            }
            public void relocate() {
                if (
                    this.source.offset == -1 ||
                    this.destination.offset == -1
                ) throw new RuntimeException("Cannot relocate offset branch to unset destination offset");
                int offset = this.destination.offset - this.source.offset;
                byte[] ba = new byte[] {
                    (byte) (offset >> 24),
                    (byte) (offset >> 16),
                    (byte) (offset >> 8),
                    (byte) offset
                };
                System.arraycopy(ba, 0, code, this.where.offset, 4);
            }
            private final Offset where, source, destination;
        }

        public Offset newOffset() {
            Offset o = new Offset();
            o.set();
            return o;
        }

        /**
         * Allocate an {@Inserter}, set it to the current offset, and insert it
         * before the current offset.
         *
         * In clear text, this means that you can continue writing to the
         * "Code" attribute, then {@link #pushInserter(ClassFile.CodeAttribute.Inserter)} the
         * {@link Inserter}, then write again (which inserts bytes into the
         * "Code" attribute at the previously remembered position), and then
         * {@link #popInserter()}.
         */
        public Inserter newInserter() {
            Inserter i = new Inserter();
            i.set();
            return i;
        }

        /**
         * Remember the current {@link Inserter}, then replace it with the
         * new one.
         */
        public void pushInserter(Inserter ins) {
            if (ins.nextInserter != null) throw new RuntimeException("An Inserter can only be pushed once at a time");
            ins.nextInserter = this.currentInserter;
            this.currentInserter = ins;
        }

        /**
         * Replace the current {@link Inserter} with the remembered one (see
         * {@link #pushInserter(ClassFile.CodeAttribute.Inserter)}).
         */
        public void popInserter() {
            Inserter ni = currentInserter.nextInserter;
            if (ni == null) throw new RuntimeException("Code inserter stack underflow");
            currentInserter.nextInserter = null; // Mark it as "unpushed".
            currentInserter = ni;
        }

        /**
         * A class that represents an offset within a "Code" attribute.
         *
         * The concept of an "offset" is that if one writes into the middle of
         * a "Code" attribute, all offsets behind the insertion point are
         * automatically shifted.
         */
        public class Offset {

            /**
             * Set this "Offset" to the offset of the current inserter; insert
             * this "Offset" before the current inserter.
             */
            public void set() {
                if (this.offset != -1) throw new RuntimeException("Cannot \"set()\" Offset more than once");

                this.offset = CodeAttribute.this.currentInserter.offset;
                this.prev = CodeAttribute.this.currentInserter.prev;
                this.next = CodeAttribute.this.currentInserter;

                this.prev.next = this;
                this.next.prev = this;
            }
            public final CodeAttribute getCodeAttribute() { return CodeAttribute.this; }
            int    offset = -1;
            Offset prev = null, next = null;
        }

        public void addException(
            Offset startPC,
            Offset endPC,
            Offset handlerPC,
            String catchTypeFD
        ) {
            this.exceptionTable.add(new ExceptioN(
                startPC,
                endPC,
                handlerPC,
                catchTypeFD == null ? 0 : ClassFile.this.addConstantClassInfo(catchTypeFD)
            ));
        }

        private class ExceptioN {
            public ExceptioN(
                Offset startPC,
                Offset endPC,
                Offset handlerPC,
                short  catchType
            ) {
                this.startPC   = startPC;
                this.endPC     = endPC;
                this.handlerPC = handlerPC;
                this.catchType = catchType;
            }
            private final Offset startPC, endPC, handlerPC;
            private final short  catchType; // 0 == "finally" clause
        }

        /**
         * A class that implements an insertion point into a "Code"
         * attribute.
         */
        public class Inserter extends Offset {
            private Inserter nextInserter = null; // null == not in "currentInserter" stack
        }
        private abstract class Relocatable {
            public abstract void relocate();
        }
        private final int INITIAL_SIZE   = 100;
        private final int SIZE_INCREMENT = 128;
        private byte[]         code = new byte[INITIAL_SIZE];
        private final Offset   beginning = new Offset();
        private final Inserter end = new Inserter();
        {
            beginning.offset = end.offset = 0;
            beginning.next = end;
            end.prev = beginning;
        }
        private Inserter currentInserter = end;

        private Vector                attributes = new Vector(); // AttributeInfo

        private short        localVariableArrayLength = 0;
        private final Stack  savedLocalVariableArrayLengths = new Stack();
        private final Vector relocatables = new Vector();

        private final Vector exceptionTable = new Vector(); // ExceptioN
    }

    /**
     * A throw-in interface that marks {@Offset}s as "fix-ups": During the
     * execution of {@link ClassFile.CodeAttribute.fixUp}, all "fix-ups" are
     * invoked and can do last touches to the code attribute.
     * <p>
     * This is currently used for inserting the "padding bytes" into the
     * TABLESWITCH and LOOKUPSWITCH instructions.
     */
    /*package*/ interface FixUp {
        void fixUp();
    }
}
