
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.util.*;
import java.io.*;

/**
 * A simplified substitute for the <tt>javac</tt> tool.
 *
 * Usage:
 * <pre>
 * java net.janino.Parser [ -o <output-dir> ] <source-file> ...
 * </pre>
 */
public class Compiler {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/Compiler.java,v 1.4 2002/04/27 13:30:27 arno Exp $";

    public static void main(String[] args) throws Exception {
        String outputDirectory = null;

        int i;
        for (i = 0; i < args.length; ++i) {
            String arg = args[i];
            if (arg.charAt(0) != '-') break;
            if (arg.equals("-o")) {
                outputDirectory = args[++i];
            } else {
                System.err.println("Unrecognized command line option \"" + arg + "\".");
                System.exit(1);
            }
        }
        for (; i < args.length; ++i) {
            String sourceFileName = args[i];

            // Parse one compilation unit.
            Scanner              scanner = null;
            Java.CompilationUnit compilationUnit;
            try {
                scanner = new Scanner(sourceFileName);
                compilationUnit = new Parser(scanner).parseCompilationUnit();
            } finally {
                if (scanner != null) scanner.close();
            }

            // Compile the compilation unit.
            Vector v = new Vector(); // ClassFile
            compilationUnit.compile(v);

            // Store the compiled classes and interfaces into class files.
            for (int j = 0; j < v.size(); ++j) {
                ClassFile cf = (ClassFile) v.elementAt(j);

                // Verify that class name and file name match if the class is
                // declared "public" (7.6).
                if ((cf.accessFlags & Mod.PUBLIC) != 0) {
                    String className = cf.thisClassName;
                    { int idx = className.indexOf('$'); if (idx != -1) className = className.substring(0, idx); }
                    if (!sourceFileName.endsWith(className.replace('.', File.separatorChar) + ".java")) {
                        System.err.println("Class or interface \"" + className + "\" is declared \"public\", but the source file name \"" + sourceFileName + "\" does not match the class name.");
                        System.exit(1);
                    }
                }

                String classFileName = cf.thisClassName.replace('.', File.separatorChar) + ".class";
                if (outputDirectory != null) {
                    classFileName = outputDirectory + File.separator + classFileName;
                }

                // Create directory for class file is it does not exist.
                {
                    File dir = new File(classFileName).getParentFile();
                    if (dir != null && !dir.isDirectory()) {
                        if (!dir.mkdirs()) throw new IOException("Cannot create directory for class file \"" + classFileName + "\"");
                    }
                }

                FileOutputStream fos = new FileOutputStream(classFileName);
                try {
                    cf.store(fos);
                } finally {
                    fos.close();
                }
            }
        }
    }
}
