
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.util.*;

/**
 * Helper class that defines useful methods for handling "field descriptors"
 * (JVMS 4.3.2) and "method descriptors" (JVMS 4.3.3).
 */
class Descriptor {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/Descriptor.java,v 1.5 2002/05/25 20:59:03 arno Exp $";

    public static boolean isReference(String d) {
        return d.length() > 1;
    }
    public static boolean isClassOrInterfaceReference(String d) {
        return d.charAt(0) == 'L';
    }
    public static boolean isArrayReference(String d) {
        return d.charAt(0) == '[';
    }
    public static short size(String d) {
        if (d.equals(Descriptor.VOID)) return 0;
        if (hasSize1(d)) return 1;
        if (hasSize2(d)) return 2;
        throw new RuntimeException("No size defined for type \"" + toString(d) + "\"");
    }
    public static boolean hasSize1(String d) {
        if (d.length() == 1) return "BCFISZ".indexOf(d) != -1;
        return isReference(d);
    }
    public static boolean hasSize2(String d) {
        return d.equals(LONG) || d.equals(DOUBLE);
    }

    // Pretty-print.
    public static String toString(String d) {
        int idx = 0;
        StringBuffer sb = new StringBuffer();
        if (d.charAt(0) == '(') {
            ++idx;
            sb.append("(");
            while (idx < d.length() && d.charAt(idx) != ')') {
                if (idx != 1) sb.append(", ");
                idx = toString(d, idx, sb);
            }
            if (idx >= d.length()) throw new RuntimeException("Invalid descriptor \"" + d + "\"");
            sb.append(") => ");
            ++idx;
        }
        toString(d, idx, sb);
        return sb.toString();
    }
    private static int toString(String d, int idx, StringBuffer sb) {
        int dimensions = 0;
        while (idx < d.length() && d.charAt(idx) == '[') {
            ++dimensions;
            ++idx;
        }
        if (idx >= d.length()) throw new RuntimeException("Invalid descriptor \"" + d + "\"");
        switch (d.charAt(idx)) {
        case 'L':
            {
                int idx2 = d.indexOf(';', idx);
                if (idx2 == -1) throw new RuntimeException("Invalid descriptor \"" + d + "\"");
                sb.append(d.substring(idx + 1, idx2).replace('/', '.'));
                idx = idx2;
            }
            break;
        case 'V': sb.append("void");    break;
        case 'B': sb.append("byte");    break;
        case 'C': sb.append("char");    break;
        case 'D': sb.append("double");  break;
        case 'F': sb.append("float");   break;
        case 'I': sb.append("int");     break;
        case 'J': sb.append("long");    break;
        case 'S': sb.append("short");   break;
        case 'Z': sb.append("boolean"); break;
        default:
            throw new RuntimeException("Invalid descriptor \"" + d + "\"");
        }
        for (; dimensions > 0; --dimensions) sb.append("[]");
        return idx + 1;
    }

    public static String fromClassName(String className) {
        if (className.equals("void"   )) return Descriptor.VOID;
        if (className.equals("byte"   )) return Descriptor.BYTE;
        if (className.equals("char"   )) return Descriptor.CHAR;
        if (className.equals("double" )) return Descriptor.DOUBLE;
        if (className.equals("float"  )) return Descriptor.FLOAT;
        if (className.equals("int"    )) return Descriptor.INT;
        if (className.equals("long"   )) return Descriptor.LONG;
        if (className.equals("short"  )) return Descriptor.SHORT;
        if (className.equals("boolean")) return Descriptor.BOOLEAN;
        return 'L' + className.replace('.', '/') + ';';
    }

    public static String toClassName(String d) {
        char firstChar = d.charAt(0);
        if (firstChar == 'L') return d.substring(1, d.length() - 1).replace('/', '.');
        if (firstChar == '[') return d.replace('/', '.');
        throw new RuntimeException("Descriptor \"" + d + "\" does not describe a class");
    }

    public final static String VOID    = "V";
    public final static String BYTE    = "B";
    public final static String CHAR    = "C";
    public final static String DOUBLE  = "D";
    public final static String FLOAT   = "F";
    public final static String INT     = "I";
    public final static String LONG    = "J";
    public final static String SHORT   = "S";
    public final static String BOOLEAN = "Z";
    public final static String OBJECT  = "Ljava/lang/Object;";
    public final static String STRING  = "Ljava/lang/String;";
}
