
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/**
 * An expression evaluator that evaluates expressions in Java bytecode.
 */

public class ExpressionEvaluator {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/ExpressionEvaluator.java,v 1.9 2002/05/19 22:10:48 arno Exp $";

    /**
     * Construct an expression evaluator that evaluates the given expression
     * with the given type, parameter names and types.
     *
     * <p>
     *   "parameterNames" and "parameterTypes" must have the same length.
     * </p>
     *
     * @param expression The expression to evaluate, e.g. "i * j".
     *
     * @param expressionType The type of the expression, e.g.
     * "Double.TYPE".
     *
     * @param parameterNames The names of the expression parameters, e.g.
     * "i" and "j".
     *
     * @param parameterTypes The types of the expression parameters, e.g.
     * "Integer.TYPE" or "Double.TYPE".
     */
    public ExpressionEvaluator(
        String   expression,
        Class    expressionType,
        String[] parameterNames,
        Class[]  parameterTypes
    ) throws Java.CompileException, Parser.ParseException, Scanner.ScanException, IOException {
        if (parameterNames.length != parameterTypes.length) throw new RuntimeException("Lengths of \"parameterNames\" and \"parameterTypes\" do not match");

        // Parse Java source code.
        Java.CompilationUnit compilationUnit = new Java.CompilationUnit();
        Scanner scanner = new Scanner(null, new StringReader(expression));
        Parser parser = new Parser(scanner);
        while (scanner.peek().isKeyword("import")) parser.parseImportDeclaration(compilationUnit);
        Java.ClassDeclaration cd = new Java.ClassDeclaration(
            scanner.peek().getLocation(), // location
            (Java.Scope) compilationUnit, // enclosingScope
            Mod.PUBLIC,                   // modifiers
            "SC",                         // name
            (Java.ReferenceType) null,    // extendedType
            new Java.ReferenceType[0]     // implementedTypes
        );
        compilationUnit.addClassOrInterfaceDeclaration(cd);
        Java.FormalParameter[] fps = new Java.FormalParameter[parameterNames.length];
        for (int i = 0; i < fps.length; ++i) {
            fps[i] = new Java.FormalParameter(
                scanner.peek().getLocation(),     // location
                true,                             // finaL
                new ClassType(                    // type
                    scanner.peek().getLocation(),
                    parameterTypes[i]
                ),
                parameterNames[i]                 // name
            );
        }
        Java.MethodDeclarator md = new Java.MethodDeclarator(
            scanner.peek().getLocation(), // location
            cd,                           // declaringClassOrInterface
            (short) (                     // modifiers
                Mod.PUBLIC |
                Mod.STATIC
            ),
            new ClassType(                // type
                scanner.peek().getLocation(),
                expressionType
            ),
            "eval",                       // name
            fps,                          // formalParameters
            new Java.ReferenceType[0]     // thrownExceptions
        );
        cd.addDeclaredMethod(md);
        Java.Block b = new Java.Block(
            scanner.peek().getLocation(), // location
            (Java.Scope) md               // enclosingScope
        );
        md.setBody(b);
        b.addStatement(new Java.ReturnStatement(
            scanner.peek().getLocation(), // location
            b,                            // enclosingBlock
            parser.parseExpression(       // returnValue
                (Java.Scope) b
            ).toRvalue()
        ));
        if (!scanner.peek().isEOF()) throw new Parser.ParseException("Unexpected token \"" + scanner.peek() + "\"", scanner.peek().getLocation());

        // Compile.
        Vector v = new Vector(); // ClassFile
        compilationUnit.compile(v);

        // Get compiled class.
        if (v.size() != 1) throw new RuntimeException("Compiles to " + v.size() + " classes!?");

        // Generate bytecode and store it in a byte array.
        ClassFile cf = (ClassFile) v.elementAt(0);
        ByteArrayOutputStream baos2 = new ByteArrayOutputStream();
        try {
            cf.store(baos2);
        } catch (IOException ex) {
            throw new RuntimeException(ex.toString());
        }
        final byte[] ba = baos2.toByteArray();

        // Load bytecode.
        Class c = new ClassLoader() {
            public Class main() {
                return defineClass(null, ba, 0, ba.length);
            }
        }.main();

        // Get method "eval()".
        this.parameterTypes = parameterTypes;
        try {
            this.method = c.getMethod("eval", parameterTypes);
        } catch (NoSuchMethodException ex) {
            throw new RuntimeException(ex.toString());
        }
        if (this.method == null) throw new RuntimeException("Eval method not found");
    }

    private static class ClassType extends Java.Type {
        public ClassType(
            Scanner.Location location,
            Class            clazz
        ) {
            super(location);
            this.clazz = clazz;
        }
        public IClass getType() { return ReflectionIClass.fromClass(clazz); }
        private final Class clazz;
    }

    /**
     * Evaluates an expression with concrete parameter values.
     *
     * <p>
     *   Each parameter value must have the same type as specified through
     *   the "parameterTypes" parameter of
     *   {@link net.janino.ExpressionEvaluator#ExpressionEvaluator(String,
     *   Class, String[], Class[])}.
     * </p>
     * <p>
     *   Parameters of primitive type must passed with their wrapper class
     *   objects.
     * </p>
     * <p>
     *   The object returned has the class specified through the "returnType"
     *   parameter of
     *   {@link net.janino.ExpressionEvaluator#ExpressionEvaluator(String,
     *   Class, String[], Class[])}.
     * </p>
     *
     * @param parameterValues The concrete parameter values.
     */
    public Object evaluate(Object[] parameterValues) throws InvocationTargetException {
        try {
            return this.method.invoke(null, parameterValues);
        } catch (IllegalAccessException ex) {
            throw new RuntimeException(ex.toString());
        }
    }

    private final Class[] parameterTypes;
    private final Method  method;
}
