
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.util.*;

/**
 * A simplified equivalent to "java.lang.reflect".
 */
public abstract class IClass {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/IClass.java,v 1.8 2002/05/25 20:59:03 arno Exp $";

    public abstract IConstructor[] getDeclaredIConstructors();
    public abstract IMethod[]      getDeclaredIMethods();
    public abstract IField[]       getDeclaredIFields();
    public abstract IClass[]       getDeclaredIClasses();

    /**
     * Returns "null" for class "Object" and for primitive types.
     *
     * @return The superclass.
     */
    public abstract IClass         getSuperclass() throws Java.CompileException;
    public abstract IClass[]       getImplementedInterfaces() throws Java.CompileException;
    public abstract String         getDescriptor();
    public abstract boolean        isArray();
    public abstract IClass         getComponentType();
    public final IClass getArrayType() {
        if (this.myArrayType == null) this.myArrayType = new ArrayIClass(this);
        return this.myArrayType;
    }
    public abstract boolean        isPrimitive();
    public abstract boolean        isPrimitiveNumeric();

    /**
     * Determine if "this" is assignable from "that". This is true if "this"
     * is identical with "that", or if "this" is assignable from "that"'s
     * superclass (if any), or if "this" is assignable from any of the
     * interfaces which "that" implements.
     */
    public boolean isAssignableFrom(IClass that) throws Java.CompileException {
        if (this == that) return true;

        IClass tsc = that.getSuperclass();
        if (tsc != null && this.isAssignableFrom(tsc)) return true;

        IClass[] tii = that.getImplementedInterfaces();
        for (int i = 0; i < tii.length; ++i) {
            if (this.isAssignableFrom(tii[i])) return true;
        }
        return false;
    }

    /** The singleton that represents the "void" type. */
    public static final IClass VOID    = ReflectionIClass.fromClass(Void.TYPE);
    /** The singleton that represents the "byte" type. */
    public static final IClass BYTE    = ReflectionIClass.fromClass(Byte.TYPE);
    /** The singleton that represents the "char" type. */
    public static final IClass CHAR    = ReflectionIClass.fromClass(Character.TYPE);
    /** The singleton that represents the "double" type. */
    public static final IClass DOUBLE  = ReflectionIClass.fromClass(Double.TYPE);
    /** The singleton that represents the "float" type. */
    public static final IClass FLOAT   = ReflectionIClass.fromClass(Float.TYPE);
    /** The singleton that represents the "int" type. */
    public static final IClass INT     = ReflectionIClass.fromClass(Integer.TYPE);
    /** The singleton that represents the "long" type. */
    public static final IClass LONG    = ReflectionIClass.fromClass(Long.TYPE);
    /** The singleton that represents the "short" type. */
    public static final IClass SHORT   = ReflectionIClass.fromClass(Short.TYPE);
    /** The singleton that represents the "boolean" type. */
    public static final IClass BOOLEAN = ReflectionIClass.fromClass(Boolean.TYPE);

    /** The singleton that represents the "java.lang.Object" class. */
    public static final IClass OBJECT  = ReflectionIClass.fromClass(Object.class);
    /** The singleton that represents the "java.lang.String" class. */
    public static final IClass STRING  = ReflectionIClass.fromClass(String.class);

    /** The singleton that represents the "java.lang.Throwable" class. */
    public static final IClass THROWABLE         = ReflectionIClass.fromClass(Throwable.class);
    /** The singleton that represents the "java.lang.RuntimeException" class. */
    public static final IClass RUNTIME_EXCEPTION = ReflectionIClass.fromClass(RuntimeException.class);
    /** The singleton that represents the "java.lang.Error" class. */
    public static final IClass ERROR             = ReflectionIClass.fromClass(Error.class);

    private ArrayIClass myArrayType = null; // Determined on demand.

    public abstract class IMember {

        /**
         * @return One of {@link #PRIVATE}, {@link #PROTECTED},
         * {@link #PACKAGE} and {@link #PUBLIC}.
         */
        public abstract int getAccess();
    }
    /** Return value for {@link IMember#getAccess}. */
    public static final int PRIVATE   = 0;
    /** Return value for {@link IMember#getAccess}. */
    public static final int PROTECTED = 1;
    /** Return value for {@link IMember#getAccess}. */
    public static final int PACKAGE   = 2;
    /** Return value for {@link IMember#getAccess}. */
    public static final int PUBLIC    = 3;

    public abstract class IInvocable extends IMember {
        public IClass   getDeclaringIClass() { return IClass.this; }
        public abstract IClass[] getParameterTypes() throws Java.CompileException;
        public abstract String   getDescriptor() throws Java.CompileException;
        public abstract IClass[] getThrownExceptions() throws Java.CompileException;

        public boolean isMoreSpecificThan(IInvocable that) throws Java.CompileException {
//System.out.print("\"" + this + "\".isMoreSpecificThan(\"" + that + "\") => ");
            if (!that.getDeclaringIClass().isAssignableFrom(this.getDeclaringIClass())) {
//System.out.println("false");
                return false;
            }
            IClass[] thisParameterTypes = this.getParameterTypes();
            IClass[] thatParameterTypes = that.getParameterTypes();
            int i;
            for (i = 0; i < thisParameterTypes.length; ++i) {
                if (!thatParameterTypes[i].isAssignableFrom(thisParameterTypes[i])) {
//System.out.println("false");
                    return false;
                }
            }
//System.out.println("true");
            return true;
        }
        public boolean isLessSpecificThan(IInvocable that) throws Java.CompileException {
            return that.isMoreSpecificThan(this);
        }
    }
    public abstract class IConstructor extends IInvocable {
        public String getDescriptor() throws Java.CompileException {
            StringBuffer sb = new StringBuffer("(");
            IClass[] parameterTypes = getParameterTypes();
            for (int i = 0; i < parameterTypes.length; ++i) {
                sb.append(parameterTypes[i].getDescriptor());
            }
            sb.append(")" + Descriptor.VOID);
            return sb.toString();
        }
    }
    public abstract class IMethod extends IInvocable {
        public abstract boolean isStatic();
        public abstract IClass  getReturnType() throws Java.CompileException;
        public abstract String  getName();
        public String getDescriptor() throws Java.CompileException {
            StringBuffer sb = new StringBuffer("(");
            IClass[] parameterTypes = getParameterTypes();
            for (int i = 0; i < parameterTypes.length; ++i) {
                sb.append(parameterTypes[i].getDescriptor());
            }
            sb.append(')');
            sb.append(getReturnType().getDescriptor());
            return sb.toString();
        }
    }
    public abstract class IField extends IMember {
        public IClass           getDeclaringIClass() { return IClass.this; }
        public abstract boolean isStatic();
        public abstract IClass  getType() throws Java.CompileException;
        public abstract String  getName();
        public String getDescriptor() throws Java.CompileException {
            return getType().getDescriptor();
        }
    }

    /**
     * Get {@link #IClass} by descriptor.
     */
    public static IClass iClassFor(final String type) {
        IClass res = (IClass) iClassForCache.get(type);
        if (res == null) {
            try {
                res = ReflectionIClass.fromClass(Class.forName(Descriptor.toClassName(type)));
            } catch (ClassNotFoundException ex) {
                return null;
            }
            iClassForCache.put(type, res);
        }
        return res;
    }
    private static final HashMap iClassForCache = new HashMap(); // String => IClass

    private static class ArrayIClass extends IClass {
        public ArrayIClass(IClass componentType) {
            this.componentType = componentType;
        }

        public IConstructor[] getDeclaredIConstructors() { return new IConstructor[0]; }
        public IMethod[] getDeclaredIMethods() { return new IMethod[0]; }
        public IField[] getDeclaredIFields() { return new IField[0]; }
        public IClass[] getDeclaredIClasses() { return new IClass[0]; }
        public IClass getSuperclass() { return IClass.OBJECT; }
        public IClass[] getImplementedInterfaces() { return new IClass[0]; }
        public String getDescriptor() { return '[' + componentType.getDescriptor(); }
        public boolean isArray() { return true; }
        public IClass getComponentType() { return componentType; }
        public boolean isPrimitive() { return false; }
        public boolean isPrimitiveNumeric() { return false; }

        public String toString() { return this.componentType.toString() + "[]"; }

        private final IClass componentType;
    }
}
