
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.util.*;

/**
 * This wrapper class defines classes that represent the elements of the
 * Java<sup>TM</sup> programming language.
 */

public class Java {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/Java.java,v 1.20 2002/05/25 20:59:03 arno Exp $";

    private Java() {} // Don't instantiate me.

    /*package*/ interface Scope {

        /**
         * @return Enclusing scope or "null".
         */
        public Scope getEnclosingScope();
    }

    /**
     * This interface is implemented by objects which are associated with a
     * location in the source code.
     */
    /*package*/ interface Locatable {
        public Scanner.Location getLocation();
        public void throwParseException(String message) throws Parser.ParseException;
        public void throwCompileException(String message) throws CompileException;
    }
    private static abstract class Located implements Locatable {
        Located(Scanner.Location location) {
            this.location = location;
        }
        public Scanner.Location getLocation() { return this.location; }
        public void throwParseException(String message) throws Parser.ParseException {
            throw new Parser.ParseException(message, location);
        }
        public void throwCompileException(String message) throws CompileException {
            throw new CompileException(message, location);
        }
        private final Scanner.Location location;
    }

    /**
     * Holds the result of {@link Parser#parseCompilationUnit}.
     *
     * <p>
     *   A call to {@link #compile(Vector)} generates a set of
     *   {@link ClassFile} objects which represent the classes and
     *   interfaces defined in the compilation unit.
     * </p>
     */
    /*package*/ static class CompilationUnit implements Scope {
        CompilationUnit() {}

        // Implement "Scope".
        public Scope getEnclosingScope() { return null; }

        public void setPackage(String packagE) {
            this.packagE = packagE;
        }
        public void addImportedClass(
            Scanner.Location location,
            String           className,
            String           fullClassName
        ) throws Parser.ParseException {
            String s = (String) this.importedClasses.get(className);
            if (s != null) {
                if (!s.equals(fullClassName)) throw new Parser.ParseException("Class \"" + className + "\" was first imported from \"" + s + "\", now again from \"" + fullClassName + "\"", location);
            } else {
                this.importedClasses.put(className, fullClassName);
            }
        }
        public void addImportedPackage(String packageName) {
            this.importedPackages.add(packageName);
        }
        public void addClassOrInterfaceDeclaration(ClassOrInterfaceDeclaration cid) {
            this.classOrInterfaceDeclarations.add(cid);
        }

        /**
         * Compiles this object and returns a set of {@link ClassFile}
         * objects.
         */
        public void compile(Vector classFiles_return) throws CompileException {
            for (int i = 0; i < classOrInterfaceDeclarations.size(); ++i) {
                ClassOrInterfaceDeclaration cid = (ClassOrInterfaceDeclaration) classOrInterfaceDeclarations.elementAt(i);
                cid.compile(classFiles_return);
            }
        }

        private String  packagE                      = null;          // null == no IMPORT directive
        private Vector  importedPackages             = new Vector();  // String
        { importedPackages.add("java.lang"); }
        private HashMap importedClasses              = new HashMap(); // String => String
        private Vector  classOrInterfaceDeclarations = new Vector();  // ClassOrInterfaceDeclaration
    }

    /*package*/ abstract static class ClassOrInterfaceDeclaration extends IClass implements Locatable, Scope {
        public ClassOrInterfaceDeclaration(
            Scanner.Location location,
            Scope            enclosingScope,
            short            modifiers,
            String           name
        ) throws Parser.ParseException {
            this.location       = location;
            this.enclosingScope = enclosingScope;
            this.modifiers      = modifiers;
            this.name           = name;

            if (enclosingScope instanceof ClassOrInterfaceDeclaration) {

                // Member type.
                ClassOrInterfaceDeclaration declaringClassOrInterface = (ClassOrInterfaceDeclaration) enclosingScope;

                // Check for redefinition of member type.
                Vector v = declaringClassOrInterface.declaredClasses;
                for (int i = 0; i < v.size(); ++i) {
                    ClassDeclaration cd = (ClassDeclaration) v.elementAt(i);
                    if (cd.name.equals(name)) this.throwParseException("Redefinition of class or interface \"" + name + "\"");
                }
                v = declaringClassOrInterface.declaredInterfaces;
                for (int i = 0; i < v.size(); ++i) {
                    InterfaceDeclaration id = (InterfaceDeclaration) v.elementAt(i);
                    if (id.name.equals(name)) throwParseException("Redefinition of class or interface \"" + name + "\"");
                }
            } else {

                // Top-level type.

                // Check for forbidden modifiers (7.6).
                if ((modifiers & (
                    Mod.PROTECTED |
                    Mod.PRIVATE |
                    Mod.STATIC
                )) != 0) throwParseException("Modifiers \"protected\", \"private\" and \"static\" not allowed in top level class or interface declaration");

                // Find compilation unit.
                Scope s;
                for (s = enclosingScope; s != null && !(s instanceof CompilationUnit); s = s.getEnclosingScope());
                if (s == null) this.throwParseException("Class or interface declaration not in compilation unit scope");
                CompilationUnit compilationUnit = (CompilationUnit) s;

                // Check for conflict with single-type-import (7.6).
                if (compilationUnit.importedClasses.containsKey(name)) throwParseException("Top level class or interface declaration \"" + name + "\" conflicts with single-type-import \"" + compilationUnit.importedClasses.get(name) + "\"");

                // Check for redefinition within compilation unit (7.6).
                Vector v = compilationUnit.classOrInterfaceDeclarations;
                for (int i = 0; i < v.size(); ++i) {
                    ClassOrInterfaceDeclaration id = (ClassOrInterfaceDeclaration) v.elementAt(i);
                    if (id.name.equals(name)) throwParseException("Redefinition of class or interface \"" + name + "\"");
                }
            }
        }

        public void addDeclaredMethod(MethodDeclarator method) {
            this.declaredMethods.add(method);
        }

        // Implement "Scope".
        public Scope getEnclosingScope() {
            return this.enclosingScope;
        }

        protected String getClassName() {
            if (this.thisClassName == null) {
                StringBuffer sb = new StringBuffer(this.name);
                Scope s = this.getEnclosingScope();
                while (s instanceof ClassOrInterfaceDeclaration) {
                    sb.insert(0, ((ClassOrInterfaceDeclaration) s).name + '$');
                    s = s.getEnclosingScope();
                }
                CompilationUnit compilationUnit = (CompilationUnit) s;
                if (compilationUnit.packagE != null) {
                    sb.insert(0, compilationUnit.packagE + '.');
                }
                this.thisClassName = sb.toString();
            }
            return this.thisClassName;
        }

        // Implement "Locatable".
        public Scanner.Location getLocation() { return this.location; }
        public void throwParseException(String message) throws Parser.ParseException {
            throw new Parser.ParseException(message, location);
        }
        public void throwCompileException(String message) throws CompileException {
            throw new CompileException(message, location);
        }

        // Implement "IClass".
        public IClass.IMethod[] getDeclaredIMethods() {
            IClass.IMethod[] res = new IClass.IMethod[this.declaredMethods.size()];
            for (int i = 0; i < res.length; ++i) {
                res[i] = ((MethodDeclarator) this.declaredMethods.elementAt(i)).toIMethod();
            }
            return res;
        }
        public IClass[] getDeclaredIClasses() {
            return (IClass[]) this.declaredClasses.toArray(new IClass[this.declaredClasses.size()]);
        }
        public String getDescriptor() {
            return Descriptor.fromClassName(this.getClassName());
        }
        public boolean isArray() { return false; }
        public IClass  getComponentType() { return null; }
        public boolean isPrimitive() { return false; }
        public boolean isPrimitiveNumeric() { return false; }

        public String toString() { return this.getClassName(); }

        // Compiles itself and adds itself to the compilation unit.
        abstract public void compile(Vector classFiles_return) throws CompileException;

        private /*final*/ Scanner.Location location;
        private /*final*/ Scope            enclosingScope;
        protected /*final*/ short          modifiers;
        protected /*final*/ String         name;
        protected final Vector         declaredMethods = new Vector(); // MethodDeclarator
        protected final Vector         declaredClasses = new Vector(); // ClassDeclaration
        protected final Vector         declaredInterfaces = new Vector(); // InterfaceDeclaration

        private String thisClassName = null;
    }

    /*package*/ static class ClassDeclaration extends ClassOrInterfaceDeclaration {
        public ClassDeclaration(
            Scanner.Location location,
            Scope            enclosingScope,
            short            modifiers,
            String           name,
            ReferenceType    extendedType, // null == none
            ReferenceType[]  implementedTypes
        ) throws Parser.ParseException {
            super(
                location,
                enclosingScope,
                modifiers,
                name
            );

            this.extendedType = extendedType == null ? new ReferenceType(
                location,
                null, // scope
                new String[] { "java", "lang", "Object" }
            ) : extendedType;
            this.implementedTypes = implementedTypes;
        }

        public void addConstructor(ConstructorDeclarator cd) {
            this.constructors.add(cd);
        }
        public void addInstanceVariableDeclaratorOrInstanceInitializer(
            BlockStatement bs
        ) {
            this.instanceVariableDeclaratorsAndInstanceInitializers.add(bs);
        }
        public void addClassVariableDeclaratorOrClassInitializer(
            BlockStatement bs
        ) {
            this.classVariableDeclaratorsAndClassInitializers.add(bs);
        }

        private /*final*/ ReferenceType   extendedType;
        private /*final*/ ReferenceType[] implementedTypes;
        private final Vector          constructors = new Vector(); // ConstructorDeclarator
        private final Vector          instanceVariableDeclaratorsAndInstanceInitializers = new Vector(); // BlockStatement
        private final Vector          classVariableDeclaratorsAndClassInitializers = new Vector(); // BlockStatement

        // Compile time members.

        // Implement "IClass".
        public IClass getSuperclass() throws CompileException {
            return this.extendedType.getType();
        }
        public IClass[] getImplementedInterfaces() throws CompileException {
            IClass[] res = new IClass[this.implementedTypes.length];
            for (int i = 0; i < res.length; ++i) {
                res[i] = this.implementedTypes[i].getType();
            }
            return res;
        }
        public IClass.IConstructor[] getDeclaredIConstructors() {
            IClass.IConstructor[] res = new IClass.IConstructor[this.constructors.size()];
            for (int i = 0; i < res.length; ++i) {
                res[i] = ((ConstructorDeclarator) this.constructors.elementAt(i)).toIConstructor();
            }
            return res;
        }
        public IClass.IField[] getDeclaredIFields() {
            Vector res = new Vector();

            // Determine static fields.
            for (int i = 0; i < this.classVariableDeclaratorsAndClassInitializers.size(); ++i) {
                BlockStatement cVDOCI = (BlockStatement) this.classVariableDeclaratorsAndClassInitializers.elementAt(i);
                if (cVDOCI instanceof FieldDeclarator) {
                    FieldDeclarator fd = (FieldDeclarator) cVDOCI;
                    IClass.IField[] flds = fd.getIFields();
                    for (int j = 0; j < flds.length; ++j) res.add(flds[j]);
                }
            }

            // Determine non-static fields.
            for (int i = 0; i < this.instanceVariableDeclaratorsAndInstanceInitializers.size(); ++i) {
                BlockStatement iVDOII = (BlockStatement) this.instanceVariableDeclaratorsAndInstanceInitializers.elementAt(i);
                if (iVDOII instanceof FieldDeclarator) {
                    FieldDeclarator fd = (FieldDeclarator) iVDOII;
                    IClass.IField[] flds = fd.getIFields();
                    for (int j = 0; j < flds.length; ++j) res.add(flds[j]);
                }
            }

            return (IClass.IField[]) res.toArray(new IClass.IField[res.size()]);
        }

        public void compile(Vector classFiles_return) throws CompileException {
            if (
                this.getEnclosingScope() instanceof ClassOrInterfaceDeclaration &&
                (this.modifiers & Mod.STATIC) == 0
            ) this.throwCompileException("Non-static member types (inner classes) NYI");

            // Determine superclass.
            this.superclass = extendedType.getType();

            // Determine implemented interfaces.
            this.interfaces = new IClass[this.implementedTypes.length];
            String[] interfaceDescriptors = new String[this.interfaces.length];
            for (int i = 0; i < implementedTypes.length; ++i) {
                this.interfaces[i] = this.implementedTypes[i].getType();
                interfaceDescriptors[i] = this.interfaces[i].getDescriptor();
            }

            // Create "ClassFile" object.
            ClassFile cf = new ClassFile(
                (short) (this.modifiers | Mod.SUPER), // accessFlags
                this.getClassName(),                  // thisClassName
                this.superclass.getDescriptor(),      // superClassFD
                interfaceDescriptors                  // interfaceNames
            );

            // Class variables.
            for (int i = 0; i < this.classVariableDeclaratorsAndClassInitializers.size(); ++i) {
                BlockStatement bs = (BlockStatement) this.classVariableDeclaratorsAndClassInitializers.elementAt(i);
                if (!(bs instanceof FieldDeclarator)) continue;
                FieldDeclarator fd = (FieldDeclarator) bs;
                for (int j = 0; j < fd.variableDeclarators.length; ++j) {
                    VariableDeclarator vd = fd.variableDeclarators[j];
                    Type type = fd.type;
                    for (int k = 0; k < vd.brackets; ++k) type = new ArrayType(type);

                    cf.addFieldInfo(
                        fd.modifiers,                   // modifiers
                        type.getType().getDescriptor(), // fieldTypeFD
                        vd.name,                        // fieldName
                        (                               // optionalConstantValue
                            (fd.modifiers & Mod.FINAL) != 0 &&
                            vd.optionalInitializer != null
                        ) ? vd.optionalInitializer.getConstantValue() : null
                    );
                }
            }

            // Instance variables.
            for (int i = 0; i < this.instanceVariableDeclaratorsAndInstanceInitializers.size(); ++i) {
                BlockStatement bs = (BlockStatement) this.instanceVariableDeclaratorsAndInstanceInitializers.elementAt(i);
                if (!(bs instanceof FieldDeclarator)) continue;
                FieldDeclarator fd = (FieldDeclarator) bs;
                for (int j = 0; j < fd.variableDeclarators.length; ++j) {
                    VariableDeclarator vd = fd.variableDeclarators[j];
                    Type type = fd.type;
                    for (int k = 0; k < vd.brackets; ++k) type = new ArrayType(type);
                    cf.addFieldInfo(
                        fd.modifiers,                   // modifiers,
                        type.getType().getDescriptor(), // fieldTypeFD,
                        vd.name,                        // fieldName,
                        null                            // optionalConstantValue
                    );
                }
            }

            // Class initialization method.
            if (!this.classVariableDeclaratorsAndClassInitializers.isEmpty()) {
                MethodDeclarator classInitializationMethod = new MethodDeclarator(
                    this.getLocation(),     // location
                    this,                   // declaringClassOrInterface
                    (short) (               // modifiers
                        Mod.STATIC |
                        Mod.PUBLIC
                    ),
                    new BasicType(          // type
                        this.getLocation(),
                        BasicType.VOID
                    ),
                    "<clinit>",             // name
                    new FormalParameter[0], // formalParameters
                    new ReferenceType[0]    // thrownExceptions
                );
                Block b = new Block(
                    this.getLocation(),
                    classInitializationMethod
                );
                b.addStatements(this.classVariableDeclaratorsAndClassInitializers);
                classInitializationMethod.setBody(b);
                classInitializationMethod.compile(cf);
            }

            // Constructors.
            for (int i = 0; i < constructors.size(); ++i) {
                ConstructorDeclarator cd = (ConstructorDeclarator) constructors.elementAt(i);
                cd.compile(cf);
            }
            // Add default constructor.
            if (constructors.size() == 0) {
                ConstructorDeclarator cd = new ConstructorDeclarator(
                    this.getLocation(),     // location
                    this,                   // declaringClass
                    Mod.PUBLIC,             // modifiers
                    new FormalParameter[0], // formalParameters
                    new ReferenceType[0]    // thrownExceptions
                );
                cd.setBody(new Block(this.getLocation(), (Scope) cd));
                cd.compile(cf);
            }

            // Methods, including class initialization method.
            for (int i = 0; i < declaredMethods.size(); ++i) {
                MethodDeclarator md = (MethodDeclarator) this.declaredMethods.elementAt(i);
                md.compile(cf);
            }

            // Member classes.
            for (int i = 0; i < this.declaredClasses.size(); ++i) {
                ((ClassDeclaration) this.declaredClasses.elementAt(i)).compile(
                    classFiles_return
                );
            }

            // Member interfaces.
            for (int i = 0; i < this.declaredInterfaces.size(); ++i) {
                ((InterfaceDeclaration) this.declaredInterfaces.elementAt(i)).compile(
                    classFiles_return
                );
            }

            classFiles_return.add(cf);
        }

        private IClass   superclass = null;
        private IClass[] interfaces = null;
    }

    /*package*/ static class InterfaceDeclaration extends ClassOrInterfaceDeclaration {
        public InterfaceDeclaration(
            Scanner.Location location,
            Scope            enclosingScope,
            short            modifiers,
            String           name,
            ReferenceType[]  extendedTypes
        ) throws Parser.ParseException {
            super(
                location,
                enclosingScope,
                modifiers,
                name
            );
            this.extendedTypes = extendedTypes;
        }

        public void addConstantDeclaration(FieldDeclarator fd) {
            this.constantDeclarations.add(fd);
        }

        // Implement "IClass".
        public IClass getSuperclass() {
            return IClass.OBJECT;
        }
        public IClass[] getImplementedInterfaces() throws CompileException {
            IClass[] res = new IClass[this.extendedTypes.length];
            for (int i = 0; i < res.length; ++i) {
                res[i] = this.extendedTypes[i].getType();
            }
            return res;
        }
        public IClass.IConstructor[] getDeclaredIConstructors() {
            return new IClass.IConstructor[0];
        }
        public IClass.IField[] getDeclaredIFields() {
            Vector res = new Vector();

            // Determine static fields.
            for (int i = 0; i < this.constantDeclarations.size(); ++i) {
                BlockStatement bs = (BlockStatement) this.constantDeclarations.elementAt(i);
                if (bs instanceof FieldDeclarator) {
                    FieldDeclarator fd = (FieldDeclarator) bs;
                    res.add(fd.getIFields());
                }
            }

            return (IClass.IField[]) res.toArray(new IClass.IField[res.size()]);
        }

        public void compile(Vector classFiles_return) throws CompileException {

            // Determine extended interfaces.
            this.interfaces = new IClass[this.extendedTypes.length];
            String[] interfaceDescriptors = new String[this.interfaces.length];
            for (int i = 0; i < extendedTypes.length; ++i) {
                this.interfaces[i] = this.extendedTypes[i].getType();
                interfaceDescriptors[i] = this.interfaces[i].getDescriptor();
            }

            // Create "ClassFile" object.
            ClassFile cf = new ClassFile(
                (short) (            // accessFlags
		    this.modifiers |
		    Mod.SUPER |
		    Mod.INTERFACE
                ),
                this.getClassName(), // thisClassName
                Descriptor.OBJECT,   // superClassFD
                interfaceDescriptors // interfaceNames
            );

            // Class variables.
            for (int i = 0; i < this.constantDeclarations.size(); ++i) {
                BlockStatement bs = (BlockStatement) this.constantDeclarations.elementAt(i);
                if (!(bs instanceof FieldDeclarator)) continue;
                FieldDeclarator fd = (FieldDeclarator) bs;
                for (int j = 0; j < fd.variableDeclarators.length; ++j) {
                    VariableDeclarator vd = fd.variableDeclarators[j];
                    Type type = fd.type;
                    for (int k = 0; k < vd.brackets; ++k) type = new ArrayType(type);
                    cf.addFieldInfo(
                        fd.modifiers,                   // accessFlags
                        type.getType().getDescriptor(), // fieldTypeFD
                        vd.name,                        // fieldName
                        (                               // optionalConstantValue
                            (fd.modifiers & Mod.FINAL) != 0 &&
                            vd.optionalInitializer != null
                        ) ? vd.optionalInitializer.getConstantValue() : null
                    );
                }
            }

            // Interface initialization method.
            if (!this.constantDeclarations.isEmpty()) {
                MethodDeclarator interfaceInitializationMethod = new MethodDeclarator(
                    this.getLocation(),     // location
                    this,                   // declaringClassOrInterface
                    (short) (               // modifiers
                        Mod.STATIC |
                        Mod.PUBLIC
                    ),
                    new BasicType(          // type
                        this.getLocation(),
                        BasicType.VOID
                    ),
                    "<clinit>",             // name
                    new FormalParameter[0], // formalParameters
                    new ReferenceType[0]    // thrownExcaptions
                );
                Block b = new Block(
                    this.getLocation(),
                    interfaceInitializationMethod
                );
                b.addStatements(this.constantDeclarations);
                interfaceInitializationMethod.setBody(b);
                interfaceInitializationMethod.compile(cf);
            }

            // Methods.
            for (int i = 0; i < declaredMethods.size(); ++i) {
                MethodDeclarator md = (MethodDeclarator) declaredMethods.elementAt(i);
                md.compile(cf);
            }

            // Member classes.
            for (int i = 0; i < this.declaredClasses.size(); ++i) {
                ((ClassDeclaration) this.declaredClasses.elementAt(i)).compile(
                    classFiles_return
                );
            }

            // Member interfaces.
            for (int i = 0; i < this.declaredInterfaces.size(); ++i) {
                ((InterfaceDeclaration) this.declaredInterfaces.elementAt(i)).compile(
                    classFiles_return
                );
            }

            classFiles_return.add(cf);
        }

        private final ReferenceType[] extendedTypes;
        private final Vector          constantDeclarations = new Vector(); // FieldDeclarator

        // Set during "compile()".
        private IClass[] interfaces = null;
    }

    /**
     * Abstract base class for {@link Java.ConstructorDeclarator} and
     * {@link Java.MethodDeclarator}.
     */
    /*package*/ abstract static class FunctionDeclarator extends Located implements Scope {
        public FunctionDeclarator(
            Scanner.Location            location,
            ClassOrInterfaceDeclaration declaringClassOrInterface,
            short                       modifiers,
            Type                        type,
            String                      name,
            FormalParameter[]           formalParameters,
            Type[]                      thrownExceptions
        ) {
            super(location);
            this.declaringClassOrInterface = declaringClassOrInterface;
            this.modifiers                 = modifiers;
            this.type                      = type;
            this.name                      = name;
            this.formalParameters          = formalParameters;
            this.thrownExceptions          = thrownExceptions;
        }

        // Implement "Scope".
        public Scope getEnclosingScope() {
            return this.declaringClassOrInterface;
        }

        public short getModifiers() { return this.modifiers; }

        public IClass getReturnType() throws CompileException {
            if (this.returnType == null) {
                this.returnType = this.type.getType();
            }
            return this.returnType;
        }

        public String getName() { return this.name; }

        public FormalParameter[] getFormalParameters() { return this.formalParameters; }

        public void compile(ClassFile classFile) throws CompileException {

            // Create MethodInfo.
            StringBuffer sb = new StringBuffer();
            sb.append('(');
            for (int i = 0; i < this.formalParameters.length; ++i) {
                sb.append(this.formalParameters[i].type.getType().getDescriptor());
            }
            sb.append(')');
            sb.append(this.getReturnType().getDescriptor());

            ClassFile.MethodInfo mi = classFile.addMethodInfo(
                this.modifiers, // accessFlags
                this.name,      // name
                sb.toString()   // methodMD
            );

            if ((this.modifiers & Mod.ABSTRACT) != 0) return;

            // Create CodeAttribute.
            this.codeAttribute = mi.getCodeAttribute();

            // Define special parameter "this".
            if ((modifiers & Mod.STATIC) == 0) {
                this.codeAttribute.allocateLocalVariable((short) 1);
            }

            // Add function parameters.
            IClass dci = this.declaringClassOrInterface;
            for (int i = 0; i < formalParameters.length; ++i) {
                FormalParameter fp = formalParameters[i];
                if (parameters.containsKey(fp.name)) this.throwCompileException("Redefinition of formal parameter \"" + name + "\"");
                IClass fpt = fp.type.getType();
                this.parameters.put(fp.name, new LocalVariable(
                    fp.finaL,
                    fpt,
                    this.codeAttribute.allocateLocalVariable(Descriptor.size(fpt.getDescriptor()))
                ));
            }

            // Add "Exceptions" attribute -- TBD.

            // Compile the function statements.
            Vector ss = this.getBlockStatementsToCompile(); // BlockStatement
            boolean previousStatementCanCompleteNormally = true;
            for (int i = 0; i < ss.size(); ++i) {
                if (!previousStatementCanCompleteNormally) this.throwCompileException("Statement is unreachable");
                BlockStatement bs = (BlockStatement) ss.elementAt(i);
                previousStatementCanCompleteNormally = bs.compile(codeAttribute);
            }
            if (previousStatementCanCompleteNormally) {
                if (this.getReturnType() != IClass.VOID) this.throwCompileException("Method must return a value");
                this.codeAttribute.writeOpcode(Opcode.RETURN);
            }

            // Fix up.
            this.codeAttribute.fixUp();

            // Relocate.
            this.codeAttribute.relocate();

            // Do flow analysis.
try {
            this.codeAttribute.flowAnalysis();
} catch (RuntimeException ex) {
    ex.printStackTrace();
}
        }

        protected abstract Vector getBlockStatementsToCompile(); // BlockStatement

        protected /*final*/ ClassOrInterfaceDeclaration declaringClassOrInterface;
        protected /*final*/ short                       modifiers;
        private /*final*/   Type                        type;
        private /*final*/   String                      name;
        private /*final*/   FormalParameter[]           formalParameters;
        protected /*final*/ Type[]                      thrownExceptions;

        // Set by "compile()".
        private IClass                  returnType = null;
        private final HashMap           parameters = new HashMap();   // String name => LocalVariable
        private ClassFile.CodeAttribute codeAttribute = null;
    }

    /*package*/ static class ConstructorDeclarator extends FunctionDeclarator {
        public ConstructorDeclarator(
            Scanner.Location  location,
            ClassDeclaration  declaringClass,
            short             modifiers,
            FormalParameter[] formalParameters,
            Type[]            thrownExceptions
        ) {
            super(
                location,                                // location
                declaringClass,                          // declaringClassOrInterface
                modifiers,                               // modifiers
                new BasicType(location, BasicType.VOID), // type
                "<init>",                                // name
                formalParameters,                        // formalParameters
                thrownExceptions                         // thrownExceptions
            );

            this.declaringClass = declaringClass;

            this.iConstructor = declaringClass.new IConstructor() {

                // Implement IMember.
                public int getAccess() {
                    switch (ConstructorDeclarator.this.modifiers & Mod.PPP) {
                    case Mod.PRIVATE:
                        return IClass.PRIVATE;
                    case Mod.PROTECTED:
                        return IClass.PROTECTED;
                    case Mod.PACKAGE:
                        return IClass.PACKAGE;
                    case Mod.PUBLIC:
                        return IClass.PUBLIC;
                    default:
                        throw new RuntimeException("Invalid access");
                    }
                }

                // Implement IInvocable.
                public IClass[] getParameterTypes() throws CompileException {
                    FormalParameter[] fps = ConstructorDeclarator.this.getFormalParameters();
                    IClass[] res = new IClass[fps.length];
                    for (int i = 0; i < fps.length; ++i) {
                        res[i] = fps[i].type.getType();
                    }
                    return res;
                }
                public IClass[] getThrownExceptions() throws CompileException {
                    IClass[] res = new IClass[ConstructorDeclarator.this.thrownExceptions.length];
                    for (int i = 0; i < res.length; ++i) {
                        res[i] = ConstructorDeclarator.this.thrownExceptions[i].getType();
                    }
                    return res;
                }

                public String toString() {
                    StringBuffer sb = new StringBuffer();
                    sb.append(ConstructorDeclarator.this.declaringClassOrInterface.getClassName());
                    sb.append('(');
                    FormalParameter[] fps = ConstructorDeclarator.this.getFormalParameters();
                    for (int i = 0; i < fps.length; ++i) {
                        if (i != 0) sb.append(", ");
                        try {
                            sb.append(fps[i].type.getType().toString());
                        } catch (CompileException ex) {
                            sb.append("???");
                        }
                    }
                    return sb.append(')').toString();
                }
            };
        }
        public void setAlternateConstructorInvocation(
            ConstructorInvocation alternateConstructorInvocation
        ) {
            this.alternateConstructorInvocation = alternateConstructorInvocation;
        }
        public void setSuperclassConstructorInvocation(
            ConstructorInvocation superclassConstructorInvocation
        ) {
            this.superclassConstructorInvocation = superclassConstructorInvocation;
        }
        public void setBody(Block body) {
            if (this.body != null) throw new RuntimeException("Body must be set exactly once");
            this.body = body;
        }

        // Implement "FunctionDeclarator":
        protected Vector getBlockStatementsToCompile() { // BlockStatement
            Vector v = new Vector();
            if (this.alternateConstructorInvocation != null) {
                v.add((BlockStatement) this.alternateConstructorInvocation);
            } else
            if (this.superclassConstructorInvocation != null) {
                v.add((BlockStatement) this.superclassConstructorInvocation);
                v.addAll(this.declaringClass.instanceVariableDeclaratorsAndInstanceInitializers);
            } else {
                v.add((BlockStatement) new ExpressionStatement(
                    new ConstructorInvocation(
                        this.getLocation(), // location
                        declaringClass,     // declaringClass
                        this,               // declaringConstructor
                        true,               // targetIsSuperclass
                        new Rvalue[0]       // arguments
                    ),
                    (Scope) this.declaringClass
                ));
                v.addAll(this.declaringClass.instanceVariableDeclaratorsAndInstanceInitializers);
            }
            if (this.body == null) throw new RuntimeException("Constructor lacks body");
            v.add((BlockStatement) this.body);

            return v;
        }

        public IClass.IConstructor toIConstructor() { return this.iConstructor; }

        private final ClassDeclaration    declaringClass;
        private final IClass.IConstructor iConstructor;

        private ConstructorInvocation     alternateConstructorInvocation = null;
        private ConstructorInvocation     superclassConstructorInvocation = null;
        private Block                     body = null;
    }

    /*package*/ static class MethodDeclarator extends FunctionDeclarator {
        public MethodDeclarator(
            Scanner.Location            location,
            ClassOrInterfaceDeclaration declaringClassOrInterface,
            short                       modifiers,
            Type                        type,
            String                      name,
            FormalParameter[]           formalParameters,
            Type[]                      thrownExceptions
        ) {
            super(
                location,                  // location
                declaringClassOrInterface, // declaringClassOrInterface
                modifiers,                 // modifiers
                type,                      // type
                name,                      // name
                formalParameters,          // formalParameters
                thrownExceptions           // thrownExceptions
            );
            this.iMethod = declaringClassOrInterface.new IMethod() {

                // Implement IMember.
                public int getAccess() {
                    switch (MethodDeclarator.this.modifiers & Mod.PPP) {
                    case Mod.PRIVATE:
                        return IClass.PRIVATE;
                    case Mod.PROTECTED:
                        return IClass.PROTECTED;
                    case Mod.PACKAGE:
                        return IClass.PACKAGE;
                    case Mod.PUBLIC:
                        return IClass.PUBLIC;
                    default:
                        throw new RuntimeException("Invalid access");
                    }
                }

                // Implement IInvocable.
                public IClass[] getParameterTypes() throws CompileException {
                    FormalParameter[] fps = MethodDeclarator.this.getFormalParameters();
                    IClass[] res = new IClass[fps.length];
                    for (int i = 0; i < fps.length; ++i) {
                        res[i] = fps[i].type.getType();
                    }
                    return res;
                }
                public IClass[] getThrownExceptions() throws CompileException {
                    IClass[] res = new IClass[MethodDeclarator.this.thrownExceptions.length];
                    for (int i = 0; i < res.length; ++i) {
                        res[i] = MethodDeclarator.this.thrownExceptions[i].getType();
                    }
                    return res;
                }

                // Implement IMethod.
                public boolean isStatic() { return (MethodDeclarator.this.getModifiers() & Mod.STATIC) != 0; }
                public IClass getReturnType() throws CompileException {
                    return MethodDeclarator.this.getReturnType();
                }
                public String getName() { return MethodDeclarator.this.getName(); }
            };
        }

        public void setBody(Block body) {
            if (this.body != null) throw new RuntimeException("Body must be set exactly once");
            this.body = body;
        }

        // Implement "FunctionDeclarator":
        protected Vector getBlockStatementsToCompile() { // BlockStatement
            Vector v = new Vector();
            if (body == null) throw new RuntimeException("Body not set");
            v.add((BlockStatement) body);
            return v;
        }

        public IClass.IMethod toIMethod() { return this.iMethod; }

        private final IClass.IMethod iMethod;
        private Block body = null;
    }

    /**
     * This class is derived from "Statement", because it provides for the
     * initialization of the field. In other words, "compile()" generates the
     * code that initializes the field.
     */
    /*package*/ static class FieldDeclarator extends Statement {
        public FieldDeclarator(
            Scanner.Location            location,
            ClassOrInterfaceDeclaration declaringClassOrInterface,
            short                       modifiers,
            Type                        type,
            VariableDeclarator[]        variableDeclarators
        ) {
            super(
                location,
                null      // enclosingBlock
            );
            this.modifiers                 = modifiers;
            this.declaringClassOrInterface = declaringClassOrInterface;
            this.type                      = type;
            this.variableDeclarators       = variableDeclarators;
        }

        public IClass.IField[] getIFields() {
            IClass.IField[] res = new IClass.IField[this.variableDeclarators.length];
            for (int i = 0; i < res.length; ++i) {
                final VariableDeclarator vd = this.variableDeclarators[i];
                res[i] = this.declaringClassOrInterface.new IField() {

                    // Implement IMember.
                    public int getAccess() {
                        switch (FieldDeclarator.this.modifiers & Mod.PPP) {
                        case Mod.PRIVATE:
                            return IClass.PRIVATE;
                        case Mod.PROTECTED:
                            return IClass.PROTECTED;
                        case Mod.PACKAGE:
                            return IClass.PACKAGE;
                        case Mod.PUBLIC:
                            return IClass.PUBLIC;
                        default:
                            throw new RuntimeException("Invalid access");
                        }
                    }

                    // Implement "IField".
                    public boolean isStatic() { return (FieldDeclarator.this.modifiers & Mod.STATIC) != 0; }
                    public IClass getType() throws CompileException {
                        IClass res2 = FieldDeclarator.this.type.getType();
                        for (int j = 0; j < vd.brackets; ++j) res2 = res2.getArrayType();
                        return res2;
                    }
                    public String getName() { return vd.name; }
                };
            }
            return res;
        }

        private final ClassOrInterfaceDeclaration declaringClassOrInterface;
        private final short                       modifiers;
        private final Type                        type;
        private final VariableDeclarator[]        variableDeclarators;

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            for (int i = 0; i < variableDeclarators.length; ++i) {
                VariableDeclarator vd = variableDeclarators[i];
                Rvalue initializer = vd.optionalInitializer;

                // No code generated if no initializer.
                if (initializer == null) continue;

                // If field is FINAL STATIC and initializer is a constant
                // value, then continue, because that field is already
                // initialized through a "ConstantValue" attribute (4.7.2).
                if (
                    (this.modifiers & Mod.STATIC) != 0 &&
                    (this.modifiers & Mod.FINAL) != 0 &&
                    initializer.getConstantValue() != null
                ) continue;

                if ((modifiers & Mod.STATIC) == 0) {
                    codeAttribute.writeOpcode(Opcode.ALOAD_0);
                }
                IClass initializerType = initializer.compileGetValue(codeAttribute);
                IClass fieldType = this.type.getType();
                for (int j = 0; j < vd.brackets; ++j) fieldType = fieldType.getArrayType();
                Java.assignmentConversion(
                    codeAttribute,    // codeAttribute
                    (Locatable) this, // locatable
                    initializerType,  // sourceType
                    fieldType         // destinationType
                );
                if ((modifiers & Mod.STATIC) != 0) {
                    codeAttribute.writeOpcode(Opcode.PUTSTATIC);
                } else {
                    codeAttribute.writeOpcode(Opcode.PUTFIELD);
                }
                codeAttribute.writeShort(codeAttribute.getClassFile().addConstantFieldrefInfo(
                    this.declaringClassOrInterface.getDescriptor(), // classFD
                    vd.name,                                        // fieldName
                    fieldType.getDescriptor()                       // fieldFD
                ));
            }
            return true;
        }
    }

    /*package*/ static class VariableDeclarator {
        public VariableDeclarator(
            String name,
            int    brackets,
            Rvalue optionalInitializer
        ) {
            this.name                = name;
            this.brackets            = brackets;
            this.optionalInitializer = optionalInitializer;
        }
        private final String name;
        private final int    brackets;
        private final Rvalue optionalInitializer;
    }

    /*package*/ static class FormalParameter extends Located {
        public FormalParameter(
            Scanner.Location location,
            boolean          finaL,
            Type             type,
            String           name
        ) {
            super(location);
            this.finaL = finaL;
            this.type  = type;
            this.name  = name;
        }
        private final boolean finaL;
        private final Type    type;
        private final String  name;
    }

    /**
     * Base of all statements that can appear in a block.
     */
    /*package*/ interface BlockStatement extends Locatable {

        /**
         * @return <tt>false</tt> if this statement cannot complete normally (14.20)
         */
        public abstract boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException;
    }

    /*package*/ static abstract class Statement implements BlockStatement, Scope {
        protected Statement(
            Scanner.Location location,
            Scope            enclosingScope
        ) {
            this.location       = location;
            this.enclosingScope = enclosingScope;
        }

        // Implement "Scope".
        public Scope getEnclosingScope() { return this.enclosingScope; }

        //Implement "Locatable".
        public Scanner.Location getLocation() { return location; }
        public void throwParseException(String message) throws Parser.ParseException {
            throw new Parser.ParseException(message, location);
        }
        public void throwCompileException(String message) throws CompileException {
            throw new CompileException(message, location);
        }

        private final Scanner.Location location;
        protected final Scope          enclosingScope;
    }

    /*package*/ static class LabeledStatement extends BreakableStatement {
        protected LabeledStatement(
            Scanner.Location location,
            Scope            enclosingScope,
            String           label
        ) {
            super(location, enclosingScope);
            this.label = label;
        }

        public void setBody(Statement body) {
            this.body = body;
        }

        private /*final*/ String label;
        private Statement    body = null;

        // Compile time members:

        public final boolean compile2(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            return body.compile(codeAttribute);
        }
    }

    /**
     * Representation of a Java<sup>TM</sup> "block" (JLS 14.2).
     * <p>
     * The statements that the block defines are executed in sequence.
     */
    /*package*/ static class Block extends Statement {
        public Block(
            Scanner.Location location,
            Scope            enclosingScope
        ) {
            super(location, enclosingScope);
        }

        public void addStatement(BlockStatement statement) {
            this.statements.add(statement);
        }
        public void addStatements(
            Vector statements // BlockStatement
        ) {
            this.statements.addAll(statements);
        }
        public BlockStatement[] getStatements() {
            return (BlockStatement[]) statements.toArray(new BlockStatement[statements.size()]);
        }

        private final Vector statements = new Vector(); // BlockStatement

        // Compile time members.

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            codeAttribute.saveLocalVariables();
            try {
                boolean previousStatementCanCompleteNormally = true;
                for (int i = 0; i < statements.size(); ++i) {
                    BlockStatement bs = (BlockStatement) statements.elementAt(i);
                    if (!previousStatementCanCompleteNormally) bs.throwCompileException("Statement is unreachable");
                    if (bs instanceof LocalVariableDeclarationStatement) {
                        this.defineLocalVariables(
                            codeAttribute,
                            (LocalVariableDeclarationStatement) bs
                        );
                    }
                    previousStatementCanCompleteNormally = bs.compile(codeAttribute);
                }
                return previousStatementCanCompleteNormally;
            } finally {
                codeAttribute.restoreLocalVariables();
            }
        }

        public short defineLocalVariable(
            Locatable               locatable,
            ClassFile.CodeAttribute codeAttribute,
            boolean                 finaL,
            IClass                  type,
            String                  name
        ) throws CompileException {
            Scope s;
            for (s = this; s instanceof Statement; s = s.getEnclosingScope()) {
                if (s instanceof Block) {
                    Block b = (Block) s;
                    if (b.localVariables.containsKey(name)) locatable.throwCompileException("Redefinition of local variable \"" + name + "\"");
                }
            }
            short lvIndex = codeAttribute.allocateLocalVariable(
                Descriptor.size(type.getDescriptor())
            );
            this.localVariables.put(name, new LocalVariable(
                finaL,
                type,
                lvIndex
            ));
            return lvIndex;
        }

        public void defineLocalVariables(
            ClassFile.CodeAttribute           codeAttribute,
            LocalVariableDeclarationStatement lvds
        ) throws CompileException {
            for (int j = 0; j < lvds.variableDeclarators.length; ++j) {
                VariableDeclarator vd = lvds.variableDeclarators[j];
                if ((lvds.modifiers & ~Mod.FINAL) != 0) this.throwCompileException("The only allowed modifier in local variable declarations is \"final\"");

                Type type = lvds.type;
                for (int k = 0; k < vd.brackets; ++k) type = new ArrayType(type);
                this.defineLocalVariable(
                    (Locatable) lvds,
                    codeAttribute,
                    (lvds.modifiers & Mod.FINAL) != 0,
                    type.getType(),
                    vd.name
                );
            }
        }

        private HashMap localVariables = new HashMap(); // String name => LocalVariable
    }

    /**
     * Base class for statements that can be terminated abnormally with a
     * "break" statement.
     * <p>
     * According to the JLS, statements that can be terminated abnormally with
     * a "break" statement are: "COntinuable" statements ("for", "do" and
     * "while"), labeled statements, and the "switch" statement.
     */
    /*package*/ static abstract class BreakableStatement extends Statement {
        public BreakableStatement(
            Scanner.Location location,
            Scope            enclosingScope
        ) {
            super(location, enclosingScope);
        }

        // Implement "BlockStatement".
        public final boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            boolean canCompleteNormally = this.compile2(codeAttribute);
            if (this.whereToBreak != null) {
                this.whereToBreak.set();
                canCompleteNormally = true;
            }
            return canCompleteNormally;
        }
        public abstract boolean compile2(ClassFile.CodeAttribute codeAttribute) throws CompileException;
        public ClassFile.CodeAttribute.Offset getWhereToBreak(ClassFile.CodeAttribute codeAttribute) {
            if (this.whereToBreak == null) {
                this.whereToBreak = codeAttribute.new Offset();
            }
            return this.whereToBreak;
        }

        private ClassFile.CodeAttribute.Offset whereToBreak = null;
    }

    /*package*/ static abstract class ContinuableStatement extends BreakableStatement {
        public ContinuableStatement(
            Scanner.Location location,
            Scope            enclosingScope
        ) {
            super(location, enclosingScope);
        }

        protected ClassFile.CodeAttribute.Offset whereToContinue = null;
        protected boolean                        bodyHasContinue = false;
    }

    /*package*/ static class ExpressionStatement extends Statement {
        public ExpressionStatement(
            Rvalue rvalue,
            Scope  enclosingScope
        ) {
            super(rvalue.getLocation(), enclosingScope);
            this.rvalue = rvalue;
        }
        private final Rvalue rvalue;

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            this.rvalue.compile(codeAttribute);
            return true;
        }
    }

    /*package*/ static class IfStatement extends Statement {
        public IfStatement(
            Scanner.Location location,
            Scope            enclosingScope,
            Rvalue           condition,
            BlockStatement   trueStatement,
            BlockStatement   falseStatement // optional
        ) {
            super(location, enclosingScope);
            this.condition      = condition;
            this.trueStatement  = trueStatement;
            this.falseStatement = falseStatement;
        }

        private final Rvalue         condition;
        private final BlockStatement trueStatement;
        private final BlockStatement falseStatement; // optional

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws Java.CompileException {

            // Check for constant condition.
            Object cv = this.condition.getConstantValue();
            if (cv != null && cv instanceof Boolean) {
                return (
                    ((Boolean) cv).booleanValue() ?
                    this.trueStatement.compile(codeAttribute) :
                    this.falseStatement == null || this.falseStatement.compile(codeAttribute)
                );
            }

            if (this.falseStatement == null) {
                ClassFile.CodeAttribute.Offset end = codeAttribute.new Offset();
                this.condition.compileBoolean(codeAttribute, end, Java.Rvalue.JUMP_IF_FALSE);
                this.trueStatement.compile(codeAttribute);
                end.set();
                return true;
            }
            ClassFile.CodeAttribute.Offset fso = codeAttribute.new Offset();
            ClassFile.CodeAttribute.Offset end = codeAttribute.new Offset();
            this.condition.compileBoolean(codeAttribute, fso, Java.Rvalue.JUMP_IF_FALSE);
            boolean tsccn = this.trueStatement.compile(codeAttribute);
            codeAttribute.writeBranch(Opcode.GOTO, end);
            fso.set();
            boolean fsccn = this.falseStatement.compile(codeAttribute);
            end.set();
            return tsccn || fsccn;
        }
    }

    /*package*/ static class ForStatement extends ContinuableStatement {
        public ForStatement(
            Scanner.Location location,
            Scope            enclosingScope
        ) {

            // Trick: Insert a "Block" between the enclosing scope and the
            // "for" statement.
            super(location, new Block(location, enclosingScope));
            this.implicitBlock = (Block) this.enclosingScope;
            this.body = new Block(
                location,
                this      // enclosingScope
            );
        }

        public void set(
            BlockStatement init,      // optional
            Rvalue         condition, // optional
            BlockStatement update,    // optional
            BlockStatement body
        ) {
            this.init      = init;
            this.condition = condition;
            this.update    = update;
            this.body      = body;
        }

        public final Block     implicitBlock;
        private BlockStatement init      = null;
        private Rvalue         condition = null;
        private BlockStatement update    = null;
        private BlockStatement body;

        // Compile time members:

        public boolean compile2(ClassFile.CodeAttribute codeAttribute) throws CompileException {

            // Check for constant condition.
            boolean constantTrueCondition = false;
            {
                Object cv = (
                    condition == null ?
                    new Boolean(true) :
                    condition.getConstantValue()
                );
                if (cv != null && cv instanceof Boolean) {
                    if (!((Boolean) cv).booleanValue()) this.throwCompileException("Body of \"for\" statement is unreachable");
                    constantTrueCondition = true;
                }
            }

            this.whereToContinue = codeAttribute.new Offset();
            this.bodyHasContinue = false;

            codeAttribute.saveLocalVariables();
            try {
                if (init != null && init instanceof LocalVariableDeclarationStatement) {
                    this.implicitBlock.defineLocalVariables(
                        codeAttribute,
                        (LocalVariableDeclarationStatement) init
                    );
                }

                if (init != null) init.compile(codeAttribute);

                ClassFile.CodeAttribute.Offset toCondition = codeAttribute.new Offset();
                if (constantTrueCondition) {
                    ; // GOTO condition not necessary in this case.
                } else {
                    codeAttribute.writeBranch(Opcode.GOTO, toCondition);
                }

                ClassFile.CodeAttribute.Offset bodyOffset = codeAttribute.newOffset();
                if (
                    !body.compile(codeAttribute) &&
                    !this.bodyHasContinue &&
                    update != null
                ) this.throwCompileException("For update is unreachable");

                // Compile update.
                this.whereToContinue.set();
                if (update != null) update.compile(codeAttribute);

                // Compile condition.
                toCondition.set();
                condition.compileBoolean(codeAttribute, bodyOffset, Rvalue.JUMP_IF_TRUE);
            } finally {
                codeAttribute.restoreLocalVariables();
            }

            return !constantTrueCondition;
        }
    }

    /*package*/ static class WhileStatement extends ContinuableStatement {
        public WhileStatement(
            Scanner.Location location,
            Scope            enclosingScope,
            Rvalue           condition
        ) {
            super(location, enclosingScope);
            this.condition = condition;
        }

        public void setBody(BlockStatement body) {
            this.body = body;
        }

        private final Rvalue   condition;
        private BlockStatement body = null;

        // Compile time members:

        public boolean compile2(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            this.whereToContinue = codeAttribute.new Offset();
            this.bodyHasContinue = false;

            boolean constantTrueCondition = false;
            {
                Object cv = condition.getConstantValue();
                if (cv != null && cv instanceof Boolean) {
                    if (!((Boolean) cv).booleanValue()) this.throwCompileException("Body of \"while\" statement is unreachable");
                    constantTrueCondition = true;
                }
            }

            if (!constantTrueCondition) {
                codeAttribute.writeBranch(Opcode.GOTO, this.whereToContinue);
            }

            // Compile body.
            ClassFile.CodeAttribute.Offset bodyOffset = codeAttribute.newOffset();
            body.compile(codeAttribute);

            // Compile condition.
            this.whereToContinue.set();
            condition.compileBoolean(codeAttribute, bodyOffset, Rvalue.JUMP_IF_TRUE);

            return !constantTrueCondition;
        }
    }

    /*package*/ static class TryStatement extends Statement {
        public TryStatement(
            Scanner.Location location,
            Scope            enclosingScope
        ) {
            super(location, enclosingScope);
        }

        public void setBody(BlockStatement body) {
            this.body = body;
        }
        public void addCatchClause(CatchClause catchClause) {
            this.catchClauses.add(catchClause);
        }
        public Vector getCatchClauses() {
            return this.catchClauses;
        }
        public void setFinally(Block finallY) {
            this.finallY = finallY;
        }
        public Block getFinally() {
            return this.finallY;
        }

        private BlockStatement body         = null;
        private final Vector   catchClauses = new Vector();
        private Block          finallY      = null; // Optional

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            if (this.finallY != null) this.finallyOffset = codeAttribute.new Offset();

            ClassFile.CodeAttribute.Offset beginningOfBody = codeAttribute.newOffset();
            ClassFile.CodeAttribute.Offset afterStatement = codeAttribute.new Offset();

            boolean canCompleteNormally = this.body.compile(codeAttribute);
            ClassFile.CodeAttribute.Offset afterBody = codeAttribute.newOffset();
            if (canCompleteNormally) {
                codeAttribute.writeBranch(Opcode.GOTO, afterStatement);
            }

            codeAttribute.saveLocalVariables();
            try {

                // Local variable for exception object.
                short exceptionLvIndex = codeAttribute.allocateLocalVariable((short) 1);

                for (int i = 0; i < this.catchClauses.size(); ++i) {
                    CatchClause cc = (CatchClause) this.catchClauses.elementAt(i);
                    IClass caughtExceptionType = cc.caughtException.type.getType();
                    codeAttribute.addException(
                        beginningOfBody,                    // startPC
                        afterBody,                          // endPC
                        codeAttribute.newOffset(),          // handlerPC
                        caughtExceptionType.getDescriptor() // catchTypeFD
                    );
                    Java.store(
                        codeAttribute,       // codeAttribute
                        caughtExceptionType, // lvType
                        exceptionLvIndex     // lvIndex
                    );

                    // Kludge: Treat the exception variable like a local
                    // variable of the catch clause body.
                    cc.body.localVariables.put(
                        cc.caughtException.name,
                        new LocalVariable(
                            false,               // finaL
                            caughtExceptionType, // type
                            exceptionLvIndex     // localVariableIndex
                        )
                    );

                    if (cc.body.compile(codeAttribute)) {
                        canCompleteNormally = true;
                        if (
                            i < this.catchClauses.size() - 1 ||
                            this.finallY != null
                        ) codeAttribute.writeBranch(Opcode.GOTO, afterStatement);
                    }
                }
                if (this.finallY != null) {
                    ClassFile.CodeAttribute.Offset here = codeAttribute.newOffset();
                    codeAttribute.addException(
                        beginningOfBody, // startPC
                        here,            // endPC
                        here,            // handlerPC
                        null             // catchTypeFD
                    );

                    // Store exception object in local variable.
                    Java.store(
                        codeAttribute,   // codeAttribute
                        IClass.OBJECT,   // valueType
                        exceptionLvIndex // localVariableIndex
                    );
                    codeAttribute.writeBranch(Opcode.JSR, this.finallyOffset);
                    Java.load(
                        codeAttribute,   // codeAttribute
                        IClass.OBJECT,   // valueType
                        exceptionLvIndex // localVariableIndex
                    );
                    codeAttribute.writeOpcode(Opcode.ATHROW);

                    this.finallyOffset.set();
                    short pcLVIndex = codeAttribute.allocateLocalVariable((short) 1);
                    Java.store(
                        codeAttribute,    // codeAttribute
                        IClass.OBJECT,    // valueType
                        pcLVIndex         // localVariableIndex
                    );
                    if (this.finallY.compile(codeAttribute)) {
                        codeAttribute.writeOpcode(Opcode.RET);
                        codeAttribute.writeByte(pcLVIndex);
                    } else {
                        canCompleteNormally = false;
                    }
                }

                afterStatement.set();

                if (this.finallyOffset != null) {
                    codeAttribute.writeBranch(Opcode.JSR, this.finallyOffset);
                }
            } finally {
                codeAttribute.restoreLocalVariables();
            }

            return canCompleteNormally;
        }

        private ClassFile.CodeAttribute.Offset finallyOffset = null;
    }

    /*package*/ static class CatchClause {
        CatchClause(
            FormalParameter caughtException,
            Block           body
        ) {
            this.caughtException = caughtException;
            this.body            = body;
        }

        private final FormalParameter caughtException;
        private final Block           body;
    }

    /*package*/ static class SwitchStatement extends BreakableStatement {
        public SwitchStatement(
            Scanner.Location location,
            Scope            enclosingScope
        ) {
            super(location, enclosingScope);
        }

        public void setCondition(Rvalue condition) {
            this.condition = condition;
        }
        public void addSwitchBlockStatementGroup(SwitchBlockStatementGroup sbsg) {
            this.sbsgs.add(sbsg);
        }

        private Rvalue condition = null;
        private final Vector sbsgs = new Vector(); // SwitchBlockStatementGroup

        // Compile time members:

        public boolean compile2(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            TreeMap                          caseLabelMap = new TreeMap(); // Integer => Offset
            ClassFile.CodeAttribute.Offset   defaultLabelOffset = null;
            ClassFile.CodeAttribute.Offset[] sbsgOffsets = new ClassFile.CodeAttribute.Offset[this.sbsgs.size()];
            for (int i = 0; i < this.sbsgs.size(); ++i) {
                SwitchBlockStatementGroup      sbsg = (SwitchBlockStatementGroup) this.sbsgs.elementAt(i);
                sbsgOffsets[i] = codeAttribute.new Offset();
                for (int j = 0; j < sbsg.caseLabels.size(); ++j) {
                    Rvalue rv = (Rvalue) sbsg.caseLabels.elementAt(j);
                    Object cv = rv.getConstantValue();
                    if (cv == null) rv.throwCompileException("Value of \"case\" switch label does not pose a constant value");
                    if (!(cv instanceof Integer)) rv.throwCompileException("Value of \"case\" switch label must be of type \"int\"");
                    if (caseLabelMap.containsKey(cv)) rv.throwCompileException("Duplicate \"case\" switch label value");
                    caseLabelMap.put(cv, sbsgOffsets[i]);
                }
                if (sbsg.hasDefaultLabel) {
                    if (defaultLabelOffset != null) sbsg.throwCompileException("Duplicate \"default\" switch label");
                    defaultLabelOffset = sbsgOffsets[i];
                }
            }
            if (defaultLabelOffset == null) defaultLabelOffset = this.getWhereToBreak(codeAttribute);

            // Compute condition.
            Java.assignmentConversion(
                codeAttribute,                  // codeAttribute
                (Locatable) this,               // locatable
                this.condition.compileGetValue( // sourceType
                    codeAttribute
                ),
                IClass.INT                      // targetType
            );

            // Generate TABLESWITCH or LOOKUPSWITCH instruction.
            ClassFile.CodeAttribute.Offset switchOffset = codeAttribute.newOffset();
            if (caseLabelMap.isEmpty() ||
            2 * caseLabelMap.size() >= (
                ((Integer) caseLabelMap.lastKey()).intValue() -
                ((Integer) caseLabelMap.firstKey()).intValue()
            )) {
                int low = ((Integer) caseLabelMap.firstKey()).intValue();
                int high = ((Integer) caseLabelMap.lastKey()).intValue();

                codeAttribute.writeOpcode(Opcode.TABLESWITCH);
                new Padder(codeAttribute).set();
                codeAttribute.writeOffset(switchOffset, defaultLabelOffset);
                codeAttribute.writeInt(low);
                codeAttribute.writeInt(high);
                Iterator si = caseLabelMap.entrySet().iterator();
                int cur = low;
                while (si.hasNext()) {
                    Map.Entry me = (Map.Entry) si.next();
                    int val = ((Integer) me.getKey()).intValue();
                    while (cur < val) {
                        codeAttribute.writeOffset(switchOffset, defaultLabelOffset);
                        ++cur;
                    }
                    codeAttribute.writeOffset(
                        switchOffset,
                        (ClassFile.CodeAttribute.Offset) me.getValue()
                    );
                    ++cur;
                }
            } else {
                codeAttribute.writeOpcode(Opcode.LOOKUPSWITCH);
                new Padder(codeAttribute).set();
                codeAttribute.writeOffset(switchOffset, defaultLabelOffset);
                codeAttribute.writeInt(caseLabelMap.size());
                Iterator si = caseLabelMap.entrySet().iterator();
                while (si.hasNext()) {
                    Map.Entry me = (Map.Entry) si.next();
                    codeAttribute.writeInt(((Integer) me.getKey()).intValue());
                    codeAttribute.writeOffset(
                        switchOffset,
                        (ClassFile.CodeAttribute.Offset) me.getValue()
                    );
                }
            }

            // Compile statement groups.
            boolean canCompleteNormally = true;
            for (int i = 0; i < this.sbsgs.size(); ++i) {
                SwitchBlockStatementGroup sbsg = (SwitchBlockStatementGroup) this.sbsgs.elementAt(i);
                sbsgOffsets[i].set();
                canCompleteNormally = true;
                for (int j = 0; j < sbsg.blockStatements.size(); ++j) {
                    BlockStatement bs = (BlockStatement) sbsg.blockStatements.elementAt(j);
                    if (!canCompleteNormally) bs.throwCompileException("Statement is unreachable");
                    canCompleteNormally = bs.compile(codeAttribute);
                }
            }
            return canCompleteNormally;
        }
    }
    private static class Padder extends ClassFile.CodeAttribute.Inserter implements ClassFile.FixUp {
        public Padder(ClassFile.CodeAttribute codeAttribute) {
            codeAttribute.super();
        }
        public void fixUp() {
            int x = this.offset % 4;
            if (x != 0) {
                ClassFile.CodeAttribute ca = this.getCodeAttribute();
                ca.pushInserter(this); {
                    ca.write(new byte[4 - x]);
                } ca.popInserter();
            }
        }
    }

    /*package*/ static class SwitchBlockStatementGroup extends Located {
        public SwitchBlockStatementGroup(Scanner.Location location) {
            super(location);
        }

        public void addSwitchLabel(Rvalue value) {
            this.caseLabels.add(value);
        }
        public void addDefaultSwitchLabel() throws Parser.ParseException {
            if (this.hasDefaultLabel) this.throwParseException("Duplicate \"default\" switch label");
            this.hasDefaultLabel = true;
        }
        public void setBlockStatements(Vector blockStatements) {
            this.blockStatements = blockStatements;
        }

        private final Vector caseLabels = new Vector(); // Rvalue
        private boolean      hasDefaultLabel = false;
        private Vector       blockStatements; // BlockStatement
    }

    /*package*/ static class DoStatement extends ContinuableStatement {
        public DoStatement(
            Scanner.Location location,
            Scope            enclosingScope
        ) {
            super(location, enclosingScope);
        }

        public void setBody(BlockStatement body) {
            this.body = body;
        }
        public void setCondition(Rvalue condition) {
            this.condition = condition;
        }

        private BlockStatement body      = null;
        private Rvalue         condition = null;

        // Compile time members:

        public boolean compile2(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            this.whereToContinue = codeAttribute.new Offset();
            this.bodyHasContinue = false;

            boolean constantTrueCondition = false;
            {
                Object cv = condition.getConstantValue();
                if (cv != null && cv instanceof Boolean) {
                    constantTrueCondition = ((Boolean) cv).booleanValue();
                }
            }

            ClassFile.CodeAttribute.Offset bodyOffset = codeAttribute.newOffset();

            // Compile body.
            if (!body.compile(codeAttribute) && !this.bodyHasContinue) this.throwCompileException("\"do\" statement never tests its condition");

            // Compile condition.
            this.whereToContinue.set();
            condition.compileBoolean(codeAttribute, bodyOffset, Rvalue.JUMP_IF_TRUE);

            return !constantTrueCondition;
        }
    }

    /*package*/ static class LocalVariableDeclarationStatement extends Statement {
        public LocalVariableDeclarationStatement(
            Scanner.Location     location,
            Block                declaringBlock,
            short                modifiers,
            Type                 type,
            VariableDeclarator[] variableDeclarators
        ) {
            super(location, declaringBlock);
            this.declaringBlock      = declaringBlock;
            this.modifiers           = modifiers;
            this.type                = type;
            this.variableDeclarators = variableDeclarators;
        }

        private final Block                declaringBlock;
        private final short                modifiers;
        private final Type                 type;
        private final VariableDeclarator[] variableDeclarators;

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            for (int i = 0; i < variableDeclarators.length; ++i) {
                VariableDeclarator vd = variableDeclarators[i];
                if (vd.optionalInitializer == null) continue; // No initializer.

                LocalVariable lv = (LocalVariable) this.declaringBlock.localVariables.get(vd.name);
                if (lv == null) throw new RuntimeException("Local variable unknown when initialized!?");

                IClass initializerType = vd.optionalInitializer.compileGetValue(codeAttribute);
                store(codeAttribute, (Locatable) this, initializerType, lv);
            }
            return true;
        }
    }

    /*package*/ static class ReturnStatement extends Statement {
        public ReturnStatement(
            Scanner.Location location,
            Scope            enclosingScope,
            Rvalue           returnValue
        ) {
            super(location, enclosingScope);
            this.returnValue = returnValue;
        }

        private final Rvalue returnValue;

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {

            // Determine enclosing block, function and compilation Unit.
            FunctionDeclarator enclosingFunction = null;
            {
                Scope s = this.enclosingScope;
                for (s = s.getEnclosingScope(); s instanceof Statement; s = s.getEnclosingScope());
                enclosingFunction = (FunctionDeclarator) s;
            }

            IClass returnType = enclosingFunction.getReturnType();
            if (returnType == IClass.VOID) {
                if (this.returnValue != null) this.throwCompileException("Method must not return a value");
                executeFinallyClauses(
                    codeAttribute,
                    this.enclosingScope,
                    enclosingFunction
                );
                codeAttribute.writeOpcode(Opcode.RETURN);
                return false;
            }
            if (this.returnValue == null) this.throwCompileException("Method must return a value");
            IClass type = returnValue.compileGetValue(codeAttribute);
            Java.assignmentConversion(
                codeAttribute,    // codeAttribute
                (Locatable) this, // locatable
                type,             // sourceType
                returnType        // targetType
            );
            executeFinallyClauses(
                codeAttribute,
                this.enclosingScope,
                enclosingFunction
            );
            codeAttribute.writeOpcode(Opcode.IRETURN + Java.ilfda(returnType));
            return false;
        }
    }

    /*package*/ static class ThrowStatement extends Statement {
        public ThrowStatement(
            Scanner.Location location,
            Scope            enclosingScope,
            Rvalue           expression
        ) {
            super(location, enclosingScope);
            this.expression = expression;
        }

        private final Rvalue expression;

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            IClass expressionType = this.expression.compileGetValue(codeAttribute);
            Java.checkThrownException(
                (Locatable) this,   // locatable
                expressionType,     // type
                this.enclosingScope // scope
            );
            codeAttribute.writeOpcode(Opcode.ATHROW);
            return false;
        }
    }

    private static void checkThrownException(
        Locatable locatable,
        IClass    type,
        Scope     scope
    ) throws CompileException {

        // Thrown object must be assignable to "Throwable".
        if (!IClass.THROWABLE.isAssignableFrom(type)) locatable.throwCompileException("Thrown object of type \"" + type + "\" is not assignable to \"Throwable\"");

        // "RuntimeException" and "Error" are never checked.
        if (
            IClass.RUNTIME_EXCEPTION.isAssignableFrom(type) ||
            IClass.ERROR.isAssignableFrom(type)
        ) return;

        // Match against enclosing "try...catch" blocks.
        while (scope instanceof Statement) {
            if (scope instanceof TryStatement) {
                TryStatement ts = (TryStatement) scope;
                for (int i = 0; i < ts.catchClauses.size(); ++i) {
                    CatchClause cc = (CatchClause) ts.catchClauses.elementAt(i);
                    IClass caughtType = cc.caughtException.type.getType();
                    if (caughtType.isAssignableFrom(type)) return;
                }
            }
            scope = scope.getEnclosingScope();
        }

        // Match against "throws" clause of declaring function.
        if (scope instanceof FunctionDeclarator) {
            FunctionDeclarator fd = (FunctionDeclarator) scope;
            for (int i = 0; i < fd.thrownExceptions.length; ++i) {
                IClass te = fd.thrownExceptions[i].getType();
                if (te.isAssignableFrom(type)) return;
            }
        }

        locatable.throwCompileException("Thrown exception of type \"" + type + "\" is neither caught by a \"try...catch\" block nor declared in the \"throws\" clause of the declaring function");
    }

    /**
     * Representation of the Java<sup>TM</sup> "break" statement (JLS 14.14).
     */
    /*package*/ static class BreakStatement extends Statement {
        public BreakStatement(
            Scanner.Location location,
            Scope            enclosingScope,
            String           optionalLabel
        ) {
            super(location, enclosingScope);
            this.optionalLabel = optionalLabel;
        }

        final String optionalLabel;

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            BreakableStatement brokenStatement = null;
            for (Scope s = this.enclosingScope; s instanceof Statement; s = s.getEnclosingScope()) {
                if (s instanceof BreakableStatement) {
                    BreakableStatement bs = (BreakableStatement) s;
                    if (bs instanceof LabeledStatement) {
                        if (
                            this.optionalLabel != null &&
                            ((LabeledStatement) s).label.equals(this.optionalLabel)
                        ) {
                            brokenStatement = bs;
                            break;
                        }
                    } else {
                        brokenStatement = bs;
                        break;
                    }
                }
            }
            if (brokenStatement == null) this.throwCompileException("\"break\" statement is not enclosed by a breakable statement");

            executeFinallyClauses(
                codeAttribute,
                this.enclosingScope,
                brokenStatement.enclosingScope
            );
            codeAttribute.writeBranch(Opcode.GOTO, brokenStatement.getWhereToBreak(codeAttribute));
            return false;
        }
    }

    /**
     * Representation of the Java<sup>TM</sup> "continue" statement (JLS
     * 14.15).
     */
    /*package*/ static class ContinueStatement extends Statement {
        public ContinueStatement(
            Scanner.Location location,
            Scope            enclosingScope,
            String           optionalLabel
        ) {
            super(location, enclosingScope);
            this.optionalLabel = optionalLabel;
        }

        final String optionalLabel;

        // Compile time members:

        public boolean compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            if (optionalLabel != null) this.throwCompileException("\"continue\" with label NYI");

            ContinuableStatement cs = null;
            for (
                Scope s = this.enclosingScope;
                s instanceof Statement;
                s = s.getEnclosingScope()
            ) {
                if (this.optionalLabel != null) {
                    if (s instanceof LabeledStatement) {
                        LabeledStatement ls = (LabeledStatement) s;
                        if (ls.label.equals(this.optionalLabel)) {
                            Statement st = ls.body;
                            while (st instanceof LabeledStatement) st = ((LabeledStatement) st).body;
                            if (!(st instanceof ContinuableStatement)) st.throwCompileException("Labeled statement is not continuable");
                            cs = (ContinuableStatement) st;
                            break;
                        }
                    }
                } else
                if (s instanceof ContinuableStatement) {
                    cs = (ContinuableStatement) s;
                    break;
                }
            }
            if (cs == null) this.throwCompileException("\"continue\" statement is not enclosed by a continuable statement");

            cs.bodyHasContinue = true;
            executeFinallyClauses(
                codeAttribute,
                this.enclosingScope,
                cs.enclosingScope
            );
            codeAttribute.writeBranch(Opcode.GOTO, cs.whereToContinue);
            return false;
        }
    }

    /**
     * Statements that jump out of blocks ("return", "break", "continue")
     * must call this method to make sure that the "finally" clauses of all
     * "try...catch" statements are executed.
     */
    private static void executeFinallyClauses(
        ClassFile.CodeAttribute codeAttribute,
        Scope                   from,
        Scope                   to
    ) {
        for (Scope s = from; s != to; s = s.getEnclosingScope()) {
            if (s instanceof TryStatement) {
                TryStatement ts = (TryStatement) s;
                if (ts.finallyOffset != null) {
                    codeAttribute.writeBranch(Opcode.JSR, ts.finallyOffset);
                }
            }
        }
    }

    /**
     * Abstract base class for {@link Java.Type}, {@link Java.Rvalue} and
     * {@link Java.Lvalue}.
     */
    /*package*/ static abstract class Atom extends Located {
        public Atom(Scanner.Location location) {
            super(location);
        }

        // Parse time members:

        public Type toType() throws Parser.ParseException {
            if (!(this instanceof Type)) this.throwParseException("Expression is not a type");
            return (Type) this;
        }
        public Rvalue toRvalue() throws Parser.ParseException {
            if (!(this instanceof Rvalue)) this.throwParseException("Expression is not an rvalue");
            return (Rvalue) this;
        }
        public Lvalue toLvalue() throws Parser.ParseException {
            if (!(this instanceof Lvalue)) this.throwParseException("Expression is not an lvalue");
            return (Lvalue) this;
        }

        // Compile time members:

        public abstract IClass getType() throws CompileException;
        public boolean isType() throws CompileException {
            return this instanceof Type;
        }
    }

    /**
     * Representation of a Java<sup>TM</sup> type.
     */
    /*package*/ static abstract class Type extends Atom {
        public Type(Scanner.Location location) {
            super(location);
        }
    }

    /**
     * Representation of a Java<sup>TM</sup> "basic type" (obviously
     * equaivalent to a "primitive type") (JLS 4.2).
     */
    /*package*/ static class BasicType extends Type {
        public BasicType(Scanner.Location location, int index) {
            super(location);
            this.index = index;
        }

        public IClass getType() {
            return basicTypes[index];
        }
        public String toString() { return basicTypes[index].toString(); }

        public static final int VOID    = 0;
        public static final int BYTE    = 1;
        public static final int SHORT   = 2;
        public static final int CHAR    = 3;
        public static final int INT     = 4;
        public static final int LONG    = 5;
        public static final int FLOAT   = 6;
        public static final int DOUBLE  = 7;
        public static final int BOOLEAN = 8;

        private static final IClass[] basicTypes = {
            IClass.VOID,
            IClass.BYTE,
            IClass.SHORT,
            IClass.CHAR,
            IClass.INT,
            IClass.LONG,
            IClass.FLOAT,
            IClass.DOUBLE,
            IClass.BOOLEAN,
        };

        private final int index;
    }
    private static final String[] basicTypeNames = { "void", "byte", "short", "char", "int", "long", "float", "double", "boolean", };

    /*package*/ static class ReferenceType extends Type {
        public ReferenceType(
            Scanner.Location location,
            Scope            scope,
            String[]         identifiers
        ) {
            super(location);
            this.scope       = scope;
            this.identifiers = identifiers;
        }

        public IClass getType() throws CompileException {
            Statement                   scopeStatement = null;
            ClassOrInterfaceDeclaration scopeClassOrInterface = null;
            CompilationUnit             scopeCompilationUnit;
            {
                Scope s = this.scope;

                // Determine scope statement.
                if (s instanceof Statement) {
                    scopeStatement = (Statement) s;
                    for (s = s.getEnclosingScope(); s instanceof Statement; s = s.getEnclosingScope());
                }

                if (s instanceof FunctionDeclarator) {
                    s = s.getEnclosingScope();
                }

                // Determine scope class or interface.
                if (s instanceof ClassOrInterfaceDeclaration) {
                    scopeClassOrInterface = (ClassOrInterfaceDeclaration) s;
                }

                // Determine scope compilationUnit.
                if (s != null) while (!(s instanceof CompilationUnit)) s = s.getEnclosingScope();
                scopeCompilationUnit = (CompilationUnit) s;
            }

            if (identifiers.length == 1) {

                // Simple type name (single identifier).
                String simpleTypeName = identifiers[0];

                // 6.5.5.1.1 Local class.
                // NYI.

                // 6.5.5.1.2 Member type.
                for (Scope s = scopeClassOrInterface; s != null; s = s.getEnclosingScope()) {
                    if (s instanceof ClassOrInterfaceDeclaration) {
                        ClassOrInterfaceDeclaration cid = (ClassOrInterfaceDeclaration) s;
                        Vector v = cid.declaredClasses;
                        v.addAll(cid.declaredInterfaces);
                        for (int i = 0; i < v.size(); ++i) {
                            ClassOrInterfaceDeclaration mt = (ClassOrInterfaceDeclaration) v.elementAt(i);
                            if (mt.name.equals(identifiers[0])) return mt;
                        }
                    }
                }

                if (scopeCompilationUnit != null) {

                    // 6.5.5.1.4a Single-type import.
                    String importedClassName = (String) scopeCompilationUnit.importedClasses.get(identifiers[0]);
                    if (importedClassName != null) {
                        IClass res = IClass.iClassFor(Descriptor.fromClassName(importedClassName));
                        if (res == null) this.throwCompileException("Imported class \"" + this.identifiers[0] + "\" could not be loaded");
                        return res;
                    }

                    // 6.5.5.1.4b Type declared in same compilation unit.
                    for (
                        int i = 0;
                        i < scopeCompilationUnit.classOrInterfaceDeclarations.size();
                        ++i
                    ) {
                        ClassOrInterfaceDeclaration cid = (ClassOrInterfaceDeclaration) scopeCompilationUnit.classOrInterfaceDeclarations.elementAt(i);
                        if (cid.name.equals(simpleTypeName)) return cid;
                    }
                }

                // 6.5.5.1.5 Type declared in other compilation unit of same
                // package. NYI.

                // 6.5.5.1.6 Type-import-on-demand declaration.
                {
                    IClass importedClass = null;
                    for (int i = 0; i < scopeCompilationUnit.importedPackages.size(); ++i) {
                        String className = (String) scopeCompilationUnit.importedPackages.elementAt(i) + '.' + identifiers[0];
                        IClass c = IClass.iClassFor(Descriptor.fromClassName(className));
                        if (importedClass != null) this.throwCompileException("Ambiguous class name: \"" + importedClass + "\" vs. \"" + c + "\"");
                        importedClass = c;
                    }
                    if (importedClass != null) return importedClass;
                }

                // 6.5.5.1.8 Give up.
                this.throwCompileException("Cannot determine simple type name \"" + simpleTypeName + "\"");
                /* NEVER REACHED */ return null;
            } else {

                // Qualified type name (two or more identifiers).
                String qualifiedTypeName = join(identifiers, ".");

                // Fully-qualified class name?
                {
                    IClass res = IClass.iClassFor(Descriptor.fromClassName(qualifiedTypeName));
                    if (res != null) return res;
                }

                // Give up.
                this.throwCompileException("Cannot determine qualified type name \"" + qualifiedTypeName + "\"");
                /* NEVER REACHED */ return null;
            }
        }
        public String toString() { return join(identifiers, "."); }

        private final String[] identifiers;
        private final Scope    scope;
    }

    /**
     * Representation of a Java<sup>TM</sup> array type (JLS 10.1).
     */
    /*package*/ static class ArrayType extends Type {
        public ArrayType(Type componentType) {
            super(componentType.getLocation());
            this.componentType = componentType;
        }
        public Type getComponentType() {
            return this.componentType;
        }
        public IClass getType() throws CompileException {
            return this.componentType.getType().getArrayType();
        }

        public String toString() {
            return this.componentType.toString() + "[]";
        }

        private final Type componentType;
    }

    /**
     * Representation of an "rvalue", i.e. an expression that has a type and
     * a value, but cannot be assigned to: An expression that can be the
     * right-hand-side of an assignment.
     */
    /*package*/ static abstract class Rvalue extends Atom {
        protected Rvalue(
            Scanner.Location location
        ) {
            super(location);
        }

        /**
         * Some {@link Java.Rvalue}s compile more efficiently when their value
         * is not needed, e.g. "i++".
         */
        public void compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            Java.pop(codeAttribute, this.compileGetValue(codeAttribute));
        }

        /**
         * Convenience function that calls {@link
         * #compileContext(ClassFile.CodeAttribute)} and {@link
         * #compileGet(ClassFile.CodeAttribute)}.
         * @return The type of the Rvalue
         */
        public final IClass compileGetValue(
            ClassFile.CodeAttribute codeAttribute
        ) throws CompileException {
            this.compileContext(codeAttribute);
            return this.compileGet(codeAttribute);
        }

        /**
         * Generates code that determines the context of the {@link
         * Java.Rvalue} and puts it on the operand stack. Most expressions
         * do not have a "context", but some do. E.g. for "x[y]", the context
         * is "x, y". The bottom line is that for statements like "x[y] += 3"
         * the context is only evaluated once.
         *
         * @return The size of the context on the operand stack
         */
        public int compileContext(
            ClassFile.CodeAttribute codeAttribute
        ) throws CompileException {
            return 0;
        }

        /**
         * Generates code that determines the value of the {@link Java.Rvalue}
         * and puts it on the operand stack. This method relies on that the
         * "context" of the {@link Java.Rvalue} is on top of the operand stack
         * (see {@link #compileContext(ClassFile.CodeAttribute)}).
         *
         * @return The type of the {@link Java.Rvalue}
         */
        public abstract IClass compileGet(
            ClassFile.CodeAttribute codeAttribute
        ) throws CompileException;

        /**
         * Some {@Rvalue}s compile more efficiently when their value is the
         * condition for a branch.
         */
        public void compileBoolean(
            ClassFile.CodeAttribute        codeAttribute,
            ClassFile.CodeAttribute.Offset dst,        // Where to jump.
            boolean                        orientation // JUMP_IF_TRUE or JUMP_IF_FALSE.
        ) throws CompileException {
            Object cv = this.getConstantValue();
            if (cv != null && cv instanceof Boolean) {
                if (orientation == JUMP_IF_TRUE ^ ((Boolean) cv).booleanValue()) {
                    codeAttribute.writeBranch(Opcode.GOTO, dst);
                }
                return;
            }

            if (this.compileGetValue(codeAttribute) != IClass.BOOLEAN) this.throwCompileException("Not a boolean expression");

            codeAttribute.writeBranch(orientation == JUMP_IF_TRUE ? Opcode.IFNE : Opcode.IFEQ, dst);
        }

        /**
         * Attempts to evaluate as a constant expression.
         *
         * @return null if value is not constant; otherwise a String, Byte,
         * Short, Integer, Boolean, Character, Float, Long or Double
         */
        public Object getConstantValue() {
            return null;
        }

        public final static boolean JUMP_IF_TRUE  = true;
        public final static boolean JUMP_IF_FALSE = false;
    }

    /**
     * Base class for {@link Java.Rvalue}s that compile better as conditional
     * branches.
     */
    /*package*/ static abstract class BooleanRvalue extends Rvalue {
        protected BooleanRvalue(Scanner.Location location) {
            super(location);
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            ClassFile.CodeAttribute.Offset isTrue = codeAttribute.new Offset();
            this.compileBoolean(codeAttribute, isTrue, JUMP_IF_TRUE);
            codeAttribute.writeOpcode(Opcode.ICONST_0);
            ClassFile.CodeAttribute.Offset end = codeAttribute.new Offset();
            codeAttribute.writeBranch(Opcode.GOTO, end);
            isTrue.set();
            codeAttribute.writeOpcode(Opcode.ICONST_1);
            end.set();

            return IClass.BOOLEAN;
        }
    }

    /**
     * Representation of an "lvalue", i.e. an expression that has a type and
     * a value, and can be assigned to: An expression that can be the
     * left-hand-side of an assignment.
     */
    /*package*/ static abstract class Lvalue extends Rvalue {
        protected Lvalue(Scanner.Location location) {
            super(location);
        }

        /**
         * Generates code that stores a value in the {@link Java.Lvalue}.
         * Expects the {@link Java.Lvalue}'s context (see {@link
         * #compileContext}) and a value of type "valueType" on the operand
         * stack.
         */
        public abstract void compileSet(
            ClassFile.CodeAttribute codeAttribute,
            IClass                  valueType
        ) throws CompileException;
    }

    /**
     * This class is special: It does not extend/implement the Atom subclasses,
     * but overrides Atom's "to...()" methods.
     */
    /*package*/ static class AmbiguousName extends Lvalue {
        public AmbiguousName(
            Scanner.Location location,
            Scope            scope,
            String[]         identifiers,
            int              n
        ) {
            super(location);
            this.scope       = scope;
            this.identifiers = identifiers;
            this.n           = n;
        }
        public AmbiguousName(
            Scanner.Location location,
            Scope            scope,
            String[]         identifiers
        ) {
            this(location, scope, identifiers, identifiers.length);
        }

        // Override "Atom.toType()".
        public Type toType() throws Parser.ParseException {
            return new Type(this.getLocation()) {
                public boolean isType() throws CompileException {
                    return AmbiguousName.this.reclassify().isType();
                }
                public IClass getType() throws CompileException {
                    try {
                        return AmbiguousName.this.reclassify().toType().getType();
                    } catch (Parser.ParseException ex) {
                        this.throwCompileException(ex.getMessage());
                        /* NEVER REACHED */ return null;
                    }
                }
            };
        }

        // Compile time members.

        // Implement "Type".
        public IClass getType() throws CompileException {
            try {
                return AmbiguousName.this.reclassify().toLvalue().getType();
            } catch (Parser.ParseException ex) {
                this.throwCompileException(ex.getMessage());
                /* NEVER REACHED */ return null;
            }
        }
        public boolean isType() throws CompileException {
            return this.reclassify().isType();
        }
        // Implement "Rvalue".
        public int compileContext(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            try {
                return AmbiguousName.this.reclassify().toRvalue().compileContext(codeAttribute);
            } catch (Parser.ParseException ex) {
                this.throwCompileException(ex.getMessage());
                /* NEVER REACHED */ return 0;
            }
        }
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            try {
                return AmbiguousName.this.reclassify().toRvalue().compileGet(codeAttribute);
            } catch (Parser.ParseException ex) {
                this.throwCompileException(ex.getMessage());
                /* NEVER REACHED */ return null;
            }
        }
        // Implement "Lvalue".
        public void compileSet(
            ClassFile.CodeAttribute codeAttribute,
            IClass                  valueType
        ) throws CompileException {
            try {
                AmbiguousName.this.reclassify().toLvalue().compileSet(
                    codeAttribute,
                    valueType
                );
            } catch (Parser.ParseException ex) {
                this.throwCompileException(ex.getMessage());
            }
        }

        private Atom reclassify() throws CompileException {
            if (this.reclassified == null) {
                this.reclassified = this.uncachedReclassify();
            }
            return this.reclassified;
        }
        private Atom reclassified = null;

        private Atom uncachedReclassify() throws CompileException {

            // Determine scope block, function and compilation Unit.
            Statement                         scopeStatement = null;
            final FunctionDeclarator          scopeFunction;
            final ClassOrInterfaceDeclaration scopeClassOrInterface;
            CompilationUnit                   scopeCompilationUnit = null;
            {
                Scope s = this.scope;
                if (s instanceof Statement) {
                    scopeStatement = (Statement) s;
                    for (s = s.getEnclosingScope(); s instanceof Statement; s = s.getEnclosingScope());
                }
                if (s instanceof FunctionDeclarator) {
                    scopeFunction = (FunctionDeclarator) s;
                    s = s.getEnclosingScope();
                } else {
                    scopeFunction = null;
                }
                if (s instanceof ClassOrInterfaceDeclaration) {
                    scopeClassOrInterface = (ClassOrInterfaceDeclaration) s;
                    s = s.getEnclosingScope();
                } else {
                    scopeClassOrInterface = null;
                }
                while (!(s instanceof CompilationUnit)) s = s.getEnclosingScope();
                scopeCompilationUnit = (CompilationUnit) s;
            }

            if (this.n == 1) {

                // 6.5.2.1

                // 6.5.2.1.1.1 Local variable.
                // 6.5.2.1.1.2 Parameter.
                {
                    LocalVariable lv = null;
                    for (Scope s = this.scope; s instanceof Statement; s = s.getEnclosingScope()) {
                        if (s instanceof Block) {
                            Block b = (Block) s;
                            lv = (LocalVariable) b.localVariables.get(identifiers[0]);
                            if (lv != null) break;
                        }
                    }
                    if (lv == null && scopeFunction != null) {
                        lv = (LocalVariable) scopeFunction.parameters.get(this.identifiers[0]);
                    }
                    if (lv != null) {
                        return new LocalVariableAccess(this.getLocation(), lv);
                    }
                }

                // 6.5.2.1.1.3 Field.
                for (Scope s = scopeClassOrInterface; s != null; s = s.getEnclosingScope()) {
                    if (s instanceof ClassOrInterfaceDeclaration) {
                        final ClassOrInterfaceDeclaration cid = (ClassOrInterfaceDeclaration) s;
                        IClass.IField[] fields = cid.getDeclaredIFields();
                        for (int i = 0; i < fields.length; ++i) {
                            final IClass.IField f = fields[i];
                            if (this.identifiers[0].equals(f.getName())) {
                                Java.Type ct = new Java.Type(scopeClassOrInterface.getLocation()) {
                                    public boolean isType() { return true; }
                                    public IClass getType() { return scopeClassOrInterface; }
                                };
                                return new FieldAccess(
                                    this.getLocation(),
                                    (
                                        scopeFunction == null ?
                                        (Atom) ct /* TBD: Instance initializer */ :
                                        f.isStatic() ?
                                        ct :
                                        (scopeFunction.modifiers & Mod.STATIC) != 0 ?
                                        (Atom) ct :
                                        (Atom) new ThisReference(this.getLocation(), scopeFunction)
                                    ),
                                    f
                                );
                            }
                        }
                    }
                }

                // 6.5.2.1.2.1 Local class -- TBD.

                // 6.5.2.1.2.2 Member type.
                for (Scope s = scopeClassOrInterface; s != null; s = s.getEnclosingScope()) {
                    if (s instanceof ClassOrInterfaceDeclaration) {
                        ClassOrInterfaceDeclaration cid = (ClassOrInterfaceDeclaration) s;
                        Vector v = cid.declaredClasses;
                        v.addAll(cid.declaredInterfaces);
                        for (int i = 0; i < v.size(); ++i) {
                            final ClassOrInterfaceDeclaration mt = (ClassOrInterfaceDeclaration) v.elementAt(i);
                            if (mt.name.equals(identifiers[0])) {
                                return new Type(this.getLocation()) {
                                    public IClass getType() throws CompileException {
                                        return mt;
                                    }
                                };
                            }
                        }
                    }
                }

                // 6.5.2.1.3.1 Single-type-import.
                if (scopeCompilationUnit != null) {
                    String importedClassName = (String) scopeCompilationUnit.importedClasses.get(identifiers[0]);
                    if (importedClassName != null) {
                        final IClass res = IClass.iClassFor(Descriptor.fromClassName(importedClassName));
                        if (res == null) this.throwCompileException("Cannot load imported class \"" + importedClassName + "\"");
                        return new Type(this.getLocation()) {
                            public IClass getType() throws CompileException {
                                return res;
                            }
                        };
                    }
                }

                // 6.5.2.1.3.2 Top level class/interface in this compilation unit.
                if (scopeCompilationUnit != null) {
                    Vector coids = scopeCompilationUnit.classOrInterfaceDeclarations;
                    for (int i = 0; i < coids.size(); ++i) {
                        final ClassOrInterfaceDeclaration coid = (ClassOrInterfaceDeclaration) coids.elementAt(i);
                        if (coid.name.equals(identifiers[0])) {
                            return new Type(this.getLocation()) {
                                public IClass getType() throws CompileException {
                                    return coid;
                                }
                            };
                        }
                    }
                }

                // 6.5.2.1.4 Class or interface in same package.
                if (scopeCompilationUnit != null && scopeCompilationUnit.packagE != null) {
                    String className = scopeCompilationUnit.packagE + '.' + identifiers[0];
                    final IClass res = IClass.iClassFor(Descriptor.fromClassName(className));
                    if (res != null) {
                        return new Type(this.getLocation()) {
                            public IClass getType() throws CompileException {
                                return res;
                            }
                        };
                    }
                }

                // 6.5.2.1.5, 6.5.2.1.6 Type-import-on-demand.
                if (scopeCompilationUnit != null) {
                    IClass importedClass = null;
                    for (int i = 0; i < scopeCompilationUnit.importedPackages.size(); ++i) {
                        String className = (String) scopeCompilationUnit.importedPackages.elementAt(i) + '.' + identifiers[0];
                        IClass c = IClass.iClassFor(Descriptor.fromClassName(className));
                        if (importedClass != null) this.throwCompileException("Ambiguous class name: \"" + importedClass + "\" vs. \"" + c + "\"");
                        importedClass = c;
                    }
                    final IClass fic = importedClass;
                    return new Type(this.getLocation()) {
                        public IClass getType() throws CompileException {
                            return fic;
                        }
                    };
                }

                this.throwCompileException("Cannot determine ambiguous name \"" + identifiers[0] + "\"");
            } // this.n == 1

            // 6.5.2.2
            Atom lhs = new AmbiguousName(
                this.getLocation(),
                this.scope,
                this.identifiers, this.n - 1
            ).uncachedReclassify();
            IClass lhsType = lhs.getType();
            String rhs = this.identifiers[this.n - 1];

            // 6.5.2.2.2
            // 6.5.2.2.3
            if (rhs.equals("length")) {
                try {
                    return new ArrayLength(this.getLocation(), lhs.toRvalue());
                } catch (Parser.ParseException ex) {
                    this.throwCompileException(ex.getMessage());
                }
            }

            IClass.IMethod[] methods = lhsType.getDeclaredIMethods();
            for (int i = 0; i < methods.length; ++i) {
                if (methods[i].getName().equals(rhs)) {

                    // 6.5.2.2.2.1
                    // 6.5.2.2.3.1
                    this.throwCompileException("Method name \"" + rhs + "\" cannot be used in this context");
                }
            }
            IClass.IField[] fields = lhsType.getDeclaredIFields();
            for (int i = 0; i < fields.length; ++i) {
                IClass.IField f = fields[i];
                if (f.getName().equals(rhs)) {
                    // 6.5.2.2.2.2 TYPE.FIELD
                    // 6.5.2.2.3 EXPRESSION.FIELD
                    return new FieldAccess(
                        this.getLocation(), // location
                        lhs,                // lhs
                        f
                    );
                }
            }
            IClass[] classes = lhsType.getDeclaredIClasses();
            for (int i = 0; i < classes.length; ++i) {
                final IClass memberType = classes[i];
                String name = Descriptor.toClassName(memberType.getDescriptor());
                name = name.substring(name.lastIndexOf('.') + 1);
                if (name.equals(rhs)) {

                    // 6.5.2.2.2.3 TYPE.TYPE
                    // 6.5.2.2.3.3 EXPRESSION.TYPE
                    return new Type(this.getLocation()) {
                        public IClass getType() { return memberType; }
                    };
                }
            }

            this.throwCompileException("\"" + rhs + "\" is neither a method, a field, nor a member class of \"" + lhsType + "\"");
            /* NEVER REACHED */ return null;
        }

        private final Scope    scope;
        private final String[] identifiers;
        private final int      n;
    }

    private static class LocalVariableAccess extends Lvalue {
        public LocalVariableAccess(
            Scanner.Location location,
            LocalVariable    localVariable
        ) {
            super(location);
            this.localVariable = localVariable;
        }

        private /*final*/ LocalVariable localVariable;

        // Compile time members.

        public IClass getType() { return this.localVariable.type; }
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            return Java.load(codeAttribute, this.localVariable);
        }
        public void compileSet(
            ClassFile.CodeAttribute codeAttribute,
            IClass                  valueType
        ) throws CompileException {
            Java.store(
                codeAttribute,
                (Locatable) this,
                valueType,
                this.localVariable
            );
        }
    }

    static private class FieldAccess extends Lvalue {
        public FieldAccess(
            Scanner.Location location,
            Atom             lhs,
            IClass.IField    field
        ) {
            super(location);
            this.lhs   = lhs;
            this.field = field;
        }

        private final Atom          lhs;
        private final IClass.IField field;

        // Compile time members.

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.field.getType();
        }
        // Implement "Rvalue".
        public int compileContext(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            try {
                if (this.field.isStatic()) {
                    this.lhs.toType().getType();
                    return 0;
                } else {
                    this.lhs.toRvalue().compileGetValue(codeAttribute);
                    return 1;
                }
            } catch (Parser.ParseException ex) {
                this.throwCompileException(ex.getMessage());
                /* NEVER REACHED */ return 0;
            }
        }
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            if (this.field.isStatic()) {
                codeAttribute.writeOpcode(Opcode.GETSTATIC);
            } else {
                codeAttribute.writeOpcode(Opcode.GETFIELD);
            }
            codeAttribute.writeShort(codeAttribute.getClassFile().addConstantFieldrefInfo(
                this.field.getDeclaringIClass().getDescriptor(), // classFD
                this.field.getName(),                            // fieldName
                this.field.getType().getDescriptor()             // fieldFD
            ));
            return this.field.getType();
        }
        public void compileSet(
            ClassFile.CodeAttribute codeAttribute,
            IClass                  valueType
        ) throws CompileException {
            Java.assignmentConversion(
                codeAttribute,       // codeAttribute
                (Locatable) this,    // locatable
                valueType,           // sourceType
                this.field.getType() // targetType
            );
            short cfri = codeAttribute.getClassFile().addConstantFieldrefInfo(
                this.field.getDeclaringIClass().getDescriptor(), // classFD
                this.field.getName(),                            // fieldName
                this.field.getDescriptor()                       // fieldFD
            );
            codeAttribute.writeOpcode(
                this.field.isStatic() ?
                Opcode.PUTSTATIC :
                Opcode.PUTFIELD
            );
            codeAttribute.writeShort(cfri);
        }
    }

    static private class ArrayLength extends Rvalue {
        public ArrayLength(
            Scanner.Location location,
            Rvalue           lhs
        ) {
            super(location);
            this.lhs   = lhs;
        }

        private final Rvalue lhs;

        // Compile time members.

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return IClass.INT;
        }
        // Implement "Rvalue".
        public int compileContext(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            if (!this.lhs.compileGetValue(codeAttribute).isArray()) this.throwCompileException("Cannot determine length of non-array type");
            return 1;
        }
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            codeAttribute.writeOpcode(Opcode.ARRAYLENGTH);
            return IClass.INT;
        }
    }

    private static class ThisReference extends Rvalue {
        public ThisReference(
            Scanner.Location   location,
            FunctionDeclarator function
        ) throws CompileException {
            super(location);
            if ((function.modifiers & Mod.STATIC) != 0) this.throwCompileException("No current instance available in static method");
            this.function = function;
        }

        private final FunctionDeclarator function;

        // Compile time members.

        //Implement "Atom".
        public IClass getType() throws CompileException {
            return this.function.declaringClassOrInterface;
        }
        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            codeAttribute.writeOpcode(Opcode.ALOAD_0);
            return this.function.declaringClassOrInterface;
        }
    }

    /**
     * Loads the value of "field" of "classType" onto the stack.
     *
     * If "staticContext" is true, then the field must be static. Otherwise,
     * a reference to an instance of "classType" is expected on the operand
     * stack, and the field must be non-static.
     *
     * @return The type of the field value.
     */
/*
    private static IClass compileGetFieldValue(
        ClassFile.CodeAttribute codeAttribute,
        Locatable               locatable,
        IClass                  classType,
        boolean                 staticContext,
        IClass.IField           field
    ) throws CompileException {
        if (staticContext) {
            if (!field.isStatic()) locatable.throwCompileException("Non-static field \"" + field.getName() + "\" of type \"" + classType + "\" cannot be accessed in static context");
            codeAttribute.writeOpcode(Opcode.GETSTATIC);
        } else {
            if (field.isStatic()) locatable.throwCompileException("Static field \"" + field.getName() + "\" of type \"" + classType + "\" cannot be accessed in non-static context");
            codeAttribute.writeOpcode(Opcode.GETFIELD);
        }
        IClass ft = field.getType();
        codeAttribute.writeShort(codeAttribute.getClassFile().addConstantFieldrefInfo(
            classType.getDescriptor(), // classFD
            field.getName(),           // fieldName
            ft.getDescriptor()         // fieldFD
        ));
        return ft;
    }
*/
    /**
     * Sets the value of "field" of "classType".
     *
     * If "staticContext" is true, then the field must be static. The top value
     * on the operand stack of "valueType" undergoes assignment conversion and
     * is then assigned to the static field.
     *
     * If "staticContext" is then, then the field must be non-static. Two
     * values are expected on the operand stack: A reference to an instance
     * of "classType" and a value of type "valueType", which undergoes
     * assignment conversion and is then assigned to the non-static field.
     */
/*
    private static void compileSetFieldValue(
        ClassFile.CodeAttribute codeAttribute,
        Locatable               locatable,
        IClass                  classType,
        IClass                  valueType,
        boolean                 staticContext,
        IClass.IField           field
    ) throws CompileException {
        IClass fieldType = field.getType();
        Java.assignmentConversion(codeAttribute, locatable, valueType, fieldType);
        if (staticContext) {
            if (!field.isStatic()) locatable.throwCompileException("Non-static field \"" + field.getName() + "\" of type \"" + classType + "\" cannot be accessed in static context");
            codeAttribute.writeOpcode(Opcode.PUTSTATIC);
        } else {
            if (field.isStatic()) locatable.throwCompileException("Static field \"" + field.getName() + "\" of type \"" + classType + "\" cannot be accessed in non-static context");
            codeAttribute.writeOpcode(Opcode.PUTFIELD);
        }
        codeAttribute.writeShort(codeAttribute.getClassFile().addConstantFieldrefInfo(
            classType.getDescriptor(), // classFD
            field.getName(),           // fieldName
            fieldType.getDescriptor()  // fieldFD
        ));
    }
*/
    /*package*/ static class Assignment extends Rvalue {
        public Assignment(
            Scanner.Location location,
            Lvalue           lhs,
            String           operator,
            Rvalue           rhs
        ) {
            super(location);
            this.lhs      = lhs;
            this.operator = operator;
            this.rhs      = rhs;
        }

        private final Lvalue lhs;
        private final String operator;
        private final Rvalue rhs;

        // Compile time members.

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.lhs.getType();
        }

        // Implement "Rvalue".
        public void compile(ClassFile.CodeAttribute codeAttribute) throws Java.CompileException {
            if (operator == "=") {
                this.lhs.compileContext(codeAttribute);
                IClass rhsType = this.rhs.compileGetValue(codeAttribute);
                this.lhs.compileSet(codeAttribute, rhsType);
                return;
            }

            // Implement "|= ^= &= *= /= %= += -= <<= >>= >>>=".
            int lhsCS = this.lhs.compileContext(codeAttribute);
            Java.dup(codeAttribute, lhsCS);
            IClass lhsType = (IClass) this.lhs.compileGet(codeAttribute);
            ClassFile.CodeAttribute.Inserter convertLhsInserter = codeAttribute.newInserter();
            IClass rhsType = this.rhs.compileGetValue(codeAttribute);
            IClass resultType = Java.compileNonBooleanBinaryOperation(
                (Locatable) this,           // locatable
                codeAttribute,              // codeAttribute
                lhsType,                    // lhsType
                convertLhsInserter,         // convertLhsInserter
                operator.substring(         // op
                    0,
                    operator.length() - 1
                ).intern(), // <= IMPORTANT!
                rhsType                     // rhsType
            );
            if (
                !Java.tryIdentityConversion(resultType, lhsType) &&
                !Java.tryNarrowingPrimitiveConversion(
                    codeAttribute,    // codeAttribute
                    (Locatable) this, // locatable
                    resultType,       // sourceType
                    lhsType           // destinationType
                )
            ) throw new RuntimeException();
            this.lhs.compileSet(codeAttribute, lhsType);
        }
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws Java.CompileException {
            if (operator != "=") this.throwCompileException("Operator\"" + this.operator + "\" as an rvalue NYI");

            // Compile LHS context.
            int lhsCS = this.lhs.compileContext(codeAttribute);
            // Compile RHS.
            IClass rhsType = this.rhs.compileGetValue(codeAttribute);
            // DUPX the RHS value.
            Java.dupx(
                codeAttribute, // codeAttribute
                rhsType,       // type
                lhsCS          // x
            );
            // Set LHS.
            this.lhs.compileSet(codeAttribute, rhsType);
            return rhsType;
        }
    }

    /*package*/ static class ConditionalExpression extends Rvalue {
        public ConditionalExpression(
            Scanner.Location location,
            Rvalue           lhs,
            Rvalue           mhs,
            Rvalue           rhs
        ) {
            super(location);
            this.lhs = lhs;
            this.mhs = mhs;
            this.rhs = rhs;
        }

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return IClass.BOOLEAN;
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            ClassFile.CodeAttribute.Offset toRhs = codeAttribute.new Offset();
            ClassFile.CodeAttribute.Offset toEnd = codeAttribute.new Offset();

            lhs.compileBoolean(codeAttribute, toRhs, JUMP_IF_FALSE);
            IClass mhsType = mhs.compileGetValue(codeAttribute);
            codeAttribute.writeBranch(Opcode.GOTO, toEnd);
            toRhs.set();
            IClass rhsType = rhs.compileGetValue(codeAttribute);
            toEnd.set();

            if (mhsType != rhsType) this.throwCompileException("Inconsistent expression type : \"" + mhsType + "\" vs. \"" + rhsType + "\"");
            return mhsType;
        }

        private final Rvalue lhs, mhs, rhs;
    }

    /**
     * Objects of this class represent represent one pre- or post-increment
     * or decrement.
     */
    /*package*/ static class Crement extends Rvalue {
        public Crement(Scanner.Location location, String operator, Lvalue operand) {
            super(location);
            this.pre      = true;
            this.operator = operator;
            this.operand  = operand;
        }
        public Crement(Scanner.Location location, Lvalue operand, String operator) {
            super(location);
            this.pre      = false;
            this.operator = operator;
            this.operand  = operand;
        }

        private final boolean pre;
        private final String  operator;
        private final Lvalue  operand;

        // Compile time members.

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.operand.getType();
        }

        // Implement "Rvalue".
        public void compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {

            // Optimized crement of integer local variable.
            LocalVariable lv = this.isIntLV();
            if (lv != null) {
                codeAttribute.writeOpcode(Opcode.IINC);
                codeAttribute.writeByte(lv.localVariableArrayIndex);
                codeAttribute.writeByte(operator == "++" ? 1 : -1);
                return;
            }

            int cs = this.operand.compileContext(codeAttribute);
            Java.dup(codeAttribute, cs);
            IClass type = this.operand.compileGet(codeAttribute);
            IClass promotedType = unaryNumericPromotion(
                codeAttribute,
                (Locatable) this,
                type
            );
            codeAttribute.writeOpcode(Java.ilfd(
                promotedType,
                Opcode.ICONST_1,
                Opcode.LCONST_1,
                Opcode.FCONST_1,
                Opcode.DCONST_1
            ));
            if (this.operator == "++") {
                codeAttribute.writeOpcode(Opcode.IADD + Java.ilfd(promotedType));
            } else
            if (this.operator == "--") {
                codeAttribute.writeOpcode(Opcode.ISUB + Java.ilfd(promotedType));
            } else {
                this.throwCompileException("Unexpected operator \"" + this.operator + "\"");
            }

            if (
                !Java.tryIdentityConversion(promotedType, type) &&
                !Java.tryNarrowingPrimitiveConversion(
                    codeAttribute,    // locatable
                    (Locatable) this, // locatable
                    promotedType,     // sourceType
                    type              // targetType
                )
            ) throw new RuntimeException();
            this.operand.compileSet(codeAttribute, type);
        }

        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {

            // Optimized crement of integer local variable.
            LocalVariable lv = this.isIntLV();
            if (lv != null) {
                if (!this.pre) Java.load(codeAttribute, lv);
                codeAttribute.writeOpcode(Opcode.IINC);
                codeAttribute.writeByte(lv.localVariableArrayIndex);
                codeAttribute.writeByte(operator == "++" ? 1 : -1);
                if (this.pre) Java.load(codeAttribute, lv);
                return lv.type;
            }

            // Compile operand context.
            int cs = this.operand.compileContext(codeAttribute);
            // DUP operand context.
            Java.dup(codeAttribute, cs);
            // Get operand value.
            IClass type = this.operand.compileGet(codeAttribute);
            // DUPX operand value.
            if (!this.pre) Java.dupx(codeAttribute, type, cs);
            // Apply "unary numeric promotion".
            IClass promotedType = unaryNumericPromotion(
                codeAttribute,
                (Locatable) this,
                type
            );
            // Crement.
            codeAttribute.writeOpcode(Java.ilfd(
                promotedType,
                Opcode.ICONST_1,
                Opcode.LCONST_1,
                Opcode.FCONST_1,
                Opcode.DCONST_1
            ));
            if (this.operator == "++") {
                codeAttribute.writeOpcode(Opcode.IADD + Java.ilfd(promotedType));
            } else
            if (this.operator == "--") {
                codeAttribute.writeOpcode(Opcode.ISUB + Java.ilfd(promotedType));
            } else {
                this.throwCompileException("Unexpected operator \"" + this.operator + "\"");
            }
            // Reverse "unary numeric promotion".
            if (
                !Java.tryIdentityConversion(promotedType, type) &&
                !Java.tryNarrowingPrimitiveConversion(
                    codeAttribute,    // locatable
                    (Locatable) this, // locatable
                    promotedType,     // sourceType
                    type              // targetType
                )
            ) throw new RuntimeException();
            // DUPX cremented operand value.
            if (this.pre) Java.dupx(codeAttribute, type, cs);
            // Set operand.
            this.operand.compileSet(codeAttribute, type);

            return type;
        }

        /**
         * Checks whether the operand is an integer-like local variable.
         */
        private LocalVariable isIntLV() throws CompileException {
            if (!(this.operand instanceof AmbiguousName)) return null;
            AmbiguousName an = (AmbiguousName) this.operand;

            Atom rec = an.reclassify();
            if (!(rec instanceof LocalVariableAccess)) return null;
            LocalVariableAccess lva = (LocalVariableAccess) rec;

            LocalVariable lv = lva.localVariable;
            if (lv.finaL) lva.throwCompileException("Must not increment or decrement \"final\" local variable");
            if (
                lv.type == IClass.BYTE  ||
                lv.type == IClass.SHORT ||
                lv.type == IClass.INT   ||
                lv.type == IClass.CHAR
            ) return lv;
            return null;
        }
    }

    /**
     * This class implements an array access.
     */
    /*package*/ static class ArrayAccessExpression extends Lvalue {
        public ArrayAccessExpression(
            Scanner.Location location,
            Rvalue           lhs,
            Rvalue           index
        ) {
            super(location);
            this.lhs = lhs;
            this.index = index;
        }

        private final Rvalue lhs;
        private final Rvalue index;

        // Compile time members:

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.lhs.getType().getComponentType();
        }

        // Implement "Rvalue".
        public int compileContext(ClassFile.CodeAttribute codeAttribute) throws Java.CompileException {
            IClass lhsType = this.lhs.compileGetValue(codeAttribute);
            if (!lhsType.isArray()) this.throwCompileException("Subscript not allowed on non-array type \"" + lhsType.toString() + "\"");

            IClass indexType = this.index.compileGetValue(codeAttribute);
            if (
                !Java.tryIdentityConversion(indexType, IClass.INT) &&
                !Java.tryWideningPrimitiveConversion(
                    codeAttribute,         // codeAttribute
                    (Java.Locatable) this, // locatable
                    indexType,             // sourceType
                    IClass.INT             // targetType
                )
            ) this.throwCompileException("Index expression of type \"" + indexType + "\" cannot be widened to \"int\"");

            return 2;
        }
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws Java.CompileException {
            IClass lhsComponentType = this.getType();
            codeAttribute.writeOpcode(Opcode.IALOAD + Java.ilfda(lhsComponentType));
            return lhsComponentType;
        }

        // Implement Lvalue
        public void compileSet(
            ClassFile.CodeAttribute codeAttribute,
            IClass                  valueType
        ) throws Java.CompileException {
            IClass lhsComponentType = this.getType();
            Java.assignmentConversion(
                codeAttribute,    // codeAttribute
                (Locatable) this, // locatable
                valueType,        // sourceType
                lhsComponentType  // targetType
            );
            codeAttribute.writeOpcode(Opcode.IASTORE + Java.ilfda(lhsComponentType));
        }
    }

    /**
     * This class implements class field access.
     */
    /*package*/ static class FieldAccessExpression extends Lvalue {
        public FieldAccessExpression(
            Scanner.Location location,
            Atom             lhs,
            String           fieldName
        ) {
            super(location);
            this.lhs       = lhs;
            this.fieldName = fieldName;
        }

        private final Atom   lhs;
        private final String fieldName;

        // Compile time members:

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return (
                this.fieldName.equals("length") ?
                IClass.INT :
                this.findField().getType()
            );
        }

        // Implement "Rvalue".
        public int compileContext(ClassFile.CodeAttribute codeAttribute) throws Java.CompileException {
            try {
                this.value = (
                    this.fieldName.equals("length") ?
                    (Rvalue) new ArrayLength(
                        this.getLocation(),
                        this.lhs.toRvalue()
                    ) :
                    (Rvalue) new FieldAccess(
                        this.getLocation(),
                        this.lhs,
                        this.findField()
                    )
                );
            } catch (Parser.ParseException ex) {
                this.throwCompileException(ex.getMessage());
            }
            return this.value.compileContext(codeAttribute);
        }
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws Java.CompileException {
            return this.value.compileGet(codeAttribute);
        }

        // Implement Lvalue
        public void compileSet(
            ClassFile.CodeAttribute codeAttribute,
            IClass                  valueType
        ) throws Java.CompileException {
            try {
                this.value.toLvalue().compileSet(codeAttribute, valueType);
            } catch (Parser.ParseException ex) {
                this.throwCompileException(ex.getMessage());
            }
        }

        private IClass.IField findField() throws CompileException {
            IClass lhsType = this.lhs.getType();
            IClass.IField[] fields = lhsType.getDeclaredIFields();
            for (int i = 0; i < fields.length; ++i) {
                IClass.IField f = fields[i];
                if (f.getName().equals(this.fieldName)) return f;
            }
            this.throwCompileException("\"" + lhsType + "\" has no field \"" + this.fieldName + "\"");
            /* NEVER REACHED */ return null;
        }

        private Rvalue value = null;
    }

    /**
     * This class implements the unary operators "+", "-", "~" and "!".
     */
    /*package*/ static class UnaryOperation extends BooleanRvalue {
        public UnaryOperation(
            Scanner.Location location,
            String           operator,
            Rvalue           operand
        ) {
            super(location);
            this.operator = operator;
            this.operand  = operand;
        }

        // Implement "Atom".
        public IClass getType() throws CompileException {
            if (this.operator == "!") return IClass.BOOLEAN;
            if (this.operator == "+") return this.operand.getType();
            if (
                this.operator == "-" ||
                this.operator == "~"
            ) return unaryNumericPromotion(this, this.operand.getType());

            this.throwCompileException("Unexpected operator \"" + operator + "\"");
            /* NEVER REACHED */ return null;
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            if (operator == "!") {
                return super.compileGetValue(codeAttribute);
            }

            if (operator == "+") {
                return operand.compileGetValue(codeAttribute);
            }

            if (operator == "-") {
                IClass operandType = operand.compileGetValue(codeAttribute);

                IClass promotedType = unaryNumericPromotion(codeAttribute, (Locatable) this, operandType);
                codeAttribute.writeOpcode(Opcode.INEG + Java.ilfd(promotedType));
                return promotedType;
            }

            if (operator == "~") {
                IClass operandType = operand.compileGetValue(codeAttribute);

                IClass promotedType = unaryNumericPromotion(codeAttribute, (Locatable) this, operandType);
                if (promotedType == IClass.INT) {
                    codeAttribute.writeOpcode(Opcode.ICONST_M1);
                    codeAttribute.writeOpcode(Opcode.IXOR);
                    return IClass.INT;
                }
                if (promotedType == IClass.LONG) {
                    codeAttribute.writeOpcode(Opcode.LDC2_W);
                    codeAttribute.writeShort(codeAttribute.getClassFile().addConstantLongInfo(-1L));
                    codeAttribute.writeOpcode(Opcode.LXOR);
                    return IClass.LONG;
                }
                this.throwCompileException("Operator \"~\" not applicable to type \"" + promotedType + "\"");
            }

            this.throwCompileException("Unexpected operator \"" + operator + "\"");
            /* NEVER REACHED */ return null;
        }

        // Implement Rvalue
        public void compileBoolean(
            ClassFile.CodeAttribute        codeAttribute,
            ClassFile.CodeAttribute.Offset dst,        // Where to jump.
            boolean                        orientation // JUMP_IF_TRUE or JUMP_IF_FALSE.
        ) throws CompileException {
            if (operator == "!") {
                operand.compileBoolean(codeAttribute, dst, !orientation);
                return;
            }

            this.throwCompileException("Boolean expression expected");
        }

        private final String operator;
        private final Rvalue operand;
    }

    /*package*/ static class Instanceof extends Rvalue {
        public Instanceof(
            Scanner.Location location,
            Rvalue           lhs,
            ReferenceType    rhs
        ) {
            super(location);
            this.lhs = lhs;
            this.rhs = rhs;
        }

        private final Rvalue        lhs;
        private final ReferenceType rhs;

        // Compile time members.

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return IClass.BOOLEAN;
        }

        // Implement Rvalue
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            IClass lhsType = this.lhs.compileGetValue(codeAttribute);
            IClass rhsType = this.rhs.getType();
            if (
                !lhsType.isAssignableFrom(rhsType) &&
                !rhsType.isAssignableFrom(lhsType)
            ) this.throwCompileException("\"" + lhsType + "\" can never be an instance of \"" + rhsType + "\"");
            codeAttribute.writeOpcode(Opcode.INSTANCEOF);
            codeAttribute.writeShort(
                codeAttribute.getClassFile().addConstantClassInfo(
                    rhsType.getDescriptor()
                )
            );
            return IClass.BOOLEAN;
        }
    }

    /**
     * Representation of all non-operand-modifying Java<sup>TM</sup> binary
     * operations.
     * <p>
     * Operations with boolean result:<br>
     * <tt>|| && == != < > <= >=</tt>
     * <p>
     * Operations with non-boolean result:<br>
     * <tt>| ^ & * / % + - << >> >>></tt>
     */
    /*package*/ static class BinaryOperation extends BooleanRvalue {
        public BinaryOperation(
            Scanner.Location location,
            Rvalue           lhs,
            String           op,
            Rvalue           rhs
        ) {
            super(location);
            this.lhs = lhs;
            this.op = op;
            this.rhs = rhs;
        }

        private final Rvalue lhs;
        private final String op;
        private final Rvalue rhs;

        // Compile time members.

        // Implement "Atom".
        public IClass getType() throws CompileException {
            if (
                this.op == "||" ||
                this.op == "&&" ||
                this.op == "==" ||
                this.op == "!=" ||
                this.op == "<"  ||
                this.op == ">"  ||
                this.op == "<=" ||
                this.op == ">="
            ) return IClass.BOOLEAN;

            if (
                this.op == "|" ||
                this.op == "^" ||
                this.op == "&"
            ) return (
                this.lhs.getType() == IClass.BOOLEAN ?
                IClass.BOOLEAN :
                Java.binaryNumericPromotion(
                    (Locatable) this,
                    this.lhs.getType(),
                    this.rhs.getType()
                )
            );
            if (
                this.op == "*"   ||
                this.op == "/"   ||
                this.op == "%"   ||
                this.op == "+"   ||
                this.op == "-"   ||
                this.op == "<<"  ||
                this.op == ">>"  ||
                this.op == ">>>"
            ) {
                IClass lhsType = this.lhs.getType();
                IClass rhsType = this.rhs.getType();
                return (
                    this.op == "+" && (
                        lhsType == IClass.STRING ||
                        rhsType == IClass.STRING
                    ) ?
                    IClass.STRING :
                    Java.binaryNumericPromotion((Locatable) this, lhsType, rhsType)
                );
            }

            this.throwCompileException("Unexpected operator \"" + this.op + "\"");
            /* NEVER REACHED */ return null;
        }

        // Implement Rvalue
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            if (
                op == "||" ||
                op == "&&" ||
                op == "==" ||
                op == "!=" ||
                op == "<"  ||
                op == ">"  ||
                op == "<=" ||
                op == ">="
            ) {
                // Eventually calls "compileBoolean()".
                return super.compileGetValue(codeAttribute);
            }

            IClass lhsType = lhs.compileGetValue(codeAttribute);
            ClassFile.CodeAttribute.Inserter convertLhsInserter = codeAttribute.newInserter();
            IClass rhsType = rhs.compileGetValue(codeAttribute);

            // Implements "| ^ & * / % + - << >> >>>".
            return Java.compileNonBooleanBinaryOperation(
                (Locatable) this,
                codeAttribute,
                lhsType, convertLhsInserter, this.op, rhsType
            );
        }

        // Implement Rvalue
        public void compileBoolean(
            ClassFile.CodeAttribute        codeAttribute,
            ClassFile.CodeAttribute.Offset dst,        // Where to jump.
            boolean                        orientation // JUMP_IF_TRUE or JUMP_IF_FALSE.
        ) throws CompileException {
            if (op == "|" || op == "^" || op == "&") {
                super.compileBoolean(codeAttribute, dst, orientation);
                return;
            }

            if (op == "||" || op == "&&") {
                if (op == "||" ^ orientation == JUMP_IF_FALSE) {
                    lhs.compileBoolean(codeAttribute, dst, JUMP_IF_TRUE ^ orientation == JUMP_IF_FALSE);
                    rhs.compileBoolean(codeAttribute, dst, JUMP_IF_TRUE ^ orientation == JUMP_IF_FALSE);
                } else {
                    ClassFile.CodeAttribute.Offset end = codeAttribute.newOffset();
                    lhs.compileBoolean(codeAttribute, end, JUMP_IF_FALSE ^ orientation == JUMP_IF_FALSE);
                    rhs.compileBoolean(codeAttribute, dst, JUMP_IF_TRUE ^ orientation == JUMP_IF_FALSE);
                    end.set();
                }
                return;
            }

            if (
                op == "==" ||
                op == "!=" ||
                op == "<=" ||
                op == ">=" ||
                op == "<"  ||
                op == ">"
            ) {
                int opIdx = (
                    op == "==" ? 0 :
                    op == "!=" ? 1 :
                    op == "<"  ? 2 :
                    op == ">=" ? 3 :
                    op == ">"  ? 4 :
                    op == "<=" ? 5 : Integer.MIN_VALUE
                );
                if (orientation == JUMP_IF_FALSE) opIdx ^= 1;

                IClass lhsType = lhs.compileGetValue(codeAttribute);
                ClassFile.CodeAttribute.Inserter convertLhsInserter = codeAttribute.newInserter();
                IClass rhsType = rhs.compileGetValue(codeAttribute);

                if (
                    lhsType.isPrimitiveNumeric() &&
                    rhsType.isPrimitiveNumeric()
                ) {
                    IClass promotedType = binaryNumericPromotion(codeAttribute, (Locatable) this, lhsType, convertLhsInserter, rhsType);
                    if (promotedType == IClass.INT) {
                        codeAttribute.writeBranch(Opcode.IF_ICMPEQ + opIdx, dst);
                    } else
                    if (promotedType == IClass.LONG) {
                        codeAttribute.writeOpcode(Opcode.LCMP);
                        codeAttribute.writeBranch(Opcode.IFEQ + opIdx, dst);
                    } else
                    if (promotedType == IClass.FLOAT) {
                        codeAttribute.writeOpcode(Opcode.FCMPG);
                        codeAttribute.writeBranch(Opcode.IFEQ + opIdx, dst);
                    } else
                    if (promotedType == IClass.DOUBLE) {
                        codeAttribute.writeOpcode(Opcode.DCMPG);
                        codeAttribute.writeBranch(Opcode.IFEQ + opIdx, dst);
                    } else
                    {
                        throw new RuntimeException("Unexpected promoted type \"" + promotedType + "\"");
                    }
                    return;
                }

                if (
                    lhsType == IClass.BOOLEAN &&
                    rhsType == IClass.BOOLEAN
                ) {
                    if (op != "==" && op != "!=") {
                        this.throwCompileException("Operator \"" + op + "\" not allowed on boolean operands");
                    }
                    codeAttribute.writeBranch(Opcode.IF_ICMPEQ + opIdx, dst);
                    return;
                }

                if (
                    !lhsType.isPrimitive() &&
                    !rhsType.isPrimitive()
                ) this.throwCompileException("Reference equality operators NYI");

                this.throwCompileException("Cannot compare types \"" + lhsType + "\" and \"" + rhsType + "\"");
            }

            this.throwCompileException("Boolean expression expected");
        }
        public Object getConstantValue() {
            if (op == "+") {
                Object lhsValue = lhs.getConstantValue();
                Object rhsValue = rhs.getConstantValue();
                if (lhsValue != null && rhsValue != null) {

                    // Constant string concatenation.
                    if (lhsValue instanceof String || rhsValue instanceof String) {
                        return lhsValue.toString() + rhsValue.toString();
                    }

                    // TBD: Other binary operations on constant operands.
                    ;
                }
            }
            return null;
        }
    }

    /**
     * Execute non-boolean binary operation. LHS and RHS are expected on the
     * stack. This method may need to convert the LHS which it does by
     * inserting Code through the "convertLhsInserter".
     *
     * The following operators are supported:
     *   | ^ & * / % + - << >> >>>
     *
     * @param op The operand as an interned String.
     * @return The type of the result.
     */
    private static IClass compileNonBooleanBinaryOperation(
        Locatable                        locatable,
        ClassFile.CodeAttribute          codeAttribute,
        IClass                           lhsType,
        ClassFile.CodeAttribute.Inserter convertLhsInserter,
        String                           op,
        IClass                           rhsType
    ) throws CompileException {
        if (
            op == "|" ||
            op == "^" ||
            op == "&"
        ) {
            int opIdx = (
                op == "&" ? 0 :
                op == "|" ? 2 :
                op == "^" ? 4 : Integer.MIN_VALUE
            );

            if (
                lhsType.isPrimitiveNumeric() &&
                rhsType.isPrimitiveNumeric()
            ) {
                IClass promotedType = binaryNumericPromotion(codeAttribute, locatable, lhsType, convertLhsInserter, rhsType);
                if (promotedType == IClass.INT) {
                    codeAttribute.writeOpcode(Opcode.IAND + opIdx);
                } else
                if (promotedType == IClass.LONG) {
                    codeAttribute.writeOpcode(Opcode.LAND + opIdx);
                } else
                {
                    locatable.throwCompileException("Operator \"" + op + "\" not defined on types \"" + lhsType + "\" and \"" + rhsType + "\"");
                }
                return promotedType;
            }

            if (
                lhsType == IClass.BOOLEAN &&
                rhsType == IClass.BOOLEAN
            ) {
                codeAttribute.writeOpcode(Opcode.IAND + opIdx);
                return IClass.BOOLEAN;
            }

            locatable.throwCompileException("Operator \"" + op + "\" not defined on types \"" + lhsType + "\" and \"" + rhsType + "\"");
            /* NEVER REACHED */ return null;
        }

        if (
            op == "*"   ||
            op == "/"   ||
            op == "%"   ||
            op == "+"   ||
            op == "-"   ||
            op == "<<"  ||
            op == ">>"  ||
            op == ">>>"
        ) {

            // String concatenation?
            if (op == "+" && (
                lhsType == IClass.STRING ||
                rhsType == IClass.STRING
            )) {

                // Convert LHS to "String".
                codeAttribute.pushInserter(convertLhsInserter); {
                    Java.stringConversion(codeAttribute, locatable, lhsType);
                } codeAttribute.popInserter();

                // Convert RHS to "String".
                Java.stringConversion(codeAttribute, locatable, rhsType);

                // Concatenate.
                codeAttribute.writeOpcode(Opcode.INVOKEVIRTUAL);
                codeAttribute.writeShort(codeAttribute.getClassFile().addConstantMethodrefInfo(
                    Descriptor.STRING,                                // classFD
                    "concat",                                         // methodName
                    "(" + Descriptor.STRING + ")" + Descriptor.STRING // methodMD
                ));

                return IClass.STRING;
            }

            IClass promotedType = binaryNumericPromotion(codeAttribute, locatable, lhsType, convertLhsInserter, rhsType);
            if ((
                promotedType == IClass.FLOAT  ||
                promotedType == IClass.DOUBLE
            ) && (
                op == "<<"  ||
                op == ">>"  ||
                op == ">>>"
            )) locatable.throwCompileException("Operator \"" + op + "\" not applicable to operands of type \"" + promotedType + "\"");

            int opcode = (
                op == "*"   ? Opcode.IMUL  :
                op == "/"   ? Opcode.IDIV  :
                op == "%"   ? Opcode.IREM  :
                op == "+"   ? Opcode.IADD  :
                op == "-"   ? Opcode.ISUB  :
                op == "<<"  ? Opcode.ISHL  :
                op == ">>"  ? Opcode.ISHR  :
                op == ">>>" ? Opcode.IUSHR : Integer.MAX_VALUE
            );
            if (promotedType == IClass.INT) {
                ;
            } else
            if (promotedType == IClass.LONG) {
                ++opcode;
            } else
            if (promotedType == IClass.FLOAT) {
                opcode += 2;
            } else
            if (promotedType == IClass.DOUBLE) {
                opcode += 3;
            } else
            {
                locatable.throwCompileException("Unexpected promoted type \"" + promotedType + "\"");
            }
            codeAttribute.writeOpcode(opcode);
            return promotedType;
        }

        locatable.throwCompileException("Unexpected operator \"" + op + "\"");
        /* NEVER REACHED */ return null;
    }

    /*package*/ static class Cast extends Rvalue {
        public Cast(
            Scanner.Location location,
            Type             targetType,
            Rvalue           value
        ) {
            super(location);
            this.targetType = targetType;
            this.value      = value;
        }

        private final Type   targetType;
        private final Rvalue value;

        // Compile time members.

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.targetType.getType();
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            IClass tt = targetType.getType();
            IClass vt = this.value.compileGetValue(codeAttribute);
            if (
                !Java.tryIdentityConversion(vt, tt) &&
                !Java.tryWideningPrimitiveConversion(codeAttribute, (Locatable) this, vt, tt) &&
                !Java.tryNarrowingPrimitiveConversion(codeAttribute, (Locatable) this, vt, tt) &&
                !Java.tryWideningReferenceConversion(vt, tt) &&
                !Java.tryNarrowingReferenceConversion(codeAttribute, vt, tt)
            ) this.throwCompileException("Cannot cast \"" + vt + "\" to \"" + tt + "\"");
            return tt;
        }
    }

    /*package*/ static class ConstructorInvocation extends Invocation {
        public ConstructorInvocation(
            Scanner.Location      location,
            ClassDeclaration      declaringClass,
            ConstructorDeclarator declaringConstructor,
            boolean               targetIsSuperclass,
            Rvalue[]              arguments
        ) {
            super(location, (Scope) declaringConstructor, arguments);
            this.declaringClass     = declaringClass;
            this.targetIsSuperclass = targetIsSuperclass;
        }

        // Implement "Atom".
        public IClass getType() throws CompileException {
            this.throwCompileException("A constructor invocation has no type");
            /* NEVER REACHED */ return null;
        }

        // Implement "Rvalue".
        public void compile(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            codeAttribute.writeOpcode(Opcode.ALOAD_0);
            Java.invokeConstructor(
                (Locatable) this, // locatable
                this.scope,       // scope
                (                 // targetClass
                    this.targetIsSuperclass ?
                    this.declaringClass.getSuperclass() :
                    this.declaringClass
                ),
                this.arguments,   // arguments
                codeAttribute     // codeAttribute
            );
        }
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            this.throwCompileException("A constructor invocation is not an rvalue");
            /* NEVER REACHED */ return null;
        }

        private final ClassDeclaration declaringClass;
        private final boolean          targetIsSuperclass;
    }

    /**
     * Expects the object to initialize on the stack.
     */
    private static void invokeConstructor(
        Locatable               locatable,
        Scope                   scope,
        IClass                  targetClass,
        Rvalue[]                arguments,
        ClassFile.CodeAttribute codeAttribute
    ) throws CompileException {

        // Find constructors.
        IClass.IConstructor[] iConstructors = targetClass.getDeclaredIConstructors();
        IClass.IConstructor iConstructor;
        if (iConstructors.length == 0) {
            iConstructor = targetClass.new IConstructor() {
                public int getAccess() { return IClass.PUBLIC; }
                public IClass[] getParameterTypes() { return new IClass[0]; }
                public IClass[] getThrownExceptions() { return new IClass[0]; }
            };
        } else {
            iConstructor = (IClass.IConstructor) Java.findIInvocable(
                locatable,     // locatable
                iConstructors, // iInvocables
                arguments      // arguments
            );
        }

        // Check exceptions that the constructor may throw.
        IClass[] thrownExceptions = iConstructor.getThrownExceptions();
        for (int i = 0; i < thrownExceptions.length; ++i) {
            Java.checkThrownException(
                locatable,
                thrownExceptions[i],
                scope
            );
        }

        // Evaluate constructor arguments.
        IClass[] parameterTypes = iConstructor.getParameterTypes();
        for (int i = 0; i < arguments.length; ++i) {
            Java.assignmentConversion(
                codeAttribute,                               // codeAttribute
                locatable,                                   // locatable
                arguments[i].compileGetValue(codeAttribute), // sourceType
                parameterTypes[i]                            // targetType
            );
        }

        // Invoke!
        codeAttribute.writeOpcode(Opcode.INVOKESPECIAL);
        codeAttribute.writeShort(codeAttribute.getClassFile().addConstantMethodrefInfo(
            targetClass.getDescriptor(),  // classFD
            "<init>",                     // methodName
            iConstructor.getDescriptor()  // methodMD
        ));
    }

    /*package*/ static class MethodInvocation extends Invocation {
        public MethodInvocation(
            Scanner.Location location,
            Scope            enclosingScope,
            Atom             optionalTarget,
            String           methodName,
            Rvalue[]         arguments
        ) {
            super(location, enclosingScope, arguments);
            this.optionalTarget = optionalTarget;
            this.methodName     = methodName;
        }

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.findIMethod().getReturnType();
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            IClass.IMethod iMethod = this.findIMethod();
            IClass         lhsType;
            if (optionalTarget == null) {

                // 6.5.7.1
                FunctionDeclarator          scopeFunction = null;
                ClassOrInterfaceDeclaration scopeClassOrInterface = null;
                for (Scope s = this.scope; s != null; s = s.getEnclosingScope()) {
                    if (s instanceof FunctionDeclarator) {
                        scopeFunction = (FunctionDeclarator) s;
                    }
                    if (s instanceof ClassOrInterfaceDeclaration) {
                        scopeClassOrInterface = (ClassOrInterfaceDeclaration) s;
                        break;
                    }
                }
                if (!iMethod.isStatic()) {
                    if ((scopeFunction.modifiers & Mod.STATIC) != 0) this.throwCompileException("Instance method \"" + methodName + "\" cannot be invoked in static context");
                    Java.load(codeAttribute, scopeClassOrInterface, 0);
                }
            } else {

                // 6.5.7.2
                boolean staticContext = this.optionalTarget.isType();
                try {
                    lhsType = (
                        staticContext ?
                        this.optionalTarget.getType() :
                        this.optionalTarget.toRvalue().compileGetValue(codeAttribute)
                    );
                } catch (Parser.ParseException ex) {
                    this.throwCompileException(ex.getMessage());
                    /* NEVER REACHED */ return null;
                }
                if (iMethod.isStatic()) {
                    if (!staticContext) this.throwCompileException("Class method \"" + methodName + "\" cannot be invoked in non-static context");
                } else {
                    if (staticContext) this.throwCompileException("Instance method \"" + methodName + "\" cannot be invoked in static context");
                }
            }

            // Evaluate method parameters.
            IClass[] parameterTypes = iMethod.getParameterTypes();
            for (int i = 0; i < this.arguments.length; ++i) {
                Java.assignmentConversion(
                    codeAttribute,
                    (Locatable) this,
                    this.arguments[i].compileGetValue(codeAttribute),
                    parameterTypes[i]
                );
            }

            // Invoke!
            short cmrf = codeAttribute.getClassFile().addConstantMethodrefInfo(
                iMethod.getDeclaringIClass().getDescriptor(), // classFD
                iMethod.getName(),                            // methodName
                iMethod.getDescriptor()                       // methodMD
            );
            if (iMethod.isStatic()) {
                codeAttribute.writeOpcode(Opcode.INVOKESTATIC);
                codeAttribute.writeShort(cmrf);
            } else {
                codeAttribute.writeOpcode(
                    iMethod.getAccess() == IClass.PRIVATE ?
                    Opcode.INVOKESPECIAL :
                    Opcode.INVOKEVIRTUAL
                );
                codeAttribute.writeShort(cmrf);
            }
            return iMethod.getReturnType();
        }

        /**
         * Find named methods of "targetType", evaluate the arguments, choose the
         * most specific method, and convert the arguments to the invocable's
         * parameter types. Check that only the allowed exceptions are thrown.
         *
         * @return The selected {@link IMethod}
         */
        private IClass.IMethod findIMethod() throws CompileException {
            ClassOrInterfaceDeclaration scopeClassOrInterface = null;
            {
                Scope s = this.scope;
                while (s instanceof Statement) s = s.getEnclosingScope();
                if (s instanceof FunctionDeclarator) {
                    s = s.getEnclosingScope();
                }
                if (s instanceof ClassOrInterfaceDeclaration) {
                    scopeClassOrInterface = (ClassOrInterfaceDeclaration) s;
                }
            }

            // Find methods with specified name.
            IClass.IMethod iMethod = Java.findIMethod(
                (Locatable) this,
                (                             // targetType
                    this.optionalTarget == null ?
                    scopeClassOrInterface :
                    this.optionalTarget.getType()
                ),
                this.methodName,              // methodName
                this.arguments                // arguments
            );

            // Check exceptions that the method may throw.
            IClass[] thrownExceptions = iMethod.getThrownExceptions();
            for (int i = 0; i < thrownExceptions.length; ++i) {
                Java.checkThrownException(
                    (Locatable) this,    // locatable
                    thrownExceptions[i], // type
                    scope                // scope
                );
            }

            return iMethod;
        }



        private final Atom   optionalTarget; // null == simple method name.
        private final String methodName;
    }

    /*package*/ static class SuperclassMethodInvocation extends Invocation {
        public SuperclassMethodInvocation(
            Scanner.Location location,
            Scope            enclosingScope,
            String           methodName,
            Rvalue[]         arguments
        ) {
            super(location, enclosingScope, arguments);
            this.methodName = methodName;
        }

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.findIMethod().getReturnType();
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            IClass.IMethod iMethod = this.findIMethod();

            Scope s;
            for (s = this.scope; s instanceof Statement; s = s.getEnclosingScope());
            FunctionDeclarator fd = s instanceof FunctionDeclarator ? (FunctionDeclarator) s : null;
            if (fd == null) this.throwCompileException("Cannot invoke superclass method in non-method scope");
            if ((fd.modifiers & Mod.STATIC) != 0) this.throwCompileException("Cannot invoke superclass method in static context");
            Java.load(codeAttribute, fd.declaringClassOrInterface, 0);

            // Evaluate method parameters.
            IClass[] parameterTypes = iMethod.getParameterTypes();
            for (int i = 0; i < this.arguments.length; ++i) {
                Java.assignmentConversion(
                    codeAttribute, // codeAttribute
                    (Locatable) this, // locatable
                    this.arguments[i].compileGetValue(codeAttribute), // sourceType
                    parameterTypes[i]                                 // targetType
                );
            }

            // Invoke!
            codeAttribute.writeOpcode(Opcode.INVOKESPECIAL);
            codeAttribute.writeShort(codeAttribute.getClassFile().addConstantMethodrefInfo(
                iMethod.getDeclaringIClass().getDescriptor(), // classFD
                methodName,                                   // methodName
                iMethod.getDescriptor()                       // methodMD
            ));
            return iMethod.getReturnType();
        }

        private IClass.IMethod findIMethod() throws CompileException {
            ClassDeclaration declaringClass;
            for (Scope s = this.scope;; s = s.getEnclosingScope()) {
                if (s instanceof FunctionDeclarator) {
                    FunctionDeclarator fd = (FunctionDeclarator) s;
                    if ((fd.modifiers & Mod.STATIC) != 0) this.throwCompileException("Superclass method cannot be invoked in static context");
                }
                if (s instanceof ClassDeclaration) {
                    declaringClass = (ClassDeclaration) s;
                    break;
                }
            }
            return Java.findIMethod(
                (Locatable) this, // locatable
                declaringClass.getSuperclass(), // targetType
                this.methodName,  // methodName
                this.arguments    // arguments
            );
        }

        private final String methodName;
    }

    /*package*/ static abstract class Invocation extends Rvalue {
        public Invocation(
            Scanner.Location location,
            Scope            scope,
            Rvalue[]         arguments
        ) {
            super(location);
            this.scope     = scope;
            this.arguments = arguments;
        }

        protected final Scope    scope;
        protected final Rvalue[] arguments;
    }

    /**
     * Find {@link IMethod} by name.
     */
    private static IClass.IMethod findIMethod(
        Locatable locatable,
        IClass    targetType,
        String    methodName,
        Rvalue[]  arguments
    ) throws CompileException {
        Vector v = new Vector();
        for (IClass c = targetType; c != null; c = c.getSuperclass()) {
            IClass.IMethod[] ims = c.getDeclaredIMethods();
            for (int i = 0; i < ims.length; ++i) {
                IClass.IMethod im = ims[i];
                if (im.getName().equals(methodName)) v.add(im);
            }
        }
        if (v.size() == 0) locatable.throwCompileException("Class \"" + targetType + "\" has no method named \"" + methodName + "\"");

        // Determine arguments' types, choose the most specific method
        IClass.IMethod iMethod = (IClass.IMethod) Java.findIInvocable(
            locatable,        // locatable
            (IClass.IMethod[]) v.toArray(new IClass.IMethod[v.size()]), // iInvocables
            arguments         // arguments
        );
        return iMethod;
    }

    /**
     * Determine the arguments' types and choose the most specific invocable.
     *
     * @return The selected {@link IInvocable}
     */
    private static IClass.IInvocable findIInvocable(
        Locatable           locatable,
        IClass.IInvocable[] iInvocables,
        Rvalue[]            arguments
    ) throws CompileException {

        // Determine arguments' types.
        IClass[] argumentTypes = new IClass[arguments.length];
        for (int i = 0; i < arguments.length; ++i) {
            argumentTypes[i] = arguments[i].getType();
        }

        // Select applicable methods (15.12.2.1).
        Vector applicableIInvocables = new Vector();
        nextMethod: for (int i = 0; i < iInvocables.length; ++i) {
            IClass.IInvocable ii = iInvocables[i];

            // Check parameter count.
            IClass[] parameterTypes = ii.getParameterTypes();
            if (parameterTypes.length != arguments.length) continue;

            // Check argument types vs. parameter types.
            for (int j = 0; j < arguments.length; ++j) {
                if (!parameterTypes[j].isAssignableFrom(argumentTypes[j])) continue nextMethod;
            }

            // Applicable!
            applicableIInvocables.add(ii);
        }
        if (applicableIInvocables.size() == 0) {
            StringBuffer sb = new StringBuffer();
            for (int i = 0; i < iInvocables.length; ++i) {
                if (i > 0) sb.append(", ");
                sb.append("\"" + iInvocables[i] + "\"");
            }
            locatable.throwCompileException("No applicable constructor/method found; candidates are: " + sb.toString());
        }

        // Choose the most specific invocable (15.12.2.2).
        IClass.IInvocable iInvocable;
        if (applicableIInvocables.size() == 1) {
            iInvocable = (IClass.IInvocable) applicableIInvocables.get(0);
        } else {
            Vector mostSpecificIInvocables = new Vector();
            for (int i = 0; i < applicableIInvocables.size(); ++i) {
                IClass.IInvocable applicableIInvocable = (IClass.IInvocable) applicableIInvocables.elementAt(i);
                int moreSpecific = 0, lessSpecific = 0;
                for (int j = 0; j < mostSpecificIInvocables.size(); ++j) {
                    IClass.IInvocable mostSpecificIInvocable = (IClass.IInvocable) mostSpecificIInvocables.elementAt(j);
                    if (applicableIInvocable.isMoreSpecificThan(mostSpecificIInvocable)) {
                        ++moreSpecific;
                    } else
                    if (applicableIInvocable.isLessSpecificThan(mostSpecificIInvocable)) {
                        ++lessSpecific;
                    }
                }
                if (moreSpecific == mostSpecificIInvocables.size()) {
                    mostSpecificIInvocables.clear();
                    mostSpecificIInvocables.add(applicableIInvocable);
                } else
                if (lessSpecific < mostSpecificIInvocables.size()) {
                    mostSpecificIInvocables.add(applicableIInvocable);
                } else
                {
                }
            }
            if (mostSpecificIInvocables.size() > 1) {
                StringBuffer sb = new StringBuffer("Invocation of constructor/method \"(");
                for (int i = 0; i < arguments.length; ++i) {
                    if (i > 0) sb.append(", ");
                    sb.append(Descriptor.toString(argumentTypes[i].getDescriptor()));
                }
                sb.append(")\" is ambiguous: ");
                for (int i = 0; i < mostSpecificIInvocables.size(); ++i) {
                    if (i > 0) sb.append(" vs. ");
                    sb.append("\"" + mostSpecificIInvocables.elementAt(i) + "\"");
                }
                locatable.throwCompileException(sb.toString());
            }
            iInvocable = (IClass.IInvocable) mostSpecificIInvocables.elementAt(0);
        }
        return iInvocable;
    }

    /*package*/ static class NewClassInstance extends Rvalue {
        public NewClassInstance(
            Scanner.Location location,
            Scope            scope,
            ReferenceType    referenceType,
            Rvalue[]         arguments
        ) {
            super(location);
            this.scope         = scope;
            this.referenceType = referenceType;
            this.arguments     = arguments;
        }

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.referenceType.getType();
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            ClassFile cf = codeAttribute.getClassFile();
            IClass clazz = this.referenceType.getType();

            codeAttribute.writeOpcode(Opcode.NEW);
            codeAttribute.writeShort(cf.addConstantClassInfo(
                clazz.getDescriptor() // typeFD
            ));
            codeAttribute.writeOpcode(Opcode.DUP);
            invokeConstructor(
                (Locatable) this, // locatable
                this.scope,       // scope
                clazz,            // targetClass
                this.arguments,   // arguments
                codeAttribute     // codeAttribute
            );
            return clazz;
        }

        private final Scope         scope;
        private final ReferenceType referenceType;
        private final Rvalue[]      arguments;
    }

    /*package*/ static class NewArray extends Rvalue {
        public NewArray(
            Scanner.Location location,
            Type             type,
            Rvalue[]         dimExprs,
            int              dims
        ) {
            super(location);
            this.type     = type;
            this.dimExprs = dimExprs;
            this.dims     = dims;
        }

        // Implement "Atom".
        public IClass getType() throws CompileException {
            IClass res = this.type.getType();
            for (int i = 0; i < this.dimExprs.length; ++i) res = res.getArrayType();
            for (int i = 0; i < this.dims; ++i) res = res.getArrayType();
            return res;
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            for (int i = 0; i < this.dimExprs.length; ++i) {
                IClass dimType = this.dimExprs[i].compileGetValue(codeAttribute);
                if (dimType != IClass.INT && unaryNumericPromotion(
                    codeAttribute,    // codeAttribute
                    (Locatable) this, // locatable
                    dimType           // type
                ) != IClass.INT) this.throwCompileException("Invalid array size expression type");
            }

            return Java.newArray(
                codeAttribute,        // codeAttribute
                this.dimExprs.length, // dimExprCount
                this.dims,            // dims
                this.type.getType()   // componentType
            );
        }

        private final Type     type;
        private final Rvalue[] dimExprs;
        private final int      dims;
    }

    /**
     * Expects "dimExprCount" values of type "integer" on the operand stack.
     * Creates an array of "dimExprCount" + "dims" dimensions of
     * "componentType".
     *
     * @return The type of the created array
     */
    private static IClass newArray(
        ClassFile.CodeAttribute codeAttribute,
        int                     dimExprCount,
        int                     dims,
        IClass                  componentType
    ) {
        if (dimExprCount == 1 && dims == 0 && componentType.isPrimitive()) {

            // "new <primitive>[<n>]"
            codeAttribute.writeOpcode(Opcode.NEWARRAY);
            codeAttribute.writeByte(
                componentType == IClass.BOOLEAN ? 4 :
                componentType == IClass.CHAR    ? 5 :
                componentType == IClass.FLOAT   ? 6 :
                componentType == IClass.DOUBLE  ? 7 :
                componentType == IClass.BYTE    ? 8 :
                componentType == IClass.SHORT   ? 9 :
                componentType == IClass.INT     ? 10 :
                componentType == IClass.LONG    ? 11 : -1
            );
            return componentType.getArrayType();
        }

        IClass at = componentType;
        for (int i = 0; i < dimExprCount; ++i) at = at.getArrayType();
        for (int i = 0; i < dims; ++i) at = at.getArrayType();
        ClassFile cf = codeAttribute.getClassFile();
        if (dimExprCount == 1) {

            // "new <primitive>[<n>][]..."
            // "new <class-or-interface>[<n>]..."
            codeAttribute.writeOpcode(Opcode.ANEWARRAY);
            codeAttribute.writeShort(cf.addConstantClassInfo(
                componentType.getDescriptor() // typeFD
            ));
        } else {

            // "new <anything>[<n>][<m>]..."
            codeAttribute.writeOpcode(Opcode.MULTIANEWARRAY);
            codeAttribute.writeShort(cf.addConstantClassInfo(
                at.getDescriptor() // typeFD
            ));
            codeAttribute.writeByte(dimExprCount);
        }
        return at;
    }

    /**
     * Represents a Java<sup>TM</sup> array initializer (JLS 10.6).
     * <p>
     * Allocates an array and initializes its members with (not necessarily
     * constant) values.
     */
    /*package*/ static class ArrayInitializer extends Rvalue {
        public ArrayInitializer(
            Scanner.Location location,
            ArrayType        arrayType,
            Rvalue[]         values
        ) {
            super(location);
            this.arrayType = arrayType;
            this.values    = values;
        }

        // Implement "Atom".
        public IClass getType() throws CompileException {
            return this.arrayType.getType();
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            IClass at = this.arrayType.getType();
            IClass ct = at.getComponentType();

            Java.pushConstant(
                codeAttribute,
                new Integer(this.values.length)
            );
            Java.newArray(
                codeAttribute, // codeAttribute,
                1,             // dimExprCount,
                0,             // dims,
                ct             // componentType
            );

            for (int i = 0; i < values.length; ++i) {
                Rvalue v = values[i];
                codeAttribute.writeOpcode(Opcode.DUP);
                Java.pushConstant(codeAttribute, new Integer(i));
                IClass vt = v.compileGetValue(codeAttribute);
                Java.assignmentConversion(
                    codeAttribute, // codeAttribute
                    (Locatable) v, // location
                    vt,            // sourceType
                    ct             // targetType
                );
                codeAttribute.writeOpcode(Opcode.IASTORE + Java.ilfdabcs(ct));
            }
            return at;
        }

        private final ArrayType arrayType;
        private final Rvalue[]  values;
    }

    /*package*/ static class Literal extends Rvalue {
        public Literal(Scanner.Location location, Object value) {
            super(location);
            this.value = value;
        }

        // Implement "Atom."
        public IClass getType() {
            if (this.value instanceof Integer  ) return IClass.INT;
            if (this.value instanceof Long     ) return IClass.LONG;
            if (this.value instanceof Float    ) return IClass.FLOAT;
            if (this.value instanceof Double   ) return IClass.DOUBLE;
            if (this.value instanceof String   ) return IClass.STRING;
            if (this.value instanceof Character) return IClass.CHAR;
            if (this.value instanceof Boolean  ) return IClass.BOOLEAN;
            if (this.value == null             ) return IClass.VOID;
            throw new RuntimeException();
        }

        // Implement "Rvalue".
        public IClass compileGet(ClassFile.CodeAttribute codeAttribute) throws CompileException {
            return Java.pushConstant(codeAttribute, this.value);
        }

        public Object getConstantValue() { return value; }
        private final Object value;
    }

    private static IClass pushConstant(
        ClassFile.CodeAttribute codeAttribute,
        Object                  value
    ) {
        ClassFile cf = codeAttribute.getClassFile();
        if (
            value instanceof Integer   ||
            value instanceof Short     ||
            value instanceof Character ||
            value instanceof Byte
        ) {
            int i = (
                value instanceof Character ?
                ((Character) value).charValue() :
                ((Number) value).intValue()
            );
            if (i >= -1 && i <= 5) {
                codeAttribute.writeOpcode(Opcode.ICONST_0 + i);
            } else
            if (i >= Byte.MIN_VALUE && i <= Byte.MAX_VALUE) {
                codeAttribute.writeOpcode(Opcode.BIPUSH);
                codeAttribute.writeByte((byte) i);
            } else {
                codeAttribute.writeOpcode(Opcode.LDC_W);
                codeAttribute.writeShort(cf.addConstantIntegerInfo(i));
            }
            return IClass.INT;
        }
        if (value instanceof Long) {
            long lv = ((Long) value).longValue();
            if (lv >= 0L && lv <= 1L) {
                codeAttribute.writeOpcode(Opcode.LCONST_0 + (int) lv);
            } else {
                codeAttribute.writeOpcode(Opcode.LDC2_W);
                codeAttribute.writeShort(cf.addConstantLongInfo(lv));
            }
            return IClass.LONG;
        }
        if (value instanceof Float) {
            float fv = ((Float) value).floatValue();
            if (fv == 0.0F || fv == 1.0F || fv == 2.0F) {
                codeAttribute.writeOpcode(Opcode.FCONST_0 + (int) fv);
            } else {
                codeAttribute.writeOpcode(Opcode.LDC_W);
                codeAttribute.writeShort(cf.addConstantFloatInfo(fv));
            }
            return IClass.FLOAT;
        }
        if (value instanceof Double) {
            double dv = ((Double) value).doubleValue();
            if (dv == 0.0 || dv == 1.0) {
                codeAttribute.writeOpcode(Opcode.DCONST_0 + (int) dv);
            } else {
                codeAttribute.writeOpcode(Opcode.LDC2_W);
                codeAttribute.writeShort(cf.addConstantDoubleInfo(dv));
            }
            return IClass.DOUBLE;
        }
        if (value instanceof String) {
            short index = cf.addConstantStringInfo((String) value);
            if (index <= 255) {
                codeAttribute.writeOpcode(Opcode.LDC);
                codeAttribute.writeByte(index);
            } else {
                codeAttribute.writeOpcode(Opcode.LDC_W);
                codeAttribute.writeShort(index);
            }
            return IClass.STRING;
        }
        if (value instanceof Boolean) {
            codeAttribute.writeOpcode(((Boolean) value).booleanValue() ? Opcode.ICONST_1 : Opcode.ICONST_0);
            return IClass.BOOLEAN;
        }
        if (value == null) { // Null literal?
            codeAttribute.writeOpcode(Opcode.ACONST_NULL);
            return IClass.VOID;
        }
        throw new RuntimeException("Unknown literal type \"" + value.getClass().getName() + "\"");
    }

    private static class LocalVariable {
        public LocalVariable(
            boolean finaL,
            IClass  type,
            short   localVariableArrayIndex
        ) {
            this.finaL                   = finaL;
            this.type                    = type;
            this.localVariableArrayIndex = localVariableArrayIndex;
        }
        private final boolean finaL;
        private final IClass  type;
        private final short   localVariableArrayIndex;
    }

    /**
     * Implements "assignment conversion" (5.2).
     */
    private static void assignmentConversion(
        ClassFile.CodeAttribute codeAttribute,
        Locatable               locatable,
        IClass                  sourceType,
        IClass                  targetType
    ) throws CompileException {

        // 5.2 Identity conversion.
        if (Java.tryIdentityConversion(sourceType, targetType)) return;

        // 5.2 Widening primitive conversion.
        if (Java.tryWideningPrimitiveConversion(codeAttribute, locatable, sourceType, targetType)) return;

        // 5.2 Widening reference conversion.
        if (Java.tryWideningReferenceConversion(sourceType, targetType)) return;

        // 5.2 Special narrowing primitive conversion.
        // TBD

        // TDB...
        locatable.throwCompileException("Assignment conversion not possible from type \"" + sourceType + "\" to type \"" + targetType + "\"");
    }

    /**
     * Implements "unary numeric promotion" (5.6.1)
     *
     * @return The promoted type.
     */
    private static IClass unaryNumericPromotion(
        ClassFile.CodeAttribute codeAttribute,
        Locatable               locatable,
        IClass                  type
    ) throws CompileException {
        IClass promotedType = unaryNumericPromotion(locatable, type);

        if (
            !Java.tryIdentityConversion(type, promotedType) &&
            !Java.tryWideningPrimitiveConversion(
                codeAttribute, // codeAttribute
                locatable,     // locatable
                type,          // sourceType
                promotedType   // targetType
            )
        ) throw new RuntimeException();
        return promotedType;
    }

    private static IClass unaryNumericPromotion(
        Locatable locatable,
        IClass    type
    ) throws CompileException {
        if (!type.isPrimitiveNumeric()) locatable.throwCompileException("Unary numeric promotion not possible on non-numeric-primitive type \"" + type + "\"");

        return (
            type == IClass.DOUBLE ? IClass.DOUBLE :
            type == IClass.FLOAT  ? IClass.FLOAT  :
            type == IClass.LONG   ? IClass.LONG   :
            IClass.INT
        );
    }

    /**
     * Implements "binary numeric promotion" (5.6.2)
     *
     * @return The promoted type.
     */
    private static IClass binaryNumericPromotion(
        ClassFile.CodeAttribute          codeAttribute,
        Locatable                        locatable,
        IClass                           type1,
        ClassFile.CodeAttribute.Inserter convertInserter1,
        IClass                           type2
    ) throws CompileException {
        IClass promotedType = binaryNumericPromotion(locatable, type1, type2);

        codeAttribute.pushInserter(convertInserter1); {
            if (
                !Java.tryIdentityConversion(type1, promotedType) &&
                !Java.tryWideningPrimitiveConversion(
                    codeAttribute, // codeAttribute
                    locatable,     // locatable
                    type1,         // sourceType
                    promotedType   // targetType
                )
            ) throw new RuntimeException();
        } codeAttribute.popInserter();

        if (
            !Java.tryIdentityConversion(type2, promotedType) &&
            !Java.tryWideningPrimitiveConversion(
                codeAttribute, // codeAttribute
                locatable,     // locatable
                type2,         // sourceType
                promotedType   // targetType
            )
        ) throw new RuntimeException();

        return promotedType;
    }

    private static IClass binaryNumericPromotion(
        Locatable locatable,
        IClass    type1,
        IClass    type2
    ) throws CompileException {
        if (
            !type1.isPrimitiveNumeric() ||
            !type2.isPrimitiveNumeric()
        ) locatable.throwCompileException("Binary numeric promotion not possible on types \"" + type1 + "\" and \"" + type2 + "\"");

        return (
            type1 == IClass.DOUBLE || type2 == IClass.DOUBLE ? IClass.DOUBLE :
            type1 == IClass.FLOAT  || type2 == IClass.FLOAT  ? IClass.FLOAT  :
            type1 == IClass.LONG   || type2 == IClass.LONG   ? IClass.LONG   :
            IClass.INT
        );
    }

    /**
     * Implements "identity conversion" (5.1.1).
     *
     * @return Whether the conversion succeeded
     */
    /*package*/ static boolean tryIdentityConversion(
        IClass sourceType,
        IClass targetType
    ) throws CompileException {
        return sourceType == targetType;
    }

    /**
     * Implements "widening primitive conversion" (5.1.2).
     *
     * @return Whether the conversion succeeded
     */
    /*package*/ static boolean tryWideningPrimitiveConversion(
        ClassFile.CodeAttribute codeAttribute,
        Locatable               locatable,
        IClass                  sourceType,
        IClass                  targetType
    ) throws CompileException {
        byte[] opcodes = (byte[]) PRIMITIVE_WIDENING_CONVERSIONS.get(sourceType.getDescriptor() + targetType.getDescriptor());
        if (opcodes != null) {
            codeAttribute.write(opcodes);
            return true;
        }
        return false;
    }
    private static final HashMap PRIMITIVE_WIDENING_CONVERSIONS = new HashMap();
    static { fillConversionMap(new Object[] {
        new byte[0],
        Descriptor.BYTE  + Descriptor.SHORT,
        Descriptor.BYTE  + Descriptor.INT,
        Descriptor.SHORT + Descriptor.INT,
        Descriptor.CHAR  + Descriptor.INT,

        new byte[] { Opcode.I2L },
        Descriptor.BYTE  + Descriptor.LONG,
        Descriptor.SHORT + Descriptor.LONG,
        Descriptor.CHAR  + Descriptor.LONG,
        Descriptor.INT   + Descriptor.LONG,

        new byte[] { Opcode.I2F },
        Descriptor.BYTE  + Descriptor.FLOAT,
        Descriptor.SHORT + Descriptor.FLOAT,
        Descriptor.CHAR  + Descriptor.FLOAT,
        Descriptor.INT   + Descriptor.FLOAT,

        new byte[] { Opcode.L2F },
        Descriptor.LONG  + Descriptor.FLOAT,

        new byte[] { Opcode.I2D },
        Descriptor.BYTE  + Descriptor.DOUBLE,
        Descriptor.SHORT + Descriptor.DOUBLE,
        Descriptor.CHAR  + Descriptor.DOUBLE,
        Descriptor.INT   + Descriptor.DOUBLE,

        new byte[] { Opcode.L2D },
        Descriptor.LONG  + Descriptor.DOUBLE,

        new byte[] { Opcode.F2D },
        Descriptor.FLOAT + Descriptor.DOUBLE,
    }, PRIMITIVE_WIDENING_CONVERSIONS); }
    private static void fillConversionMap(Object[] array, HashMap map) {
        byte[] opcodes = null;
        for (int i= 0; i < array.length; ++i) {
            Object o = array[i];
            if (o instanceof byte[]) {
                opcodes = (byte[]) o;
            } else {
                map.put(o, opcodes);
            }
        }
    }

    /**
     * Implements "widening reference conversion" (5.1.4).
     *
     * @return Whether the conversion succeeded
     */
    private static boolean tryWideningReferenceConversion(
        IClass sourceType,
        IClass targetType
    ) throws CompileException {
        if (
            targetType.isPrimitive() ||
            sourceType == targetType
        ) return false;

        // 5.1.4.1 Target type is superclass of source class type.
        // 5.1.4.2 Source class type implements target interface type.
        // 5.1.4.4 Source interface type implements target interface type.
        if (targetType.isAssignableFrom(sourceType)) return true;

        // 5.1.4.3 Convert "null" literal to any reference type.
        if (sourceType == IClass.VOID && !targetType.isPrimitive()) return true;

        // 5.1.4.5 From any interface to type "Object".
        // TBD

        // 5.1.4.6 From any array type to type "Object".
        // TBD

        // 5.1.4.7 From any array type to type "Cloneable".
        // TBD

        // 5.1.4.8 From any array type to type "java.io.Serializable".
        // TBD

        return false;
    }

    /**
     * Implements "narrowing primitive conversion" (5.1.3).
     *
     * @return Whether the conversion succeeded
     */
    private static boolean tryNarrowingPrimitiveConversion(
        ClassFile.CodeAttribute codeAttribute,
        Locatable               locatable,
        IClass                  sourceType,
        IClass                  targetType
    ) throws CompileException {
        byte[] opcodes = (byte[]) PRIMITIVE_NARROWING_CONVERSIONS.get(sourceType.getDescriptor() + targetType.getDescriptor());
        if (opcodes != null) {
            codeAttribute.write(opcodes);
            return true;
        }
        return false;
    }
    private static final HashMap PRIMITIVE_NARROWING_CONVERSIONS = new HashMap();
    static { fillConversionMap(new Object[] {
        new byte[0],
        Descriptor.BYTE + Descriptor.CHAR,
        Descriptor.SHORT + Descriptor.CHAR,
        Descriptor.CHAR + Descriptor.SHORT,

        new byte[] { Opcode.I2B },
        Descriptor.SHORT + Descriptor.BYTE,
        Descriptor.CHAR + Descriptor.BYTE,
        Descriptor.INT + Descriptor.BYTE,

        new byte[] { Opcode.I2S },
        Descriptor.INT + Descriptor.SHORT,
        Descriptor.INT + Descriptor.CHAR,

        new byte[] { Opcode.L2I, Opcode.I2B },
        Descriptor.LONG + Descriptor.BYTE,

        new byte[] { Opcode.L2I, Opcode.I2S },
        Descriptor.LONG + Descriptor.SHORT,
        Descriptor.LONG + Descriptor.CHAR,

        new byte[] { Opcode.L2I },
        Descriptor.LONG + Descriptor.INT,

        new byte[] { Opcode.F2I, Opcode.I2B },
        Descriptor.FLOAT + Descriptor.BYTE,

        new byte[] { Opcode.F2I, Opcode.I2S },
        Descriptor.FLOAT + Descriptor.SHORT,
        Descriptor.FLOAT + Descriptor.CHAR,

        new byte[] { Opcode.F2I },
        Descriptor.FLOAT + Descriptor.INT,

        new byte[] { Opcode.F2L },
        Descriptor.FLOAT + Descriptor.LONG,

        new byte[] { Opcode.D2I, Opcode.I2B },
        Descriptor.DOUBLE + Descriptor.BYTE,

        new byte[] { Opcode.D2I, Opcode.I2S },
        Descriptor.DOUBLE + Descriptor.SHORT,
        Descriptor.DOUBLE + Descriptor.CHAR,

        new byte[] { Opcode.D2I },
        Descriptor.DOUBLE + Descriptor.INT,

        new byte[] { Opcode.D2L },
        Descriptor.DOUBLE + Descriptor.LONG,

        new byte[] { Opcode.D2F },
        Descriptor.DOUBLE + Descriptor.FLOAT,
    }, PRIMITIVE_NARROWING_CONVERSIONS); }

    /**
     * Implements "narrowing reference conversion" (5.1.5).
     *
     * @return Whether the conversion succeeded
     */
    private static boolean tryNarrowingReferenceConversion(
        ClassFile.CodeAttribute codeAttribute,
        IClass                  sourceType,
        IClass                  targetType
    ) throws CompileException {
        if (
            sourceType.isPrimitive() ||
            sourceType == targetType ||
            !sourceType.isAssignableFrom(targetType)
        ) return false;
        codeAttribute.writeOpcode(Opcode.CHECKCAST);
        codeAttribute.writeShort(
            codeAttribute.getClassFile().addConstantClassInfo(
                targetType.getDescriptor()
            )
        );
        return true;
    }

    /**
     * Convert object of type "sourceType" to type "String".
     */
    private static void stringConversion(
        ClassFile.CodeAttribute codeAttribute,
        Locatable               locatable,
        IClass                  sourceType
    ) {
        if (sourceType == IClass.STRING) return;

        codeAttribute.writeOpcode(Opcode.INVOKESTATIC);
        codeAttribute.writeShort(
            codeAttribute.getClassFile().addConstantMethodrefInfo(
                Descriptor.STRING, // classFD
                "valueOf",         // methodName
                "(" + (            // methodMD
                    sourceType == IClass.BOOLEAN ||
                    sourceType == IClass.CHAR    ||
                    sourceType == IClass.LONG    ||
                    sourceType == IClass.FLOAT   ||
                    sourceType == IClass.DOUBLE ? sourceType.getDescriptor() :
                    sourceType == IClass.BYTE  ||
                    sourceType == IClass.SHORT ||
                    sourceType == IClass.INT ? Descriptor.INT :
                    Descriptor.OBJECT
                ) + ")" + Descriptor.STRING
            )
        );
    }

    // Load the value of a local variable onto the stack and return its type.
    private static IClass load(
        ClassFile.CodeAttribute codeAttribute,
        LocalVariable           localVariable
    ) {
        load(
            codeAttribute,
            localVariable.type,
            localVariable.localVariableArrayIndex
        );
        return localVariable.type;
    }
    private static void load(
        ClassFile.CodeAttribute codeAttribute,
        IClass                  type,
        int                     index
    ) {
        if (index <= 3) {
            codeAttribute.writeOpcode(Opcode.ILOAD_0 + 4 * Java.ilfda(type) + index);
        } else
        if (index <= 255) {
            codeAttribute.writeOpcode(Opcode.ILOAD + Java.ilfda(type));
            codeAttribute.writeByte(index);
        } else
        {
            codeAttribute.writeOpcode(Opcode.WIDE);
            codeAttribute.writeOpcode(Opcode.ILOAD + Java.ilfda(type));
            codeAttribute.writeShort(index);
        }
    }

    // Store the value on the stack in a local variable.
    private static void store(
        ClassFile.CodeAttribute codeAttribute,
        Locatable               locatable,
        IClass                  valueType,
        LocalVariable           localVariable
    ) throws CompileException {
        IClass lvType = localVariable.type;

        Java.assignmentConversion(codeAttribute, locatable, valueType, lvType);

        Java.store(codeAttribute, lvType, localVariable.localVariableArrayIndex);
    }
    private static void store(
        ClassFile.CodeAttribute codeAttribute,
        IClass                  lvType,
        short                   lvIndex
    ) {
        if (lvIndex <= 3) {
            codeAttribute.writeOpcode(Opcode.ISTORE_0 + 4 * Java.ilfda(lvType) + lvIndex);
        } else
        if (lvIndex <= 255) {
            codeAttribute.writeOpcode(Opcode.ISTORE + Java.ilfda(lvType));
            codeAttribute.writeByte(lvIndex);
        } else
        {
            codeAttribute.writeOpcode(Opcode.WIDE);
            codeAttribute.writeOpcode(Opcode.ISTORE + Java.ilfda(lvType));
            codeAttribute.writeShort(lvIndex);
        }
    }

    /*package*/ static void dup(ClassFile.CodeAttribute codeAttribute, int n) {
        switch (n) {

        case 0:
            ;
            break;

        case 1:
            codeAttribute.writeOpcode(Opcode.DUP);
            break;

        case 2:
            codeAttribute.writeOpcode(Opcode.DUP2);
            break;

        default:
            throw new RuntimeException("dup(" + n + ")");
        }
    }
    /*package*/ static void dup(ClassFile.CodeAttribute codeAttribute, IClass type) {
        codeAttribute.writeOpcode(
            type == IClass.LONG || type == IClass.DOUBLE ?
            Opcode.DUP2 :
            Opcode.DUP
        );
    }
    /*package*/ static void dupx(
        ClassFile.CodeAttribute codeAttribute,
        IClass                  type,
        int                     x
    ) {
        if (x < 0 || x > 2) throw new RuntimeException();
        int dup  = Opcode.DUP  + x;
        int dup2 = Opcode.DUP2 + x;
        codeAttribute.writeOpcode(
            type == IClass.LONG || type == IClass.DOUBLE ?
            dup2 :
            dup
        );
    }

    private static void pop(ClassFile.CodeAttribute codeAttribute, IClass type) {
        if (type == IClass.VOID) return;
        codeAttribute.writeOpcode(
            type == IClass.LONG || type == IClass.DOUBLE ?
            Opcode.POP2 :
            Opcode.POP
        );
    }

    /**
     * An exception that reflects an error during compilation.
     *
     * This exception is associated with a particular {@link Scanner.Location
     * Location} in the source code.
     */
    public static class CompileException extends Exception {
        CompileException(String message, Scanner.Location location) {
            super(message);
            this.location = location;
        }

        /**
         * Returns the message specified at creation time, preceeded
         * with nicely formatted location information.
         */
        public String getMessage() {
            return location.toString() + ": " + super.getMessage();
        }
        private final Scanner.Location location;
    }

    public static String join(String[] sa, String separator) {
        if (sa == null) return ("(null)");
        if (sa.length == 0) return ("(zero length array)");
        StringBuffer sb = new StringBuffer(sa[0]);
        for (int i = 1; i < sa.length; ++i) {
            sb.append(separator);
            sb.append(sa[i]);
        }
        return sb.toString();
    }

    private static int ilfd(IClass t) {
        if (
            t == IClass.BYTE ||
            t == IClass.CHAR ||
            t == IClass.INT ||
            t == IClass.SHORT ||
            t == IClass.BOOLEAN
        ) return 0;
        if (t == IClass.LONG  ) return 1;
        if (t == IClass.FLOAT ) return 2;
        if (t == IClass.DOUBLE) return 3;
        throw new RuntimeException("Unexpected type \"" + t + "\"");
    }
    private static int ilfd(
        IClass t,
        int    opcodeInt,
        int    opcodeLong,
        int    opcodeFloat,
        int    opcodeDouble
    ) {
        if (
            t == IClass.BYTE ||
            t == IClass.CHAR ||
            t == IClass.INT ||
            t == IClass.SHORT ||
            t == IClass.BOOLEAN
        ) return opcodeInt;
        if (t == IClass.LONG  ) return opcodeLong;
        if (t == IClass.FLOAT ) return opcodeFloat;
        if (t == IClass.DOUBLE) return opcodeDouble;
        throw new RuntimeException("Unexpected type \"" + t + "\"");
    }
    private static int ilfda(IClass t) {
        return !t.isPrimitive() ? 4 : Java.ilfd(t);
    }
    private static int ilfdabcs(IClass t) {
        if (t == IClass.INT   ) return 0;
        if (t == IClass.LONG  ) return 1;
        if (t == IClass.FLOAT ) return 2;
        if (t == IClass.DOUBLE) return 3;
        if (!t.isPrimitive()  ) return 4;
        if (t == IClass.BOOLEAN || t == IClass.BYTE) return 5;
        if (t == IClass.CHAR  ) return 6;
        if (t == IClass.SHORT ) return 7;
        throw new RuntimeException("Unexpected type \"" + t + "\"");
    }
}
