
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.io.*;
import java.util.*;

/**
 * Implementation of a simplified compiler for the Java<sup>TM</sup>
 * programming language.
 *
 * The following elements of the Java programming language are implemented:
 * <ul>
 * <li><tt>package</tt> declaration
 * <li><tt>import</tt> declaration
 * <li><tt>class</tt> declaration
 * <li><tt>interface</tt> declaration
 * <li>Inheritance (<tt>extends</tt> and <tt>implements</tt>)
 * <li>Static member type declaration
 * <li>Class initializer
 * <li>Instance initializer
 * <li>Field declaration
 * <li>Method declaration
 * <li>Local variable declaration
 * <li>Class variable initializer
 * <li>Instance variable initializer
 * <li>Block statement (<tt>{ ... }</tt>)
 * <li><tt>if ... else</tt> statement
 * <li><tt>for</tt> statement
 * <li><tt>while</tt> statement
 * <li><tt>do ... while</tt> statement
 * <li><tt>try ... catch ... finally</tt> statement
 * <li><tt>throw</tt> statement
 * <li><tt>return</tt> statement
 * <li><tt>break</tt> statement
 * <li><tt>continue</tt> statement
 * <li><tt>switch</tt> statement
 * <li>All primitive types (<tt>boolean, char, byte, short, int, long, float,
 * double</tt>)
 * <li>Assignment operator <tt>=</tt>
 * <li>Assignment operators <tt>+=, -=, *=, /=, &amp;=, |=, ^=, %=, &lt;&lt;=,
 * &gt;&gt;=, &gt;&gt;&gt;=</tt>
 * <li>Conditional operators <tt>?...:, &amp;&amp;, ||</tt>
 * <li>Boolean logical operators <tt>&amp;, ^, |</tt>
 * <li>Integer bitwise operators <tt>&amp;, ^, |</tt>
 * <li>Numeric operators <tt>*, /, %, +, -, &lt;&lt;, &gt;&gt;,
 * &gt;&gt;&gt;</tt>
 * <li>String concatenation operator <tt>+</tt>
 * <li>Operators <tt>++</tt> and <tt>--</tt>
 * <li>Type comparison operator <tt>instanceof</tt>
 * <li>Unary operators <tt>+, -, ~, !</tt>
 * <li>Parenthesized expression
 * <li>Field access (like <tt>System.out</tt>)
 * <li>Superclass member access (<tt>super.meth();</tt>, <tt>super.field =
 * x;</tt>
 * <li>Alternate constructor invocation (like <tt>this(a, b, c)</tt>)
 * <li>Superclass constructor invocation (like <tt>super(a, b, c)</tt>)
 * <li>Method invocation (like <tt>System.out.println("Hello")</tt>)
 * (partially)
 * <li>Class instance creation (like <tt>new Foo()</tt>)
 * <li>Primitive array creation(like <tt>new int[10][5][]</tt>)
 * <li>Class or interface array creation(like <tt>new Foo[10][5][]</tt>)
 * <li>Array access (like <tt>args[0]</tt>) (currently read-only)
 * <li>Local variable access
 * <li>Integer, floating-point, boolean, character, string literal
 * <li><tt>null</tt> literal
 * <li>Unary numeric conversion
 * <li>Binary numeric conversion
 * <li>Widening numeric conversion
 * <li>Narrowing numeric conversion
 * <li>Widening reference conversion
 * <li>Narrowing reference conversion
 * <li>Cast
 * <li>Assignment conversion
 * <li>String conversion (for string concatenation)
 * <li>Constant expression
 * <li>Block scope, method scope, class scope, global scope
 * <li><tt>throws</tt> clause
 * <li>Array initializer (like <tt>String[] a = { "x", "y", "z" }</tt>)
 * </ul>
 *
 * Elements of the Java programming language that are not (yet) implemented:
 *
 * <ul>
 * <li>Non-static member types ("inner" classes, including anonymous classes)
 * <li><tt>synchronized</tt> statement
 * <li>Access control
 * <li>References between uncompiled compilation units
 * <li>Line number tables
 * <li>Class literals, e.g. "String.class"
 * </ul>
 */

public class Parser {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/Parser.java,v 1.35 2002/05/21 22:40:00 arno Exp $";

    public Parser(Scanner scanner) {
        this.scanner = scanner;
    }

    /**
     * <pre>
     *   CompilationUnit := [ 'package' QualifiedIdentifier ';' ]
     *                      { ImportDeclaration }
     *                      { TypeDeclaration }
     * </pre>
     */
    /*package*/ Java.CompilationUnit parseCompilationUnit() throws ParseException, Scanner.ScanException, IOException {
        Java.CompilationUnit compilationUnit = new Java.CompilationUnit();

        if (scanner.peek().isKeyword("package")) {
            scanner.read();
            compilationUnit.setPackage(join(parseQualifiedIdentifier(), "."));
            if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected after \"package\" directive");
        }

        while (scanner.peek().isKeyword("import")) parseImportDeclaration(compilationUnit);

        while (!scanner.peek().isEOF()) {
            Java.ClassOrInterfaceDeclaration cid = parseTypeDeclaration(compilationUnit);
            if (cid != null) compilationUnit.addClassOrInterfaceDeclaration(cid);
        }

        return compilationUnit;
    }

    /**
     * <pre>
     *   ImportDeclaration := 'import' Identifier
     *                        { '.' Identifier }
     *                        [ '.' '*' ]
     *                        ';'
     * </pre>
     */
    /*package*/ void parseImportDeclaration(Java.CompilationUnit compilationUnit) throws ParseException, Scanner.ScanException, IOException {
        if (!this.scanner.read().isKeyword("import")) this.throwParseException("\"import\" expected");
        if (!this.scanner.peek().isIdentifier()) this.throwParseException("Identifier expected after \"import\"");
        String s = this.scanner.read().getIdentifier();
        StringBuffer sb = new StringBuffer(s);
        for (;;) {
            if (scanner.peek().isOperator(";")) {
                compilationUnit.addImportedClass(
                    scanner.read().getLocation(), // location
                    s,                            // className
                    sb.toString()                 // fullClassName
                );
                break;
            }
            if (!scanner.read().isOperator(".")) this.throwParseException("\";\" or \".\" expected after identifier in \"import\" directive");
            if (scanner.peek().isOperator("*")) {
                scanner.read();
                compilationUnit.addImportedPackage(sb.toString());
                if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected after \"import\" directive");
                break;
            }
            if (!scanner.peek().isIdentifier()) this.throwParseException("Identifier or \"*\" expected after \".\" in \"import\" directive");
            s = scanner.read().getIdentifier();
            sb.append(".");
            sb.append(s);
        }
    }

    /**
     * QualifiedIdentifier := Identifier { '.' Identifier }
     */
    /*package*/ String[] parseQualifiedIdentifier() throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isIdentifier()) this.throwParseException("Identifier expected");
        String s = scanner.read().getIdentifier();
        Vector v = new Vector();
        v.add(s);
        while (scanner.peek().isOperator(".")) {
            scanner.read();
            if (!scanner.peek().isIdentifier()) this.throwParseException("Identifier expected");
            s = scanner.read().getIdentifier();
            v.add(s);
        }
        return (String[]) v.toArray(new String[v.size()]);
    }

    /**
     * <pre>
     *   TypeDeclaration := ClassOrInterfaceDeclaration
     * </pre>
     */
    /*package*/ Java.ClassOrInterfaceDeclaration parseTypeDeclaration(
        Java.CompilationUnit compilationUnit
    ) throws ParseException, Scanner.ScanException, IOException {
        if (scanner.peek().isOperator(";")) {
            scanner.read();
            return null;
        }
        return parseClassOrInterfaceDeclaration(compilationUnit);
    }

    /**
     * <pre>
     *   ClassOrInterfaceDeclaration :=
     *             ModifiersOpt 'class' ClassDeclarationRest |
     *             ModifiersOpt 'interface' InterfaceDeclarationRest
     * </pre>
     */
    /*package*/ Java.ClassOrInterfaceDeclaration parseClassOrInterfaceDeclaration(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        short modifiers = parseModifiersOpt();

        Java.ClassOrInterfaceDeclaration res;
        if (scanner.peek().isKeyword("class")) {
            scanner.read();
            res = parseClassDeclarationRest(
                enclosingScope,
                modifiers
            );
        } else
        if (scanner.peek().isKeyword("interface")) {
            scanner.read();
            res = parseInterfaceDeclarationRest(
                enclosingScope,
                modifiers
            );
        } else
        {
            this.throwParseException("Unexpected token \"" + scanner.peek() + "\" in class or interface declaration");
            /* NEVER REACHED */ return null;
        }
        return res;
    }

    /**
     * <pre>
     *   ModifiersOpt := { 'public' | 'protected' | 'private' | 'static' |
     *           'abstract' | 'final' | 'native' | 'synchronized' |
     *           'transient' | 'volatile' | 'strictfp'
     * </pre>
     */
    /*package*/ short parseModifiersOpt() throws ParseException, Scanner.ScanException, IOException {
        short mod = 0;
        while (scanner.peek().isKeyword()) {
            String kw = this.scanner.peek().getKeyword();
            short x = (short) (
                kw == "public"       ? Mod.PUBLIC       :
                kw == "protected"    ? Mod.PROTECTED    :
                kw == "private"      ? Mod.PRIVATE      :
                kw == "static"       ? Mod.STATIC       :
                kw == "abstract"     ? Mod.ABSTRACT     :
                kw == "final"        ? Mod.FINAL        :
                kw == "native"       ? Mod.NATIVE       :
                kw == "synchronized" ? Mod.SYNCHRONIZED :
                kw == "transient"    ? Mod.TRANSIENT    :
                kw == "volatile"     ? Mod.VOLATILE     :
                kw == "strictfp"     ? Mod.STRICTFP     :
                -1
            );
            if (x == -1) break;

            scanner.read();
            if ((mod & x) != 0) this.throwParseException("Duplicate modifier \"" + kw + "\"");
            for (int i = 0; i < mutuals.length; ++i) {
                short m = mutuals[i];
                if ((x & m) != 0 && (mod & m) != 0) {
                    this.throwParseException("Only one of \"" + Mod.shortToString(m) + "\" allowed");
                }
            }
            mod |= x;
        }
        return mod;
    }
    private final short[] mutuals = {
        Mod.PUBLIC | Mod.PROTECTED | Mod.PRIVATE,
    };

    /**
     * <pre>
     *   ClassDeclarationRest :=
     *        Identifier
     *        [ 'extends' ReferenceType ]
     *        [ 'implements' ReferenceTypeList ]
     *        ClassBody
     * </pre>
     */
    /*package*/ Java.ClassDeclaration parseClassDeclarationRest(
        Java.Scope enclosingScope,
        short      modifiers
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isIdentifier()) this.throwParseException("Class name expected after \"class\"");
        Scanner.Location location = scanner.peek().getLocation();
        String className = scanner.read().getIdentifier();

        Java.ReferenceType extendedType = null;
        if (scanner.peek().isKeyword("extends")) {
            scanner.read();
            extendedType = parseReferenceType(enclosingScope);
        }

        Java.ReferenceType[] implementedTypes = new Java.ReferenceType[0];
        if (scanner.peek().isKeyword("implements")) {
            scanner.read();
            implementedTypes = parseReferenceTypeList(enclosingScope);
        }

        Java.ClassDeclaration res = new Java.ClassDeclaration(
            location,
            enclosingScope,
            modifiers,
            className,
            extendedType,
            implementedTypes
        );

        parseClassBody(res);
        return res;
    }

    /**
     * <pre>
     *   ClassBody := '{' {
     *     ';' |
     *     ModifiersOpt (
     *       Block |
     *       'void' Identifier MethodDeclaratorRest |
     *       'class' ClassDeclarationRest |
     *       'interface' InterfaceDeclarationRest |
     *       ConstructorDeclarator |
     *       Type Identifier (
     *         MethodDeclaratorRest |
     *         VariableDeclaratorsRest
     *       )
     *     )
     *   } '}'
     * </pre>
     */
    /*package*/ void parseClassBody(
        Java.ClassDeclaration classDeclaration
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isOperator("{")) this.throwParseException("\"{\" expected at start of class body");
        scanner.read();

        for (;;) {
            if (scanner.peek().isOperator("}")) {
                scanner.read();
                return;
            }

            if (scanner.peek().isOperator(";")) {
                scanner.read();
                continue;
            }

            short modifiers = parseModifiersOpt();

            // Initializer?
            if (scanner.peek().isOperator("{")) {
                if ((modifiers & ~Mod.STATIC) != 0) this.throwParseException("Only modifier \"static\" allowed on initializer");

                Java.Block b = parseBlock((Java.Scope) classDeclaration);
                if ((modifiers & Mod.STATIC) != 0) {
                    classDeclaration.addClassVariableDeclaratorOrClassInitializer(b);
                } else {
                    classDeclaration.addInstanceVariableDeclaratorOrInstanceInitializer(b);
                }
            } else

            // "void" method declaration.
            if (scanner.peek().isKeyword("void")) {
                Scanner.Location location = scanner.read().getLocation();
                if (!scanner.peek().isIdentifier()) this.throwParseException("Method name expected after \"void\"");
                String name = scanner.read().getIdentifier();
                classDeclaration.declaredMethods.add(parseMethodDeclaratorRest(
                    classDeclaration, // declaringClassOrInterface
                    modifiers,
                    new Java.BasicType(location, Java.BasicType.VOID),
                    name
                ));
            } else

            // Member class.
            if (scanner.peek().isKeyword("class")) {
                scanner.read();
                classDeclaration.declaredClasses.add(parseClassDeclarationRest(
                    classDeclaration, // enclosingScope
                    modifiers
                ));
            } else

            // Member interface.
            if (scanner.peek().isKeyword("interface")) {
                scanner.read();
                classDeclaration.declaredInterfaces.add(parseInterfaceDeclarationRest(
                    classDeclaration, // enclosingScope
                    (short) (modifiers | Mod.STATIC)
                ));
            } else

            // Constructor.
            if (
                scanner.peek().isIdentifier(classDeclaration.name) &&
                scanner.peekNextButOne().isOperator("(")
            ) {
                classDeclaration.addConstructor(parseConstructorDeclarator(
                    classDeclaration, // declaringClassOrInterface
                    modifiers
                ));
            } else

            // Member method or field.
            {
                Java.Type memberType = parseType(classDeclaration);
                if (!scanner.peek().isIdentifier()) this.throwParseException("Identifier expected in member declaration");
                Scanner.Location location = scanner.peek().getLocation();
                String memberName = scanner.read().getIdentifier();

                // Method declarator.
                if (scanner.peek().isOperator("(")) {
                    classDeclaration.declaredMethods.add(parseMethodDeclaratorRest(
                        classDeclaration, // declaringClassOrInterface
                        modifiers,
                        memberType,
                        memberName
                    ));
                } else

                // Field declarator.
                {
                    Java.FieldDeclarator fd = new Java.FieldDeclarator(
                        location,                     // location
                        classDeclaration,             // declaringClassOrInterface
                        modifiers,                    // modifiers
                        memberType,                   // type
                        parseVariableDeclaratorsRest( // variableDeclarators
                            classDeclaration, // enclosingScope
                            memberType,       // type
                            memberName        // name
                        )
                    );
                    if ((modifiers & Mod.STATIC) != 0) {
                        classDeclaration.addClassVariableDeclaratorOrClassInitializer(fd);
                    } else {
                        classDeclaration.addInstanceVariableDeclaratorOrInstanceInitializer(fd);
                    }
                }
            }
        }
    }

    /**
     * <pre>
     *   InterfaceDeclarationRest :=
     *     Identifier
     *     [ 'extends' ReferenceTypeList ]
     *     InterfaceBody
     * </pre>
     */
    /*package*/ Java.InterfaceDeclaration parseInterfaceDeclarationRest(
        Java.Scope enclosingScope,
        short      modifiers
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isIdentifier()) this.throwParseException("Interface name expected after \"interface\"");
        Scanner.Location location = scanner.peek().getLocation();
        String interfaceName = scanner.read().getIdentifier();

        Java.ReferenceType[] extendedTypes = new Java.ReferenceType[0];
        if (scanner.peek().isKeyword("extends")) {
            scanner.read();
            extendedTypes = parseReferenceTypeList(enclosingScope);
        }

        Java.InterfaceDeclaration res = new Java.InterfaceDeclaration(
            location,
            enclosingScope,
            modifiers,
            interfaceName,
            extendedTypes
        );

        parseInterfaceBody(res);
        return res;
    }

    /**
     * <pre>
     *   InterfaceBody := '{' {
     *     ';' |
     *     ModifiersOpt (
     *       'void' Identifier MethodDeclaratorRest |
     *       'class' ClassDeclarationRest |
     *       'interface' InterfaceDeclarationRest |
     *       Type Identifier (
     *         MethodDeclaratorRest |
     *         VariableDeclaratorsRest
     *       )
     *     )
     *   } '}'
     * </pre>
     */
    /*package*/ void parseInterfaceBody(
        Java.InterfaceDeclaration interfaceDeclaration
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.read().isOperator("{")) this.throwParseException("\"{\" expected at start of interface body");

        for (;;) {
            if (scanner.peek().isOperator("}")) {
                scanner.read();
                break;
            }

            if (scanner.peek().isOperator(";")) {
                scanner.read();
                continue;
            }

            short modifiers = parseModifiersOpt();

            // "void" method declaration.
            if (scanner.peek().isKeyword("void")) {
                Scanner.Location location = scanner.read().getLocation();
                if (!scanner.peek().isIdentifier()) this.throwParseException("Method name expected after \"void\"");
                String name = scanner.read().getIdentifier();
                interfaceDeclaration.declaredMethods.add(parseMethodDeclaratorRest(
                    interfaceDeclaration, // declaringClassOrInterface
                    (short) (modifiers | Mod.ABSTRACT | Mod.PUBLIC),
                    new Java.BasicType(location, Java.BasicType.VOID),
                    name
                ));
            } else

            // Member class.
            if (scanner.peek().isKeyword("class")) {
                scanner.read();
                interfaceDeclaration.declaredClasses.add(parseClassDeclarationRest(
                    interfaceDeclaration, // enclosingScope
                    (short) (modifiers | Mod.STATIC | Mod.PUBLIC)
                ));
            } else

            // Member interface.
            if (scanner.peek().isKeyword("interface")) {
                scanner.read();
                interfaceDeclaration.declaredInterfaces.add(parseInterfaceDeclarationRest(
                    interfaceDeclaration, // enclosingScope
                    (short) (modifiers | Mod.STATIC | Mod.PUBLIC)
                ));
            } else

            // Member method or field.
            {
                Java.Type memberType = parseType(interfaceDeclaration);
                if (!scanner.peek().isIdentifier()) this.throwParseException("Identifier expected in member declaration");
                String memberName = scanner.peek().getIdentifier();
                Scanner.Location location = scanner.read().getLocation();

                // Method declarator.
                if (scanner.peek().isOperator("(")) {
                    interfaceDeclaration.declaredMethods.add(parseMethodDeclaratorRest(
                        interfaceDeclaration, // declaringClassOrInterface
                        (short) (modifiers | Mod.ABSTRACT | Mod.PUBLIC),
                        memberType,
                        memberName
                    ));
                } else

                // Field declarator.
                {
                    interfaceDeclaration.addConstantDeclaration(new Java.FieldDeclarator(
                        location,                     // location
                        interfaceDeclaration,         // declaringClassOrInterface
                        (short) (                     // modifiers
                            modifiers |
                            Mod.PUBLIC | Mod.STATIC | Mod.FINAL
                        ),
                        memberType,                   // type
                        parseVariableDeclaratorsRest( // variableDeclarators
                            interfaceDeclaration, // enclosingScope
                            memberType,           // type
                            memberName            // name
                        )
                    ));
                }
            }
        }
    }

    /**
     * <pre>
     *   ConstructorDeclarator :=
     *     FormalParameters
     *     [ 'throws' ReferenceTypeList ]
     *     '{'
     *       [ 'this' Arguments ';' | 'super' Arguments ';' ]
     *       BlockStatements
     *     '}'
     * </pre>
     */
    /*package*/ Java.ConstructorDeclarator parseConstructorDeclarator(
        Java.ClassDeclaration declaringClass,
        short                 modifiers
    ) throws ParseException, Scanner.ScanException, IOException {
        Scanner.Location location = scanner.read().getLocation();

        // Parse formal parameters.
        Java.FormalParameter[] formalParameters = parseFormalParameters(
            declaringClass // enclosingScope
        );

        // Parse "throws" clause.
        Java.ReferenceType[] thrownExceptions;
        if (scanner.peek().isKeyword("throws")) {
            scanner.read();
            thrownExceptions = parseReferenceTypeList(declaringClass);
        } else {
            thrownExceptions = new Java.ReferenceType[0];
        }

        // Create ConstructorDeclarator object.
        Java.ConstructorDeclarator cd = new Java.ConstructorDeclarator(
            location,
            declaringClass,
            modifiers,
            formalParameters,
            thrownExceptions
        );

        // Parse constructor body.
        location = scanner.peek().getLocation();
        if (!scanner.read().isOperator("{")) this.throwParseException("\"{\" expected");
        if (
            scanner.peek().isKeyword("this") &&
            scanner.peekNextButOne().isOperator("(")
        ) {
            scanner.read();
            scanner.read();
            cd.setAlternateConstructorInvocation(new Java.ConstructorInvocation(
                scanner.read().getLocation(),  // location
                declaringClass,                // declaringClass
                cd,                            // declaringConstructor
                false,                         // targetIsSuperclass
                parseArguments(declaringClass) // arguments
            ));
            if (!scanner.read().isOperator(";")) this.throwParseException("\";\" expected");
        } else
        if (
            scanner.peek().isKeyword("super") &&
            scanner.peek().isOperator("(")
        ) {
            scanner.read();
            scanner.read();
            cd.setSuperclassConstructorInvocation(new Java.ConstructorInvocation(
                scanner.read().getLocation(),   // location
                declaringClass,                 // declaringClass
                cd,                             // declaringConstructor
                true,                           // targetIsSuperclass
                parseArguments(declaringClass)  // arguments
            ));
            if (!scanner.read().isOperator(";")) this.throwParseException("\";\" expected");
        }
        Java.Block body = new Java.Block(
            location, // location
            cd        // enclosingScope
        );
        body.addStatements(parseBlockStatements(
            body // enclosingBlock
        ));
        cd.setBody(body);

        if (!scanner.read().isOperator("}")) this.throwParseException("\"}\" expected");

        return cd;
    }

    /**
     * <pre>
     *   MethodDeclaratorRest :=
     *     FormalParameters |
     *     [ 'throws' ReferenceTypeList ]
     *     ( ';' | MethodBody )
     * </pre>
     */
    /*package*/ Java.MethodDeclarator parseMethodDeclaratorRest(
        Java.ClassOrInterfaceDeclaration declaringClassOrInterface,
        short                            modifiers,
        Java.Type                        type,
        String                           name
    ) throws ParseException, Scanner.ScanException, IOException {
        Scanner.Location location = scanner.peek().getLocation();

        Java.FormalParameter[] formalParameters = parseFormalParameters(
            declaringClassOrInterface // enclosingScope
        );

        Java.ReferenceType[] thrownExceptions;
        if (scanner.peek().isKeyword("throws")) {
            scanner.read();
            thrownExceptions = parseReferenceTypeList(declaringClassOrInterface);
        } else {
            thrownExceptions = new Java.ReferenceType[0];
        }

        Java.MethodDeclarator md = new Java.MethodDeclarator(
            location,
            declaringClassOrInterface,
            modifiers,
            type,
            name,
            formalParameters,
            thrownExceptions
        );
        Java.Block body = null;
        if (scanner.peek().isOperator(";")) {
            if ((modifiers & Mod.ABSTRACT) == 0) this.throwParseException("Non-abstract method must have a body");
            scanner.read();
        } else {
            if ((modifiers & Mod.ABSTRACT) != 0) this.throwParseException("Abstract method must not have a body");
            md.setBody(parseMethodBody(
                md // declaringFunction
            ));
        }

        return md;
    }

    /**
     * <pre>
     *   VariableInitializer :=
     *     ArrayInitializer |
     *     Expression
     * </pre>
     */
    /*package*/ Java.Rvalue parseVariableInitializer(
        Java.Scope enclosingScope,
        Java.Type  type
    ) throws ParseException, Scanner.ScanException, IOException {
        if (scanner.peek().isOperator("{")) {
            if (!(type instanceof Java.ArrayType)) this.throwParseException("Cannot initialize non-array type \"" + type + "\" with an array");
            return this.parseArrayInitializer(
                enclosingScope,
                (Java.ArrayType) type
            );
        }
        return this.parseExpression(enclosingScope).toRvalue();
    }

    /**
     * <pre>
     *   ArrayInitializer :=
     *     '{' [ VariableInitializer { ',' VariableInitializer } [ ',' ] '}'
     * </pre>
     */
    /*package*/ Java.Rvalue parseArrayInitializer(
        Java.Scope     enclosingScope,
        Java.ArrayType arrayType
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!this.scanner.peek().isOperator("{")) this.throwParseException("\"{\" expected");
        Scanner.Location location = this.scanner.read().getLocation();
        Java.Type componentType = arrayType.getComponentType();
        Vector v = new Vector(); // Rvalue
        while (!this.scanner.peek().isOperator("}")) {
            v.add(this.parseVariableInitializer(
                enclosingScope,
                componentType
            ));
            if (this.scanner.peek().isOperator("}")) break;
            if (!this.scanner.peek().isOperator(",")) this.throwParseException("\",\" or \"}\" expected");
            this.scanner.read();
        }
        this.scanner.read();
        return new Java.ArrayInitializer(
            location,
            arrayType,
            (Java.Rvalue[]) v.toArray(new Java.Rvalue[v.size()])
        );
    }

    /**
     * <pre>
     *   FormalParameters := '(' [ FormalParameter { ',' FormalParameter } ] ')'
     * </pre>
     */
    /*package*/ Java.FormalParameter[] parseFormalParameters(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.read().isOperator("(")) this.throwParseException("\"(\" expected in formal parameter list");
        if (scanner.peek().isOperator(")")) {
            scanner.read();
            return new Java.FormalParameter[0];
        }

        Vector v = new Vector();
        for (;;) {
            v.add(parseFormalParameter(enclosingScope));
            if (!scanner.peek().isOperator(",")) break;
            scanner.read();
        }
        if (!scanner.read().isOperator(")")) this.throwParseException("\")\" expected at end of formal parameter list");
        return (Java.FormalParameter[]) v.toArray(new Java.FormalParameter[v.size()]);
    }

    /**
     * <pre>
     *   FOrmalParameter := [ 'final' ] Type Identifier BracketsOpt
     * </pre>
     */
    /*package*/ Java.FormalParameter parseFormalParameter(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        boolean finaL = scanner.peek().isKeyword("final");
        if (finaL) scanner.read();

        Java.Type type = parseType(enclosingScope);
        if (!scanner.peek().isIdentifier()) this.throwParseException("Formal parameter name expected");
        Scanner.Location location = scanner.peek().getLocation();
        String name = scanner.read().getIdentifier();

        for (int i = parseBracketsOpt(); i > 0; --i) type = new Java.ArrayType(type);
        return new Java.FormalParameter(location, finaL, type, name);
    }

    /**
     * <pre>
     *   BracketsOpt := { '[' ']' }
     * </pre>
     */
    int parseBracketsOpt() throws ParseException, Scanner.ScanException, IOException {
        int res = 0;
        while (
            scanner.peek().          isOperator("[") &&
            scanner.peekNextButOne().isOperator("]")
        ) {
            scanner.read();
            scanner.read();
            ++res;
        }
        return res;
    }

    /**
     * <pre>
     *   MethodBody := Block
     * </pre>
     */
    /*package*/ Java.Block parseMethodBody(
        Java.FunctionDeclarator declaringFunction
    ) throws ParseException, Scanner.ScanException, IOException {
        return parseBlock(
            declaringFunction // enclosingScope
        );
    }

    /**
     * <pre>
     *   '{' BlockStatements '}'
     * </pre>
     */
    /*package*/ Java.Block parseBlock(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isOperator("{")) this.throwParseException("\"{\" expected");
        Java.Block block = new Java.Block(
            scanner.read().getLocation(),
            enclosingScope
        );
        block.addStatements(parseBlockStatements(block));
        if (!scanner.read().isOperator("}")) this.throwParseException("\"}\" expected");
        return block;
    }

    /**
     * <pre>
     *   BlockStatements := { BlockStatement }
     * </pre>
     */
    /*package*/ Vector parseBlockStatements(
        Java.Block enclosingBlock
    ) throws ParseException, Scanner.ScanException, IOException {
        Vector v = new Vector();
        while (
            !scanner.peek().isOperator("}") &&
            !scanner.peek().isKeyword("case")
        ) v.add(parseBlockStatement(enclosingBlock));
        return v;
    }

    /**
     * <pre>
     *   BlockStatement := { Identifier ':' } (
     *     ( Modifiers Type | ModifiersOpt BasicType ) VariableDeclarators ';' |
     *     'class' ... |
     *     Statement |
     *     Expression ';' |
     *     Expression [ Brackets ] VariableDeclarators ';'   (1)
     *   )
     * </pre>
     *
     * (1) "Expression" must pose a type.
     */
    /*package*/ Java.BlockStatement parseBlockStatement(
        Java.Block enclosingBlock
    ) throws ParseException, Scanner.ScanException, IOException {

        // Statement?
        if (
            (
                this.scanner.peek().isIdentifier() &&
                this.scanner.peekNextButOne().isOperator(":")
            ) ||
            scanner.peek().isKeyword(new String[] {
                "if", "for", "while", "do", "try", "switch", "synchronized",
                "return", "throw", "break", "continue"
            }) ||
            scanner.peek().isOperator(new String[] { "{", ";" })
        ) return parseStatement((Java.Scope) enclosingBlock);

        // Local class declaration?
        if (scanner.peek().isKeyword("class")) {
            this.throwParseException("Local class declarations NYI");
        }

        Java.Atom a = parseExpression(enclosingBlock);

        // Expression ';'
        if (scanner.peek().isOperator(";")) {
            scanner.read();
            return new Java.ExpressionStatement(a.toRvalue(), (Java.Scope) enclosingBlock);
        }

        // Expression VariableDeclarators ';'
        Java.Type variableType = a.toType();
        Java.LocalVariableDeclarationStatement lvds = new Java.LocalVariableDeclarationStatement(
            a.getLocation(),          // location
            enclosingBlock,           // declaringBlock
            (short) 0,                // modifiers
            variableType,             // type
            parseVariableDeclarators( // variableDeclarators
                (Java.Scope) enclosingBlock, // enclosingScope
                variableType                 // type
            )
        );
        if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected after local variable declarator");
        return lvds;
    }

    /**
     * <pre>
     *   VariableDeclarators := VariableDeclarator { ',' VariableDeclarator }
     * </pre>
     */
    /*package*/ Java.VariableDeclarator[] parseVariableDeclarators(
        Java.Scope enclosingScope,
        Java.Type  type
    ) throws ParseException, Scanner.ScanException, IOException {
        Vector v = new Vector();
        v.add(parseVariableDeclarator(enclosingScope, type));
        while (scanner.peek().isOperator(",")) {
            scanner.read();
            v.add(parseVariableDeclarator(enclosingScope, type));
        }
        return (Java.VariableDeclarator[]) v.toArray(new Java.VariableDeclarator[v.size()]);
    }

    /**
     * <pre>
     *   VariableDeclaratorsRest :=
     *     VariableDeclaratorRest
     *     { ',' VariableDeclarator }
     * </pre>
     */
    /*package*/ Java.VariableDeclarator[] parseVariableDeclaratorsRest(
        Java.Scope enclosingScope,
        Java.Type  type,
        String     name
    ) throws ParseException, Scanner.ScanException, IOException {
        Vector v = new Vector();
        v.add(parseVariableDeclaratorRest(enclosingScope, type, name));
        while (scanner.peek().isOperator(",")) {
            scanner.read();
            v.add(parseVariableDeclarator(enclosingScope, type));
        }
        return (Java.VariableDeclarator[]) v.toArray(new Java.VariableDeclarator[v.size()]);
    }

    /**
     * <pre>
     *   VariableDeclarator := Identifier VariableDeclaratorRest
     * </pre>
     */
    /*package*/ Java.VariableDeclarator parseVariableDeclarator(
        Java.Scope enclosingScope,
        Java.Type  type
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isIdentifier()) this.throwParseException("Variable name expected");
        String variableName = scanner.read().getIdentifier();
        return parseVariableDeclaratorRest(enclosingScope, type, variableName);
    }

    /**
     * <pre>
     *   VariableDeclaratorRest := [ '=' VariableInitializer ]
     * </pre>
     */
    /*package*/ Java.VariableDeclarator parseVariableDeclaratorRest(
        Java.Scope enclosingScope,
        Java.Type  type,
        String     name
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Rvalue initializer = null;
        if (scanner.peek().isOperator("=")) {
            scanner.read();
            initializer = parseVariableInitializer(enclosingScope, type);
        }
        return new Java.VariableDeclarator(name, parseBracketsOpt(), initializer);
    }

    /**
     * <pre>
     *   Statement :=
     *     LabeledStatement |
     *     Block |
     *     IfStatement |
     *     ForStatement |
     *     WhileStatement |
     *     DoStatement |
     *     TryStatement |
     *     'switch' ... |
     *     'synchronized' ... |
     *     ReturnStatement |
     *     ThrowStatement |
     *     BreakStatement |
     *     ContinueStatement |
     *     ';' |
     *     ExpressionStatement
     * </pre>
     */
    /*package*/ Java.Statement parseStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (
            this.scanner.peek().isIdentifier() &&
            this.scanner.peekNextButOne().isOperator(":")
        ) {
            return parseLabeledStatement(enclosingScope);
        }

        Scanner.Token t = scanner.peek();
        Java.Statement stmt = (
            t.isOperator("{")           ? parseBlock(enclosingScope) :
            t.isKeyword("if")           ? parseIfStatement(enclosingScope) :
            t.isKeyword("for")          ? parseForStatement(enclosingScope) :
            t.isKeyword("while")        ? parseWhileStatement(enclosingScope) :
            t.isKeyword("do")           ? parseDoStatement(enclosingScope) :
            t.isKeyword("try")          ? parseTryStatement(enclosingScope) :
            t.isKeyword("switch")       ? parseSwitchStatement(enclosingScope) :
            t.isKeyword("synchronized") ? null :
            t.isKeyword("return")       ? parseReturnStatement(enclosingScope) :
            t.isKeyword("throw")        ? parseThrowStatement(enclosingScope) :
            t.isKeyword("break")        ? parseBreakStatement(enclosingScope) :
            t.isKeyword("continue")     ? parseContinueStatement(enclosingScope) :
            t.isOperator(";")           ? new Java.Statement(
                scanner.read().getLocation(),
                enclosingScope
            ) {
                public boolean compile(ClassFile.CodeAttribute codeAttribute) { return true; }
            } :
            parseExpressionStatement(enclosingScope)
        );
        if (stmt == null) this.throwParseException("\"" + t.getKeyword() + "\" NYI");

        return stmt;
    }

    /**
     * <pre>
     *   LabeledStatement := Identifier ':' Statement
     * </pre>
     */
    /*package*/ Java.Statement parseLabeledStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        Java.LabeledStatement labeledStatement = new Java.LabeledStatement(
            this.scanner.peek().getLocation(),  // location
            enclosingScope,                     // enclosingScope
            this.scanner.read().getIdentifier() // label
        );
        if (!this.scanner.read().isOperator(":")) this.throwParseException("Colon expected");
        labeledStatement.setBody(this.parseStatement(labeledStatement));
        return labeledStatement;
    }

    /**
     * <pre>
     *   IfStatement := 'if' '(' Expression ')' Statement [ 'else' Statement ]
     * </pre>
     */
    /*package*/ Java.Statement parseIfStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("if")) this.throwParseException("\"if\" expected");
        Scanner.Location location = scanner.read().getLocation();
        if (!scanner.read().isOperator("(")) this.throwParseException("Opening parenthesis expected after \"if\"");
        final Java.Rvalue condition = parseExpression(enclosingScope).toRvalue();
        if (!scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis expected after \"if\" condition");
        final Java.Statement trueStatement = parseStatement(enclosingScope);
        Java.Statement fs = null;
        if (scanner.peek().isKeyword("else")) {
            scanner.read();
            fs = parseStatement(enclosingScope);
        }
        final Java.Statement falseStatement = fs;

        return new Java.IfStatement(
            location,
            enclosingScope,
            condition,
            trueStatement,
            falseStatement
        );
    }

    /**
     * <pre>
     *   ForStatement :=
     *     'for' '('
     *       [ ForInit ] ';'
     *       [ Expression ] ';'
     *       [ StatementExpressionList ]
     *     ')' Statement
     * </pre>
     */
    /*package*/ Java.Statement parseForStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("for")) this.throwParseException("\"for\" expected");
        Scanner.Location location = scanner.read().getLocation();

        Java.ForStatement forStatement = new Java.ForStatement(location, enclosingScope);

        if (!scanner.read().isOperator("(")) this.throwParseException("Opening parenthesis expected after \"for\"");

        Java.BlockStatement forInit = null;
        if (!scanner.peek().isOperator(";")) {
            forInit = parseForInit(
                forStatement.implicitBlock // enclosingBlock
            );
        }
        if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected after \"for\" initializer");

        Java.Rvalue forCondition = null;
        if (!scanner.peek().isOperator(";")) {
            forCondition = parseExpression((Java.Scope) forStatement).toRvalue();
        }
        if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected after \"for\" condition");

        Java.BlockStatement forUpdate = null;
        if (!scanner.peek().isOperator(")")) {
            forUpdate = parseStatementExpressionList((Java.Scope) forStatement);
        }
        if (!scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis expected after \"for\" update");

        forStatement.set(
            forInit,
            forCondition,
            forUpdate,
            parseStatement((Java.Scope) forStatement)
        );

        return forStatement;
    }

    /**
     * <pre>
     *   ForInit :=
     *     Modifiers Type VariableDeclarators |
     *     ModifiersOpt BasicType VariableDeclarators |
     *     Expression (
     *       VariableDeclarators |       (1)
     *       { ',' Expression }
     *     )
     * </pre>
     *
     * (1) "Expression" must pose a type.
     */
    private Java.BlockStatement parseForInit(
        Java.Block enclosingBlock
    ) throws ParseException, Scanner.ScanException, IOException {

        // Modifiers Type VariableDeclarators
        // ModifiersOpt BasicType VariableDeclarators
        if (scanner.peek().isKeyword(new String[] { "final", "byte", "short", "char", "int", "long", "float", "double", "boolean" })) {
            short modifiers = parseModifiersOpt();
            Java.Type variableType = parseType((Java.Scope) enclosingBlock);
            return new Java.LocalVariableDeclarationStatement(
                scanner.peek().getLocation(), // location
                enclosingBlock,               // declaringBlock
                modifiers,                    // modifiers
                variableType,                 // type
                parseVariableDeclarators(     // variableDeclarators
                    (Java.Scope) enclosingBlock, // enclosingScope
                    variableType                 // type
                )
            );
        }

        Java.Atom a = parseExpression((Java.Scope) enclosingBlock);

        // Expression VariableDeclarators
        if (scanner.peek().isIdentifier()) {
            Java.Type variableType = a.toType();
            return new Java.LocalVariableDeclarationStatement(
                a.getLocation(),          // location
                enclosingBlock,           // declaringBlock
                Mod.NONE,                 // modifiers
                variableType,             // type
                parseVariableDeclarators( // variableDeclarators
                    (Java.Scope) enclosingBlock, // enclosingScope
                    variableType                 // type
                )
            );
        }

        // Expression { ',' Expression }
        if (!scanner.peek().isOperator(",")) {
            return new Java.ExpressionStatement(
                a.toRvalue(),               // rvalue
                (Java.Scope) enclosingBlock // enclosingScope
            );
        }
        scanner.read();
        Vector v = new Vector();
        v.add(new Java.ExpressionStatement(
            a.toRvalue(),               // rvalue
            (Java.Scope) enclosingBlock // enclosingScope
        ));
        for (;;) {
            v.add(new Java.ExpressionStatement(
                parseExpression(            // rvalue
                    (Java.Scope) enclosingBlock
                ).toRvalue(),
                (Java.Scope) enclosingBlock // enclosingScope
            ));
            if (!scanner.peek().isOperator(",")) break;
            scanner.read();
        }

        Java.Block b = new Java.Block(
            a.getLocation(),            // location
            (Java.Scope) enclosingBlock // enclosingScope
        );
        b.addStatements(v);
        return b;
    }

    /**
     * <pre>
     *   WhileStatement := 'while' '(' Expression ')' Statement
     * </pre>
     */
    /*package*/ Java.Statement parseWhileStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("while")) this.throwParseException("\"while\" expected");
        Scanner.Location location = scanner.read().getLocation();

        if (!scanner.read().isOperator("(")) this.throwParseException("Opening parenthesis expected after \"while\"");
        Java.WhileStatement whileStatement = new Java.WhileStatement(
            location,
            enclosingScope,
            parseExpression(enclosingScope).toRvalue() // condition
        );
        if (!scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis expected after \"while\" condition");

        whileStatement.setBody(parseStatement(whileStatement));
        return whileStatement;
    }

    /**
     * <pre>
     *   DoStatement := 'do' Statement 'while' '(' Expression ')' ';'
     * </pre>
     */
    /*package*/ Java.Statement parseDoStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("do")) this.throwParseException("\"do\" expected");
        Java.DoStatement doStatement = new Java.DoStatement(
            scanner.read().getLocation(),
            enclosingScope
        );

        doStatement.setBody(parseStatement(doStatement));

        if (!scanner.read().isKeyword("while")) this.throwParseException("\"while\" expected after body of \"do\" statement");
        if (!scanner.read().isOperator("(")) this.throwParseException("Opening parenthesis expected after \"while\"");
        doStatement.setCondition(parseExpression(enclosingScope).toRvalue());
        if (!scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis expected after \"while\" condition");
        if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected at end of \"do\" statement");

        return doStatement;
    }

    /**
     * <pre>
     *   TryStatement :=
     *     'try' Block Catches [ Finally ] |
     *     'try' Block Finally
     *
     *   Catches := CatchClause { CatchClause }
     *
     *   CatchClause := 'catch' '(' FormalParameter ')' Block
     *
     *   Finally := 'finally' Block
     * </pre>
     */
    /*package*/ Java.Statement parseTryStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("try")) this.throwParseException("\"try\" expected");
        Java.TryStatement tryStatement = new Java.TryStatement(
            scanner.read().getLocation(),
            enclosingScope
        );

        tryStatement.setBody(parseBlock(tryStatement));

        while (this.scanner.peek().isKeyword("catch")) {
            Scanner.Location location = this.scanner.read().getLocation();
            if (!this.scanner.read().isOperator("(")) this.throwParseException("Opening parenthesis expected");
            Java.FormalParameter caughtException = this.parseFormalParameter(
                tryStatement
            );
            if (!this.scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis expected");
            tryStatement.addCatchClause(new Java.CatchClause(
                caughtException,              // caughtException
                this.parseBlock(tryStatement) // body
            ));
        }
        if (this.scanner.peek().isKeyword("finally")) {
            this.scanner.read();
            // Important! The enclosing scope is the block enclosing the "try"
            // statement, not the "try" statement itself! Otherwise, a "return"
            // in the "finally" clause would lead to infinite recursion!
            tryStatement.setFinally(this.parseBlock(tryStatement));
        }
        if (
            tryStatement.getCatchClauses().size() == 0 &&
            tryStatement.getFinally() == null
        ) this.throwParseException("\"try\" statement must have at least one \"catch\" clause or a \"finally\" clause");
        return tryStatement;
    }

    /**
     * <pre>
     *   SwitchStatement :=
     *     'switch' '(' Expression ')' '{' { SwitchLabels BlockStatements } '}'
     *
     *   SwitchLabels := SwitchLabels { SwitchLabels }
     *
     *   SwitchLabel := 'case' Expression ':' | 'default' ':'
     * </pre>
     */
    /*package*/ Java.Statement parseSwitchStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("switch")) this.throwParseException("\"switch\" expected");
        Java.SwitchStatement switchStatement = new Java.SwitchStatement(
            scanner.read().getLocation(),
            enclosingScope
        );

        if (!scanner.read().isOperator("(")) this.throwParseException("Opening parenthesis expected");
        switchStatement.setCondition(parseExpression(switchStatement).toRvalue());
        if (!scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis expected");

        if (!scanner.read().isOperator("{")) this.throwParseException("\"{\" expected");
        Java.Block block = new Java.Block(this.scanner.peek().getLocation(), switchStatement);
        while (!this.scanner.peek().isOperator("}")) {
            Java.SwitchBlockStatementGroup sbsg = new Java.SwitchBlockStatementGroup(this.scanner.peek().getLocation());
            do {
                if (this.scanner.peek().isKeyword("case")) {
                    this.scanner.read();
                    sbsg.addSwitchLabel(this.parseExpression((Java.Scope) block).toRvalue());
                } else
                if (this.scanner.peek().isKeyword("default")) {
                    this.scanner.read();
                    sbsg.addDefaultSwitchLabel();
                } else {
                    this.throwParseException("\"case\" or \"default\" expected");
                }
                if (!scanner.read().isOperator(":")) this.throwParseException("Colon expected");
            } while (this.scanner.peek().isKeyword(new String[] { "case", "default" }));

            sbsg.setBlockStatements(this.parseBlockStatements(block));
            switchStatement.addSwitchBlockStatementGroup(sbsg);
        }
        this.scanner.read();
        return switchStatement;
    }

    /**
     * <pre>
     *   ReturnStatement := 'return' [ Expression ] ';'
     * </pre>
     */
    /*package*/ Java.Statement parseReturnStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("return")) this.throwParseException("\"return\" expected");
        Scanner.Location location = scanner.read().getLocation();
        Java.Rvalue returnValue = scanner.peek().isOperator(";") ? null : parseExpression(enclosingScope).toRvalue();
        if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected at end of \"return\" statement");
        return new Java.ReturnStatement(location, enclosingScope, returnValue);
    }

    /**
     * <pre>
     *   ThrowStatement := 'throw' Expression ';'
     * </pre>
     */
    /*package*/ Java.Statement parseThrowStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("throw")) this.throwParseException("\"throw\" expected");
        Scanner.Location location = scanner.read().getLocation();
        final Java.Rvalue expression = parseExpression(enclosingScope).toRvalue();
        if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected");

        return new Java.ThrowStatement(
            location,
            enclosingScope,
            expression
        );
    }

    /**
     * <pre>
     *   BreakStatement := 'break' [ Identifier ] ';'
     * </pre>
     */
    /*package*/ Java.Statement parseBreakStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("break")) this.throwParseException("\"break\" expected");
        Scanner.Location location = scanner.read().getLocation();
        String optionalLabel = null;
        if (scanner.peek().isIdentifier()) optionalLabel = scanner.read().getIdentifier();
        if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected at end of \"break\" statement");
        return new Java.BreakStatement(location, enclosingScope, optionalLabel);
    }

    /**
     * <pre>
     *   ContinueStatement := 'continue' [ Identifier ] ';'
     * </pre>
     */
    /*package*/ Java.Statement parseContinueStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isKeyword("continue")) this.throwParseException("\"continue\" expected");
        Scanner.Location location = scanner.read().getLocation();
        String optionalLabel = null;
        if (scanner.peek().isIdentifier()) optionalLabel = scanner.read().getIdentifier();
        if (!scanner.read().isOperator(";")) this.throwParseException("Semicolon expected at end of \"continue\" statement");
        return new Java.ContinueStatement(location, enclosingScope, optionalLabel);
    }

    /**
     * <pre>
     *   StatementExpressionList := ExpressionStatement { ',' ExpressionStatement }
     * </pre>
     */
    /*package*/ Java.BlockStatement parseStatementExpressionList(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        Java.BlockStatement es = new Java.ExpressionStatement(
            parseExpression(enclosingScope).toRvalue(), // rvalue
            enclosingScope                              // enclosingScope
        );
        if (!scanner.peek().isOperator(",")) return es;
        scanner.read();

        Vector v = new Vector();
        v.add(es);
        for (;;) {
            v.add(new Java.ExpressionStatement(
                parseExpression(enclosingScope).toRvalue(), // rvalue
                enclosingScope                              // enclosingScope
            ));
            if (!scanner.peek().isOperator(",")) break;
            scanner.read();
        }

        Java.Block b = new Java.Block(es.getLocation(), enclosingScope);
        b.addStatements(v);
        return b;
    }

    /**
     * <pre>
     *   Type := (
     *     'byte' | 'short' | 'char' | 'int' | 'long' |
     *     'float' | 'double' | 'boolean' |
     *     QualifiedIdentifier
     *   ) { '[' ']' }
     * </pre>
     */
    /*package*/ Java.Type parseType(
        Java.Scope scope
    ) throws ParseException, Scanner.ScanException, IOException {
        Scanner.Token t = scanner.peek();
        int bt = -1;
        if (t.isKeyword("byte"   )) { bt = Java.BasicType.BYTE   ; } else
        if (t.isKeyword("short"  )) { bt = Java.BasicType.SHORT  ; } else
        if (t.isKeyword("char"   )) { bt = Java.BasicType.CHAR   ; } else
        if (t.isKeyword("int"    )) { bt = Java.BasicType.INT    ; } else
        if (t.isKeyword("long"   )) { bt = Java.BasicType.LONG   ; } else
        if (t.isKeyword("float"  )) { bt = Java.BasicType.FLOAT  ; } else
        if (t.isKeyword("double" )) { bt = Java.BasicType.DOUBLE ; } else
        if (t.isKeyword("boolean")) { bt = Java.BasicType.BOOLEAN; }
        Java.Type res;
        if (bt != -1) {
            res = new Java.BasicType(t.getLocation(), bt);
            scanner.read();
        } else {
            res = new Java.ReferenceType(
                this.scanner.peek().getLocation(),
                scope,
                parseQualifiedIdentifier()
            );
        }
        for (int i = parseBracketsOpt(); i > 0; --i) res = new Java.ArrayType(res);
        return res;
    }

    /**
     * <pre>
     *   ReferenceType := QualifiedIdentifier
     * </pre>
     */
    /*package*/ Java.ReferenceType parseReferenceType(
        Java.Scope scope
    ) throws ParseException, Scanner.ScanException, IOException {
        return new Java.ReferenceType(
            scanner.peek().getLocation(),
            scope,
            parseQualifiedIdentifier()
        );
    }

    /**
     * <pre>
     *   ReferenceTypeList := ReferenceType { ',' ReferenceType }
     * </pre>
     */
    /*package*/ Java.ReferenceType[] parseReferenceTypeList(
        Java.Scope scope
    ) throws ParseException, Scanner.ScanException, IOException {
        Vector v = new Vector();
        v.add(parseReferenceType(scope));
        while (scanner.peek().isOperator(",")) {
            scanner.read();
            v.add(parseReferenceType(scope));
        }
        return (Java.ReferenceType[]) v.toArray(new Java.ReferenceType[v.size()]);
    }

    /**
     * <pre>
     *   Expression := AssignmentExpression
     * </pre>
     */
    /*package*/ Java.Atom parseExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        return parseAssignmentExpression(enclosingScope);
    }

    /**
     * <pre>
     *   AssignmentExpression :=
     *     ConditionalExpression [ AssignmentOperator AssignmentExpression ]
     *
     *   AssignmentOperator :=
     *     '=' | '*=' | '/=' | '%=' | '+=' | '-=' | '<<=' |
     *     '>>=' | '>>>=' | '&=' | '^=' | '|='
     * </pre>
     */
    /*package*/ Java.Atom parseAssignmentExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseConditionalExpression(enclosingScope);
        if (scanner.peek().isOperator(new String[] { "=", "+=", "-=", "*=", "/=", "&=", "|=", "^=", "%=", "<<=", ">>=", ">>>=" })) {
            String operator = scanner.peek().getOperator();
            Scanner.Location location = scanner.read().getLocation();
            final Java.Lvalue lhs = a.toLvalue();
            final Java.Rvalue rhs = parseAssignmentExpression(enclosingScope).toRvalue();
            return new Java.Assignment(location, lhs, operator, rhs);
        } else
        {
            return a;
        }
    }

    /**
     * <pre>
     *   ConditionalExpression :=
     *     ConditionalOrExpression [ '?' Expression ':' ConditionalExpression ]
     * </pre>
     */
    /*package*/ Java.Atom parseConditionalExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseConditionalOrExpression(enclosingScope);
        if (!scanner.peek().isOperator("?")) return a;
        Scanner.Location location = scanner.read().getLocation();

        Java.Rvalue lhs = a.toRvalue();
        Java.Rvalue mhs = parseExpression(enclosingScope).toRvalue();
        if (!scanner.read().isOperator(":")) this.throwParseException("\":\" expected");
        Java.Rvalue rhs = parseConditionalExpression(enclosingScope).toRvalue();
        return new Java.ConditionalExpression(location, lhs, mhs, rhs);
    }

    /**
     * <pre>
     *   ConditionalOrExpression :=
     *     ConditionalAndExpression { '||' ConditionalAndExpression ]
     * </pre>
     */
    /*package*/ Java.Atom parseConditionalOrExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseConditionalAndExpression(enclosingScope);
        while (scanner.peek().isOperator("||")) {
            a = new Java.BinaryOperation(
                scanner.read().getLocation(),
                a.toRvalue(),
                "||",
                parseConditionalAndExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   ConditionalAndExpression :=
     *     InclusiveOrExpression { '&&' InclusiveOrExpression }
     * </pre>
     */
    /*package*/ Java.Atom parseConditionalAndExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseInclusiveOrExpression(enclosingScope);
        while (scanner.peek().isOperator("&&")) {
            a = new Java.BinaryOperation(
                scanner.read().getLocation(),
                a.toRvalue(),
                "&&",
                parseInclusiveOrExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   InclusiveOrExpression :=
     *     ExclusiveOrExpression { '|' ExclusiveOrExpression }
     * </pre>
     */
    /*package*/ Java.Atom parseInclusiveOrExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseExclusiveOrExpression(enclosingScope);
        while (scanner.peek().isOperator("|")) {
            a = new Java.BinaryOperation(
                scanner.read().getLocation(),
                a.toRvalue(),
                "|",
                parseExclusiveOrExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   ExclusiveOrExpression :=
     *     AndExpression { '^' AndExpression }
     * </pre>
     */
    /*package*/ Java.Atom parseExclusiveOrExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseAndExpression(enclosingScope);
        while (scanner.peek().isOperator("^")) {
            a = new Java.BinaryOperation(
                scanner.read().getLocation(),
                a.toRvalue(),
                "^",
                parseAndExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   AndExpression :=
     *     EqualityExpression { '&' EqualityExpression }
     * </pre>
     */
    /*package*/ Java.Atom parseAndExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseEqualityExpression(enclosingScope);
        while (scanner.peek().isOperator("&")) {
            a = new Java.BinaryOperation(
                scanner.read().getLocation(),
                a.toRvalue(),
                "&",
                parseEqualityExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   EqualityExpression :=
     *     RelationalExpression { ( '==' | '!=' ) RelationalExpression }
     * </pre>
     */
    /*package*/ Java.Atom parseEqualityExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseRelationalExpression(enclosingScope);

        while (scanner.peek().isOperator(new String[] { "==", "!=" })) {
            a = new Java.BinaryOperation(
                scanner.peek().getLocation(),
                a.toRvalue(),
                scanner.read().getOperator(),
                parseRelationalExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   RelationalExpression :=
     *     ShiftExpression {
     *       ( ( '<' | '>' | '<=' | '>=' ) ShiftExpression ) |
     *       ( 'instanceof' ReferenceType )
     *     }
     * </pre>
     */
    /*package*/ Java.Atom parseRelationalExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseShiftExpression(enclosingScope);

        for (;;) {
            if (scanner.peek().isKeyword("instanceof")) {
                a = new Java.Instanceof(
                    scanner.read().getLocation(),
                    a.toRvalue(),
                    this.parseReferenceType(enclosingScope)
                );
            } else
            if (scanner.peek().isOperator(new String[] { "<", ">", "<=", ">=" })) {
                a = new Java.BinaryOperation(
                    scanner.peek().getLocation(),
                    a.toRvalue(),
                    scanner.read().getOperator(),
                    parseShiftExpression(enclosingScope).toRvalue()
                );
            } else {
                break;
            }
        }
        return a;
    }

    /**
     * <pre>
     *   ShiftExpression :=
     *     AdditiveExpression { ( '<<' | '>>' | '>>>' ) AdditiveExpression }
     * </pre>
     */
    /*package*/ Java.Atom parseShiftExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseAdditiveExpression(enclosingScope);

        while (scanner.peek().isOperator(new String[] { "<<", ">>", ">>>" })) {
            a = new Java.BinaryOperation(
                scanner.peek().getLocation(),
                a.toRvalue(),
                scanner.read().getOperator(),
                parseAdditiveExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   AdditiveExpression :=
     *     MultiplicativeExpression { ( '+' | '-' ) MultiplicativeExpression }
     * </pre>
     */
    /*package*/ Java.Atom parseAdditiveExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Java.Atom a = parseMultiplicativeExpression(enclosingScope);

        while (scanner.peek().isOperator(new String[] { "+", "-" })) {
            a = new Java.BinaryOperation(
                scanner.peek().getLocation(),
                a.toRvalue(),
                scanner.read().getOperator(),
                parseMultiplicativeExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   MultiplicativeExpression :=
     *     UnaryExpression { ( '*' | '/' | '%' ) UnaryExpression }
     * </pre>
     */
    /*package*/ Java.Atom parseMultiplicativeExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        Java.Atom a = parseUnaryExpression(enclosingScope);

        while (scanner.peek().isOperator(new String[] { "*", "/", "%" })) {
            a = new Java.BinaryOperation(
                scanner.peek().getLocation(),
                a.toRvalue(),
                scanner.read().getOperator(),
                parseUnaryExpression(enclosingScope).toRvalue()
            );
        }
        return a;
    }

    /**
     * <pre>
     *   UnaryExpression :=
     *     { PrefixOperator } Primary { Selector } { PostfixOperator }
     *
     *   PrefixOperator := '++' | '--' | '+' | '-' | '~' | '!'
     *
     *   PostfixOperator := '++' | '--'
     * </pre>
     */
    /*package*/ Java.Atom parseUnaryExpression(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (scanner.peek().isOperator(new String[] { "++", "--" })) {
            return new Java.Crement(
                scanner.peek().getLocation(),                   // location
                scanner.read().getOperator(),                   // operator
                parseUnaryExpression(enclosingScope).toLvalue() // operand
            );
        }

        if (scanner.peek().isOperator(new String[] { "+", "-", "~", "!" })) {
            return new Java.UnaryOperation(
                scanner.peek().getLocation(),
                scanner.read().getOperator(),
                parseUnaryExpression(enclosingScope).toRvalue()
            );
        }

        Java.Atom a = parsePrimary(enclosingScope);

        while (this.scanner.peek().isOperator(new String[] { ".", "[" })) {
            a = this.parseSelector(enclosingScope, a);
        }

        while (scanner.peek().isOperator(new String[] { "++", "--" })) {
            a = new Java.Crement(
                scanner.peek().getLocation(), // location
                a.toLvalue(),                 // operand
                scanner.read().getOperator()  // operator
            );
        }

        return a;
    }

    /**
     * <pre>
     *   Primary :=
     *     '(' PrimitiveType { '[]' } ')' UnaryExpression | // CastExpression 15.16
     *     '(' Expression ')' UnaryExpression |    // CastExpression 15.16
     *     '(' Expression ')' |                    // ParenthesizedExpression 15.8.5
     *     Literal |                               // Literal 15.8.1
     *     Name |                                  // AmbiguousName
     *     Name Arguments |                        // MethodInvocation
     *     Name '[]' { '[]' } |                    // ArrayType 10.1
     *     Name '[]' { '[]' } '.' 'class' |        // ClassLiteral 15.8.2
     *     'this' |                                // This 15.8.3
     *     'super' '.' Identifier |                // SuperclassFieldAccess 15.11.2
     *     'super' '.' Identifier Arguments |      // SuperclassMethodInvocation 15.12.4.9
     *     'new' ReferenceType Arguments [ ClassBody ] | // ClassInstanceCreationExpression 15.9
     *     'new' Type DimExprs { '[]' } |          // ArrayCreationExpression 15.10
     *     'new' ArrayType ArrayInitializer |      // ArrayInitializer 10.6
     *     BasicType { '[]' } |                    // Type
     *     BasicType { '[]' } '.' 'class' |        // ClassLiteral 15.8.2
     *     'void' '.' 'class'                      // ClassLiteral 15.8.2
     * </pre>
     */
    /*package*/ Java.Atom parsePrimary(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (scanner.peek().isOperator("(")) {
            scanner.read();
            if (scanner.peek().isKeyword(new String[] { "boolean", "char", "byte", "short", "int", "long", "float", "double", })) {
                // '(' PrimitiveType { '[]' } ')' UnaryExpression
                Java.Type type = this.parseType(enclosingScope);
                int brackets = this.parseBracketsOpt();
                if (!scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis expected");
                for (int i = 0; i < brackets; ++i) type = new Java.ArrayType(type);
                return new Java.Cast(
                    this.scanner.peek().getLocation(),
                    type,
                    this.parseUnaryExpression(enclosingScope).toRvalue()
                );
            }
            Java.Atom a = parseExpression(enclosingScope);
            if (!scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis expected");

            if (
                scanner.peek().isLiteral() ||
                scanner.peek().isIdentifier() ||
                scanner.peek().isOperator(new String[] { "(", "~", "!", })
            ) {
                // '(' Expression ')' UnaryExpression
                return new Java.Cast(
                    this.scanner.peek().getLocation(),
                    a.toType(),
                    this.parseUnaryExpression(enclosingScope).toRvalue()
                );
            }

            // '(' Expression ')'
            return a;
        }

        if (scanner.peek().isLiteral()) {
            // Literal
            return parseLiteral();
        }

        if (this.scanner.peek().isIdentifier()) {
            Scanner.Location location = this.scanner.peek().getLocation();
            String[] qi = this.parseQualifiedIdentifier();
            if (this.scanner.peek().isOperator("(")) {
                // Name Arguments
                return new Java.MethodInvocation(
                    this.scanner.peek().getLocation(),              // location
                    enclosingScope,                                 // enclosingScope
                    qi.length == 1 ? null : new Java.AmbiguousName( // optionalTarget
                        location,       // location
                        enclosingScope, // scope
                        qi,             // identifiers
                        qi.length - 1   // n
                    ),
                    qi[qi.length - 1],                              // methodName
                    this.parseArguments(enclosingScope)             // arguments
                );
            }
            if (
                this.scanner.peek().isOperator("[") &&
                this.scanner.peekNextButOne().isOperator("]")
            ) {
                // Name '[]' { '[]' }
                // Name '[]' { '[]' } '.' 'class'
                int brackets = this.parseBracketsOpt();
                if (
                    this.scanner.peek().isOperator(".") &&
                    this.scanner.peekNextButOne().isKeyword("]")
                ) {
                    this.throwParseException("Class literals NYI");
                } else {
                    Java.Type res = new Java.ReferenceType(
                        location, // location
                        enclosingScope, // scope
                        qi // identifiers
                    );
                    for (int i = 0; i < brackets; ++i) res = new Java.ArrayType(res);
                    return res;
                }
            }
            // Name
            return new Java.AmbiguousName(
                this.scanner.peek().getLocation(), // location
                enclosingScope,                    // scope
                qi                                 // identifiers
            );
        }

        // 'this'
        if (this.scanner.peek().isKeyword("this")) this.throwParseException("\"this\" NYI");

        if (this.scanner.peek().isKeyword("super")) {
            this.scanner.read();
            if (!this.scanner.read().isOperator(".")) this.throwParseException("\".\" expected after \"super\"");
            if (!this.scanner.peek().isIdentifier()) this.throwParseException("Identifier expected after \"super\"");
            String name = this.scanner.read().getIdentifier();
            Scanner.Location location = this.scanner.peek().getLocation();
            if (this.scanner.peek().isOperator("(")) {
                // 'super' '.' Identifier Arguments
                return new Java.SuperclassMethodInvocation(
                    this.scanner.peek().getLocation(),  // location
                    enclosingScope,                     // enclosingScope
                    name,                               // methodName
                    this.parseArguments(enclosingScope) // arguments
                );
            } else {
                // 'super' '.' Identifier
                this.throwParseException("Superclass field access NYI");
            }
        }
        if (scanner.peek().isKeyword("new")) {
            Scanner.Location location = scanner.read().getLocation();
            Java.Type type = this.parseType(enclosingScope);
            if (type instanceof Java.ArrayType) {
                // 'new' ArrayType ArrayInitializer
                return this.parseArrayInitializer(
                    enclosingScope,
                    (Java.ArrayType) type
                );
            }
            if (
                type instanceof Java.ReferenceType &&
                scanner.peek().isOperator("(")
            ) {
                // 'new' ReferenceType Arguments [ ClassBody ]
                Java.Rvalue[] arguments = this.parseArguments(enclosingScope);
                if (this.scanner.peek().isOperator("{")) this.throwParseException("Anonymous classes NYI");
                return new Java.NewClassInstance(
                    location,                  // location
                    enclosingScope,            // scope
                    (Java.ReferenceType) type, // referenceType
                    arguments                  // arguments
                );
            }
            // 'new' Type DimExprs { '[]' }
            return new Java.NewArray(
                location,                      // location
                type,                          // type
                parseDimExprs(enclosingScope), // dimExprs
                this.parseBracketsOpt()        // dims
            );
        }
        if (scanner.peek().isKeyword(new String[] { "boolean", "char", "byte", "short", "int", "long", "float", "double", })) {
            Java.Type res = this.parseType(enclosingScope);
            int brackets = this.parseBracketsOpt();
            for (int i = 0; i < brackets; ++i) res = new Java.ArrayType(res);
            if (
                this.scanner.peek().isOperator(".") &&
                this.scanner.peekNextButOne().isKeyword("class")
            ) {
                // BasicType { '[]' } '.' 'class'
                this.throwParseException("Class literals NYI");
            }
            // BasicType { '[]' }
            return res;
        }
        if (this.scanner.peek().isKeyword("void")) {
            if (
                this.scanner.peek().isOperator(".") &&
                this.scanner.peekNextButOne().isKeyword("class")
            ) {
                // 'void' '.' 'class'
                this.throwParseException("Class literals NYI");
            }
            this.throwParseException("\"void\" encountered in wrong context");
        }
        this.throwParseException("Unexpected token \"" + this.scanner.peek() + "\" in primary");
        /* NEVER REACHED */ return null;
    }

    /**
     * <pre>
     *   Selector :=
     *     '.' Identifier |                       // FieldAccess 15.11.1
     *     '.' Identifier Arguments |             // MethodInvocation
     *     '.' 'this'                             // QualifiedThis 15.8.4
     *     '.' 'super' '.' Identifier |           // SuperclassFieldReference
     *     '.' 'super' '.' Identifier Arguments | // SuperclassMethodInvocation 15.12.4.9
     *    '.' 'new' Identifier Arguments [ ClassBody ] | // QualifiedClassInstanceCreationExpression  15.9
     *    '[' Expression ']'                     // ArrayAccessExpression 15.13
     * </pre>
     */
    /*package*/ Java.Atom parseSelector(
        Java.Scope enclosingScope,
        Java.Atom  atom
    ) throws ParseException, Scanner.ScanException, IOException {
        if (this.scanner.peek().isOperator(".")) {
            this.scanner.read();
            if (this.scanner.peek().isIdentifier()) {
                String identifier = this.scanner.read().getIdentifier();
                if (this.scanner.peek().isOperator("(")) {
                    // '.' Identifier Arguments
                    return new Java.MethodInvocation(
                        this.scanner.peek().getLocation(),  // location
                        enclosingScope,                     // enclosingScope
                        atom.toRvalue(),                    // optionalTarget
                        identifier,                         // methodName
                        this.parseArguments(enclosingScope) // arguments
                    );
                }
                // '.' Identifier
                return new Java.FieldAccessExpression(
                    this.scanner.peek().getLocation(), // location
                    atom.toRvalue(),                   // lhs
                    identifier                         // fieldName
                );
            }
            if (this.scanner.peek().isKeyword("this")) {
                // '.' 'this'
                this.throwParseException("Qualified \"this\" NYI");
            }
            if (this.scanner.peek().isKeyword("super")) {
                // '.' 'super' '.' Identifier
                // '.' 'super' '.' Identifier Arguments
                this.throwParseException("Superclass field access or nethod invocation NYI");
            }
            if (this.scanner.peek().isKeyword("new")) {
                // '.' 'new' Identifier Arguments [ ClassBody ]
                this.throwParseException("Qualified \"new\" NYI");
            }
            this.throwParseException("Unexpected selector \"" + this.scanner.peek() + "\" after \".\"");
        }
        if (this.scanner.peek().isOperator("[")) {
            Scanner.Location location = this.scanner.read().getLocation();
            Java.Rvalue index = this.parseExpression(enclosingScope).toRvalue();
            if (!this.scanner.read().isOperator("]")) this.throwParseException("\"]\" expected");
            return new Java.ArrayAccessExpression(
                location,        // location
                atom.toRvalue(), // lhs
                index            // index
            );
        }
        this.throwParseException("Unexpected token \"" + this.scanner.peek() + "\" in selector");
        /* NEVER REACHED */ return null;
    }

    /**
     * <pre>
     *   DimExprs := { '[' Expression ']' }
     * </pre>
     */
    /*package*/ Java.Rvalue[] parseDimExprs(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException  {
        Vector v = new Vector();
        while (
            scanner.peek().isOperator("[") &&
            !scanner.peekNextButOne().isOperator("]")
        ) {
            scanner.read();
            v.add(this.parseExpression(enclosingScope));
            if (!scanner.read().isOperator("]")) this.throwParseException("\"]\" expected");
        }
        return (Java.Rvalue[]) v.toArray(new Java.Rvalue[v.size()]);
    }

    /**
     * <pre>
     *   Arguments := '(' [ ArgumentList ] ')'
     * </pre>
     */
    /*package*/ Java.Rvalue[] parseArguments(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.read().isOperator("(")) this.throwParseException("Opening parenthesis expected");
        if (scanner.peek().isOperator(")")) {
            scanner.read();
            return new Java.Rvalue[0];
        }
        Java.Rvalue[] arguments = parseArgumentList(enclosingScope);
        if (!scanner.read().isOperator(")")) this.throwParseException("Closing parenthesis after argument list expected");
        return arguments;
    }

    /**
     * <pre>
     *   ArgumentList := Expression { ',' Expression }
     * </pre>
     */
    /*package*/ Java.Rvalue[] parseArgumentList(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        Vector v = new Vector();
        for (;;) {
            v.add(parseExpression(enclosingScope).toRvalue());
            if (!scanner.peek().isOperator(",")) break;
            scanner.read();
        }
        return (Java.Rvalue[]) v.toArray(new Java.Rvalue[v.size()]);
    }

    /*package*/ Java.Atom parseLiteral() throws ParseException, Scanner.ScanException, IOException {
        if (!scanner.peek().isLiteral()) this.throwParseException("Literal expected");
        return new Java.Literal(
            scanner.peek().getLocation(),
            scanner.read().getLiteralValue()
        );
    }

    /**
     * <pre>
     *   ExpressionStatement := Expression
     * </pre>
     */
    /*package*/ Java.Statement parseExpressionStatement(
        Java.Scope enclosingScope
    ) throws ParseException, Scanner.ScanException, IOException {
        final Java.Rvalue rvalue = parseExpression(enclosingScope).toRvalue();
        return new Java.ExpressionStatement(rvalue, enclosingScope);
    }

    private Scanner scanner;

    /**
     * An exception that reflects an error during parsing.
     *
     * This exception is associated with a particular {@link Scanner.Location
     * Location} in the source code.
     */
    public static class ParseException extends Exception {
        ParseException(String message, Scanner.Location location) {
            super(message);
            this.location = location;
        }

        /**
         * Returns the message specified at creation time, preceeded
         * with nicely formatted location information.
         */
        public String getMessage() {
            return location.toString() + ": " + super.getMessage();
        }
        final Scanner.Location location;
    }

    /**
     * Convenience method for throwing a ParseException.
     */
    private final void throwParseException(String message) throws ParseException {
        throw new ParseException(message, scanner.peek().getLocation());
    }

    private static String join(String[] sa, String separator) {
        if (sa == null) return ("(null)");
        if (sa.length == 0) return ("(zero length array)");
        StringBuffer sb = new StringBuffer(sa[0]);
        for (int i = 1; i < sa.length; ++i) {
            sb.append(separator);
            sb.append(sa[i]);
        }
        return sb.toString();
    }
}
