
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.lang.reflect.*;
import java.util.*;

/**
 * An implementation of {@link IClass} using "java.lang.reflect".
 */
public class ReflectionIClass extends IClass {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/ReflectionIClass.java,v 1.6 2002/05/11 22:01:41 arno Exp $";

    private ReflectionIClass(Class clazz) { this.clazz = clazz; }

    public IConstructor[] getDeclaredIConstructors() {
        Constructor[] constructors = clazz.getDeclaredConstructors();
        IConstructor[] res = new IConstructor[constructors.length];
        for (int i = 0; i < constructors.length; ++i) {
            res[i] = new ReflectionIConstructor(constructors[i]);
        }
        return res;
    }
    public IMethod[] getDeclaredIMethods() {
        Method[] methods = clazz.getDeclaredMethods();
        IMethod[] res = new IMethod[methods.length];
        for (int i = 0; i < methods.length; ++i) {
            res[i] = new ReflectionIMethod(methods[i]);
        }
        return res;
    }
    public IField[] getDeclaredIFields() {
        Field[] fields = clazz.getDeclaredFields();
        IField[] res = new IField[fields.length];
        for (int i = 0; i < fields.length; ++i) {
            res[i] = new ReflectionIField(fields[i]);
        }
        return res;
    }
    public IClass[] getDeclaredIClasses() {
        Class[] classes = clazz.getDeclaredClasses();
        IClass[] res = new IClass[classes.length];
        for (int i = 0; i < classes.length; ++i) {
            res[i] = ReflectionIClass.fromClass(classes[i]);
        }
        return res;
    }
    public IClass getSuperclass() {
        Class superclass = clazz.getSuperclass();
        return superclass == null ? null : ReflectionIClass.fromClass(superclass);
    }
    public IClass[] getImplementedInterfaces() {
        Class[] iis = clazz.getInterfaces();
        IClass[] res = new IClass[iis.length];
        for (int i = 0; i < iis.length; ++i) {
            res[i] = ReflectionIClass.fromClass(iis[i]);
        }
        return res;
    }
    public String getDescriptor() {
        return Descriptor.fromClassName(clazz.getName());
    }
    public boolean isArray() {
        return clazz.isArray();
    }
    public IClass getComponentType() {
        Class componentType = clazz.getComponentType();
        return componentType == null ? null : ReflectionIClass.fromClass(componentType);
    }
    public boolean isPrimitive() {
        return clazz.isPrimitive();
    }
    public boolean isPrimitiveNumeric() {
        return (
            this.clazz == Byte.TYPE    ||
            this.clazz == Short.TYPE   ||
            this.clazz == Integer.TYPE ||
            this.clazz == Long.TYPE    ||
            this.clazz == Float.TYPE   ||
            this.clazz == Double.TYPE
        );
    }

    public static IClass fromClass(Class clazz) {
        IClass iClass = (IClass) classToIClass.get(clazz);
        if (iClass == null) {
            iClass = new ReflectionIClass(clazz);
            classToIClass.put(clazz, iClass);
        }
        return iClass;
    }
    public String toString() { return clazz.toString(); }

    private final Class          clazz;
    private final static HashMap classToIClass = new HashMap();

    private class ReflectionIConstructor extends IConstructor {
        public ReflectionIConstructor(Constructor constructor) { this.constructor = constructor; }

        // Implement IMember.
        public int getAccess() {
            int mod = this.constructor.getModifiers();
            return (
                Modifier.isPrivate(mod)   ? IClass.PRIVATE   :
                Modifier.isProtected(mod) ? IClass.PROTECTED :
                Modifier.isPublic(mod)    ? IClass.PUBLIC    :
                IClass.PACKAGE
            );
        }

        // Implement "IConstructor".
        public IClass[] getParameterTypes() {
            Class[] pts = constructor.getParameterTypes();
            IClass[] res = new IClass[pts.length];
            for (int i = 0; i < pts.length; ++i) {
                res[i] = ReflectionIClass.fromClass(pts[i]);
            }
            return res;
        }
        public IClass[] getThrownExceptions() {
            Class[] ets = constructor.getExceptionTypes();
            IClass[] res = new IClass[ets.length];
            for (int i = 0; i < ets.length; ++i) {
                res[i] = ReflectionIClass.fromClass(ets[i]);
            }
            return res;
        }

        final Constructor constructor;
    };
    private class ReflectionIMethod extends IMethod {
        public ReflectionIMethod(Method method) { this.method = method; }

        // Implement IMember.
        public int getAccess() {
            int mod = this.method.getModifiers();
            return (
                Modifier.isPrivate(mod)   ? IClass.PRIVATE   :
                Modifier.isProtected(mod) ? IClass.PROTECTED :
                Modifier.isPublic(mod)    ? IClass.PUBLIC    :
                IClass.PACKAGE
            );
        }

        // Implemnt "IMethod".
        public String getName() { return method.getName(); }
        public IClass[] getParameterTypes() {
            Class[] pts = method.getParameterTypes();
            IClass[] res = new IClass[pts.length];
            for (int i = 0; i < pts.length; ++i) {
                res[i] = ReflectionIClass.fromClass(pts[i]);
            }
            return res;
        }
        public boolean isStatic() {
            return Modifier.isStatic(method.getModifiers());
        }
        public IClass getReturnType() {
            return ReflectionIClass.fromClass(method.getReturnType());
        }
        public IClass[] getThrownExceptions() {
            Class[] ets = method.getExceptionTypes();
            IClass[] res = new IClass[ets.length];
            for (int i = 0; i < ets.length; ++i) {
                res[i] = ReflectionIClass.fromClass(ets[i]);
            }
            return res;
        }

        final Method method;
    };
    private class ReflectionIField extends IField {
        public ReflectionIField(Field field) { this.field = field; }

        // Implement IMember.
        public int getAccess() {
            int mod = this.field.getModifiers();
            return (
                Modifier.isPrivate(mod)   ? IClass.PRIVATE   :
                Modifier.isProtected(mod) ? IClass.PROTECTED :
                Modifier.isPublic(mod)    ? IClass.PUBLIC    :
                IClass.PACKAGE
            );
        }

        // Implement "IField".
        public String getName() { return field.getName(); }
        public boolean isStatic() {
            return Modifier.isStatic(field.getModifiers());
        }
        public IClass getType() {
            return ReflectionIClass.fromClass(field.getType());
        }
        public String toString() {
            return (
                Descriptor.toString(getDeclaringIClass().getDescriptor()) +
                "." +
                getName()
            );
        }

        final Field field;
    }
}
