
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino;

import java.io.*;
import java.util.HashMap;

/**
 * Splits up a character stream into tokens and returns them as
 * {@link java.lang.String String} objects.
 */

public class Scanner {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/Scanner.java,v 1.10 2002/05/21 22:40:00 arno Exp $";

    public Scanner(String fileName) throws ScanException, IOException {
        this(fileName, new FileInputStream(fileName));
    }
    public Scanner(String fileName, InputStream is) throws ScanException, IOException {
        this(fileName, new BackslashUReader(is));
    }
    public Scanner(String fileName, Reader in) throws ScanException, IOException {
        this.fileName = fileName;
        this.in       = in;

        readNextChar();
        nextToken       = internalRead();
        nextButOneToken = internalRead();
    }

    public void close() throws IOException { in.close(); }

    /**
     * Read the next token from the input.
     */
    public Token read() throws ScanException, IOException {
        Token res = nextToken;
        nextToken       = nextButOneToken;
        nextButOneToken = internalRead();
        return res;
    }

    /**
     * Peek the next token, but don't remove it from the input.
     */
    public Token peek() {
        if (DEBUG) System.err.println("peek() => \"" + nextToken + "\"");
        return nextToken;
    }

    /**
     * Peek the next but one token, neither remove the next nor the next
     * but one token from the input.
     */
    public Token peekNextButOne() {
        return nextButOneToken;
    }

    public class Token {
        private Token(
            int    type,
            String value
        ) {
            this.type         = type;
            this.value        = value;
            this.fileName     = Scanner.this.fileName;
            this.lineNumber   = Scanner.this.tokenLineNumber;
            this.columnNumber = Scanner.this.tokenColumnNumber;
        }

        /**
         * Keyword token.
         */
        public boolean isKeyword() { return type == KEYWORD; }
        public boolean isKeyword(String k) { return type == KEYWORD && value == k; }
        public boolean isKeyword(String[] ks) {
            if (type != KEYWORD) return false;
            for (int i = 0; i < ks.length; ++i) {
                if (value == ks[i]) return true;
            }
            return false;
        }
        public String getKeyword() { return value; } // An interned String.

        /**
         * Identifier token.
         */
        public boolean isIdentifier() { return type == IDENTIFIER; }
        public boolean isIdentifier(String id) { return type == IDENTIFIER && value.equals(id); }
        public String getIdentifier() { return value; }

        /**
         * Literal token.
         */
        public boolean isLiteral()          { return type >= STRING_LITERAL && type <= NULL_LITERAL; }
        public boolean isStringLiteral()    { return type == STRING_LITERAL;    }
        public boolean isCharacterLiteral() { return type == CHARACTER_LITERAL; }
        public boolean isNumericLiteral()   { return type >= INTEGER_LITERAL && type <= DOUBLE_LITERAL; }
        public boolean isIntegerLiteral()   { return type == INTEGER_LITERAL;   }
        public boolean isLongLiteral()      { return type == LONG_LITERAL;      }
        public boolean isFloatLiteral()     { return type == FLOAT_LITERAL;     }
        public boolean isDoubleLiteral()    { return type == DOUBLE_LITERAL;    }
        public boolean isBooleanLiteral()   { return type == BOOLEAN_LITERAL;   }
        public boolean isNullLiteral()      { return type == NULL_LITERAL;      }
        public Object getLiteralValue() {
            return (
                type == STRING_LITERAL    ? (Object) value :
                type == CHARACTER_LITERAL ? (Object) new Character(value.charAt(0)) :
                type == INTEGER_LITERAL   ? (Object) new Integer(value) :
                type == LONG_LITERAL      ? (Object) new Long(value) :
                type == FLOAT_LITERAL     ? (Object) new Float(value) :
                type == DOUBLE_LITERAL    ? (Object) new Double(value) :
                type == BOOLEAN_LITERAL   ? (Object) new Boolean(value) :
                type == NULL_LITERAL      ? null :
                null
            );
        }

        /**
         * Operator token.
         */
        public boolean isOperator() { return type == OPERATOR; }
        public boolean isOperator(String o) { // String must be interned.
            return type == OPERATOR && value == o;
        }
        public boolean isOperator(String[] os) { // Strings must be interned.
            if (type != OPERATOR) return false;
            for (int i = 0; i < os.length; ++i) {
                if (value == os[i]) return true;
            }
            return false;
        }
        public String getOperator() { return value; } // An interned String.

        /**
         * "EOF" token.
         */
        public boolean isEOF() { return type == EOF; }

        public Location getLocation() {
            return new Location(fileName, lineNumber, columnNumber);
        }

        public String toString() { return (
            type == STRING_LITERAL    ? '"' + value + '"' :
            type == CHARACTER_LITERAL ? '\'' + value + '\'' :
            type == LONG_LITERAL      ? value + 'L' :
            type == FLOAT_LITERAL     ? value + 'F' :
            type == DOUBLE_LITERAL    ? value + 'D' :
            value
        ); }

        private final int    type;
        private final String value;
        private final String fileName;
        private final int    lineNumber;
        private final int    columnNumber;

        // Values for "type".
        // Notice: Order matters!
        private static final int KEYWORD           = 0;
        private static final int IDENTIFIER        = 1;
        private static final int STRING_LITERAL    = 2;
        private static final int CHARACTER_LITERAL = 3;
        private static final int INTEGER_LITERAL   = 4;
        private static final int LONG_LITERAL      = 5;
        private static final int FLOAT_LITERAL     = 6;
        private static final int DOUBLE_LITERAL    = 7;
        private static final int BOOLEAN_LITERAL   = 8;
        private static final int NULL_LITERAL      = 9;
        private static final int OPERATOR          = 10;
        private static final int EOF               = 11;
    }

/*
    public boolean nextToken(String s) throws ScanException, IOException {
        if (!s.equals(nextToken)) return false;
        readNextToken();
        return true;
    }
    public String nextToken(String[] sa) throws ScanException, IOException {
        for (int i = 0; i < sa.length; ++i) {
            if (nextToken.equals(sa[i])) {
                readNextToken();
                return sa[i];
            }
        }
        return null;
    }
    public String nextToken(int type) throws ScanException, IOException {
        if (!peekToken(type)) return null;
        String res = nextToken;
        readNextToken();
        return res;
    }

    public boolean peekToken(String s) throws ScanException, IOException {
        return nextToken.equals(s);
    }
    public boolean peekToken(String[] sa) throws ScanException, IOException {
        for (int i = 0; i < sa.length; ++i) if (nextToken.equals(sa[i])) return true;
        return false;
    }
    public boolean peekToken(int type) throws ScanException, IOException {
        char firstChar = nextToken.charAt(0);
        char lastChar  = nextToken.charAt(nextToken.length() - 1);

        switch (type) {
        case IDENTIFIER:
            return Character.isJavaIdentifierStart(firstChar);
        case LITERAL:
            return firstChar == '\'' || firstChar == '"' || Character.isDigit(firstChar) || nextToken.equals("true") || nextToken.equals("false");;
        case NUMERIC_LITERAL:
            return Character.isDigit(firstChar);
        case INT_LITERAL:
            return Character.isDigit(firstChar) && Character.isDigit(lastChar);
        case LONG_LITERAL:
            return Character.isDigit(firstChar) && lastChar == 'L';
        case FLOAT_LITERAL:
            return Character.isDigit(firstChar) && lastChar == 'F';
        case DOUBLE_LITERAL:
            return Character.isDigit(firstChar) && lastChar == 'D';
        case STRING_LITERAL:
            return firstChar == '"';
        case CHARACTER_LITERAL:
            return firstChar == '\'';
        case OPERATOR:
            return !Character.isJavaIdentifierStart(firstChar) && !Character.isDigit(firstChar);
        default:
            throw new ScanException("Unknown token type " + type);
        }
    }
*/

    private Token internalRead() throws ScanException, IOException {

        // Skip whitespace and comments.
        for (;;) {

            // Eat whitespace.
            while (Character.isWhitespace((char) nextChar)) {
                readNextChar();
                if (nextChar == -1) { return new Token(Token.EOF, null); }
            }

            // Skip comment.
            if (nextChar != '/') break;
            readNextChar();
            if (nextChar == -1) { return new Token(Token.OPERATOR, "/"); }
            if (nextChar == '/') {
                for (;;) {
                    readNextChar();
                    if (nextChar == -1) { return new Token(Token.EOF, null); }
                    if (nextChar == '\r' || nextChar == '\n') break;
                }
            } else
            if (nextChar == '*') {
                boolean gotStar = false;
                for (;;) {
                    readNextChar();
                    if (nextChar == -1) throw new ScanException("EOF in C-style comment");
                    if (nextChar == '*') {
                        gotStar = true;
                    } else
                    if (gotStar && nextChar == '/') {
                        readNextChar();
                        break;
                    } else {
                        gotStar = false;
                    }
                }
            } else
            {
                return new Token(Token.OPERATOR, "/");
            }
        }

        /*
         * Whitespace and comments are now skipped; "nextChar" is definitely
         * the first character of the token.
         */
        this.tokenLineNumber   = this.nextCharLineNumber;
        this.tokenColumnNumber = this.nextCharColumnNumber;

        if (nextChar == -1) return new Token(Token.EOF, null);

        // Scan identifier.
        if (Character.isJavaIdentifierStart((char) nextChar)) {
            StringBuffer sb = new StringBuffer();
            sb.append((char) nextChar);
            for (;;) {
                readNextChar();
                if (
                    nextChar == -1 ||
                    !Character.isJavaIdentifierPart((char) nextChar)
                ) break;
                sb.append((char) nextChar);
            }
            String s = sb.toString();
            if (s.equals("true") || s.equals("false")) return new Token(Token.BOOLEAN_LITERAL, s);
            if (s.equals("null")) return new Token(Token.NULL_LITERAL, s);
            {
                String v = (String) javaKeywords.get(s);
                if (v != null) return new Token(Token.KEYWORD, v);
            }
            return new Token(Token.IDENTIFIER, s);
        }

        // Scan numeric literal.
        if (Character.isDigit((char) nextChar)) {
            StringBuffer sb = new StringBuffer();
            int state = 0;
            for (;;) {
                switch (state) {

                case 0: // First character.
                    if (nextChar == '0') {
                        state = 6;
                    } else
                    /* if (Character.isDigit((char) nextChar)) */ {
                        sb.append((char) nextChar);
                        state = 1;
                    }
                    break;

                case 1: // Decimal digits.
                    if (Character.isDigit((char) nextChar)) {
                        sb.append((char) nextChar);
                    } else
                    if (nextChar == 'l' || nextChar == 'L') {
                        readNextChar();
                        return new Token(Token.LONG_LITERAL, sb.toString());
                    } else
                    if (nextChar == 'f' || nextChar == 'F') {
                        readNextChar();
                        return new Token(Token.FLOAT_LITERAL, sb.toString());
                    } else
                    if (nextChar == 'd' || nextChar == 'D') {
                        readNextChar();
                        return new Token(Token.DOUBLE_LITERAL, sb.toString());
                    } else
                    if (nextChar == '.') {
                        sb.append('.');
                        state = 2;
                    } else
                    if (nextChar == 'E' || nextChar == 'e') {
                        sb.append('E');
                        state = 3;
                    } else
                    {
                        return new Token(Token.INTEGER_LITERAL, sb.toString());
                    }
                    break;

                case 2: // After decimal point.
                    if (Character.isDigit((char) nextChar)) {
                        sb.append((char) nextChar);
                    } else
                    if (nextChar == 'e' || nextChar == 'E') {
                        sb.append('E');
                        state = 3;
                    } else
                    if (nextChar == 'f' || nextChar == 'F') {
                        readNextChar();
                        return new Token(Token.FLOAT_LITERAL, sb.toString());
                    } else
                    if (nextChar == 'd' || nextChar == 'D') {
                        readNextChar();
                        return new Token(Token.DOUBLE_LITERAL, sb.toString());
                    } else
                    {
                        return new Token(Token.DOUBLE_LITERAL, sb.toString());
                    }
                    break;

                case 3: // Read exponent.
                    if (Character.isDigit((char) nextChar)) {
                        sb.append((char) nextChar);
                        state = 5;
                    } else
                    if (nextChar == '-' || nextChar == '+') {
                        sb.append((char) nextChar);
                        state = 4;
                    } else
                    {
                        throw new ScanException("Exponent missing after \"E\"");
                    }
                    break;

                case 4: // After exponent sign.
                    if (Character.isDigit((char) nextChar)) {
                        sb.append((char) nextChar);
                        state = 5;
                    } else
                    {
                        throw new ScanException("Exponent missing after \"E\" and sign");
                    }
                    break;

                case 5: // After first exponent digit.
                    if (Character.isDigit((char) nextChar)) {
                        sb.append((char) nextChar);
                    } else
                    if (nextChar == 'f' || nextChar == 'F') {
                        readNextChar();
                        return new Token(Token.FLOAT_LITERAL, sb.toString());
                    } else
                    if (nextChar == 'd' || nextChar == 'D') {
                        readNextChar();
                        return new Token(Token.DOUBLE_LITERAL, sb.toString());
                    } else
                    {
                        return new Token(Token.DOUBLE_LITERAL, sb.toString());
                    }
                    break;

                case 6: // After leading zero
                    if ("01234567".indexOf(nextChar) != -1) {
                        sb.append((char) nextChar);
                        state = 7;
                    } else
                    if (nextChar == '.') {
                        sb.append("0.");
                        state = 2;
                    } else
                    if (nextChar == 'x' || nextChar == 'X') {
                        state = 8;
                    } else
                    {
                        return new Token(Token.INTEGER_LITERAL, "0");
                    }
                    break;

                case 7: // In octal literal.
                    if ("01234567".indexOf(nextChar) != -1) {
                        sb.append((char) nextChar);
                    } else
                    if (nextChar == 'l' || nextChar == 'L') {
                        // Octal long literal.
                        readNextChar();
                        return new Token(Token.LONG_LITERAL, Long.toString(Long.parseLong(sb.toString(), 8)));
                    } else
                    {
                        // Octal int literal
                        return new Token(Token.INTEGER_LITERAL, Integer.toString(Integer.parseInt(sb.toString(), 8)));
                    }
                    break;

                case 8: // First hex digit
                    if (Character.digit((char) nextChar, 16) != -1) {
                        sb.append((char) nextChar);
                        state = 9;
                    } else
                    {
                        throw new ScanException("Hex digit expected after \"0x\"");
                    }
                    break;

                case 9:
                    if (Character.digit((char) nextChar, 16) != -1) {
                        sb.append((char) nextChar);
                    } else
                    if (nextChar == 'l' || nextChar == 'L') {
                        // Hex long literal
                        readNextChar();
                        return new Token(Token.LONG_LITERAL, Long.toString(Long.parseLong(sb.toString(), 16)));
                    } else
                    {
                        // Hex long literal
                        return new Token(Token.INTEGER_LITERAL, Integer.toString(Integer.parseInt(sb.toString(), 16)));
                    }
                    break;
                }
                readNextChar();
            }
        }

        // Scan string literal.
        if (nextChar == '"') {
            StringBuffer sb = new StringBuffer("");
            readNextChar();
            if (nextChar == -1) throw new ScanException("EOF in string literal");
            if (nextChar == '\r' || nextChar == '\n') throw new ScanException("Line break in string literal");
            while (nextChar != '"') {
                sb.append(unescapeCharacterLiteral());
            }
            readNextChar();
            return new Token(Token.STRING_LITERAL, sb.toString());
        }

        // Scan character literal.
        if (nextChar == '\'') {
            readNextChar();
            char lit = unescapeCharacterLiteral();
            if (nextChar != '\'') throw new ScanException("Closing single quote missing");
            readNextChar();
            return new Token(Token.CHARACTER_LITERAL, new String(new char[] { lit }));
        }

        // Scan separator / operator.
        {
            String v = (String) javaOperators.get(
                new String(new char[] { (char) nextChar })
            );
            if (v != null) {
                for (;;) {
                    readNextChar();
                    String v2 = (String) javaOperators.get(v + (char) nextChar);
                    if (v2 == null) return new Token(Token.OPERATOR, v);
                    v = v2;
                }
            }
        }

        throw new ScanException("Invalid character input \"" + (char) nextChar + "\" (character code " + (int) nextChar + ")");
    }
    private char unescapeCharacterLiteral() throws ScanException, IOException {
        if (nextChar == -1) throw new ScanException("EOF in character literal");

        if (nextChar != '\\') {
            char res = (char) nextChar;
            readNextChar();
            return res;
        }
        readNextChar();
        int idx = "btnfr".indexOf(nextChar);
        if (idx != -1) {
            char res = "\b\t\n\f\r".charAt(idx);
            readNextChar();
            return res;
        }
        idx = "01234567".indexOf(nextChar);
        if (idx != -1) {
            int code = idx;
            readNextChar();
            idx = "01234567".indexOf(nextChar);
            if (idx == -1) return (char) code;
            code = 8 * code + idx;
            readNextChar();
            idx = "01234567".indexOf(nextChar);
            if (idx == -1) return (char) code;
            code = 8 * code + idx;
            if (code > 255) throw new ScanException("Invalid octal escape");
            readNextChar();
            return (char) code;
        }

        char res = (char) nextChar;
        readNextChar();
        return res;
    }

    // Read one character and store in "nextChar".
    private void readNextChar() throws IOException {
        nextChar = in.read();
        if (nextChar == '\r') {
            ++nextCharLineNumber;
            nextCharColumnNumber = 0;
            crLfPending = true;
        } else
        if (nextChar == '\n') {
            if (crLfPending) {
                crLfPending = false;
            } else {
                ++nextCharLineNumber;
                nextCharColumnNumber = 0;
            }
        } else
        {
            ++nextCharColumnNumber;
        }
//System.out.println("'" + (char) nextChar + "' = " + (int) nextChar);
    }

    /*package*/ static class Location {
        private Location(String fileName, int lineNumber, int columnNumber) {
            this.fileName = fileName;
            this.lineNumber = lineNumber;
            this.columnNumber = columnNumber;
        }
        public String toString() {
            StringBuffer sb = new StringBuffer();
            if (fileName != null) {
                sb.append(fileName);
                sb.append(':');
            }
            sb.append(lineNumber);
            sb.append(':');
            sb.append(columnNumber);
            return sb.toString();
        }
        private final String fileName;
        private final int    lineNumber;
        private final int    columnNumber;
    }

    private static final boolean DEBUG = false;

    private /*final*/ String     fileName;  // null == unspecified
    private /*final*/ Reader     in;
    private int              nextChar  = -1;
    private boolean          crLfPending = false;
    private int              nextCharLineNumber   = 1;
    private int              nextCharColumnNumber = 0;

    private Token nextToken, nextButOneToken;
    private int   tokenLineNumber;
    private int   tokenColumnNumber;

    private static final HashMap javaKeywords = new HashMap();
    static {
        String[] ks = {
            "abstract", "boolean", "break", "byte", "case", "catch", "char",
            "class", "const", "continue", "default", "do", "double", "else",
            "extends", "final", "finally", "float", "for", "goto", "if",
            "implements", "import", "instanceof", "int", "interface", "long",
            "native", "new", "package", "private", "protected", "public",
            "return", "short", "static", "strictfp", "super", "switch",
            "synchronized", "this", "throw", "throws", "transient", "try",
            "void", "volatile", "while"
        };
        for (int i = 0; i < ks.length; ++i) javaKeywords.put(ks[i], ks[i]);
    }
    private static final HashMap javaOperators = new HashMap();
    static {
        String[] ops = {
            // Separators:
            "(", ")", "{", "}", "[", "]", ";", ",", ".",
            // Operators:
            "=",  ">",  "<",  "!",  "~",  "?",  ":",
            "==", "<=", ">=", "!=", "&&", "||", "++", "--",
            "+",  "-",  "*",  "/",  "&",  "|",  "^",  "%",  "<<",  ">>",  ">>>",
            "+=", "-=", "*=", "/=", "&=", "|=", "^=", "%=", "<<=", ">>=", ">>>=",
        };
        for (int i = 0; i < ops.length; ++i) javaOperators.put(ops[i], ops[i]);
    }


    /**
     * An exception that reflects an error during parsing.
     */
    public static class ScanException extends Exception {
        public ScanException(String s) { super(s); }
    }

    /**
     * Reads an ISO 8859-1-encoded input stream. Escapes like "\u1234" (where
     * "xxxx" is a four-digit hex number) are decoded as unicode characters.
     */
    private static class BackslashUReader extends Reader {
        public BackslashUReader(InputStream is) { this.is = is; }
        public int read(char[] cbuf, int off, int len) throws IOException {
            int res = 0;
            while (len > 0) {
                int c;
                if (unreadChar == -1) {
                    c = is.read();
                    if (c == -1) break;
                } else {
                    c = unreadChar;
                    unreadChar = -1;
                }
                if (c == '\\') {
                    c = is.read();
                    if (c != 'u') {
                        unreadChar = c;
                        c = '\\';
                    } else {
                        byte[] ba = new byte[4];
                        if (is.read(ba) != 4) break;
                        try {
                            c = 0xffff & Integer.parseInt(new String(ba), 16);
                        } catch (NumberFormatException ex) {
                            throw new IOException("Invalid escape sequence \"\\u" + new String(ba) + "\"");
                        }
                    }
                }
                cbuf[off++] = (char) c;
                --len;
                ++res;
            }
            return res == 0 ? -1 : res;
        }
        public void close() throws IOException { is.close(); }

        private final InputStream is;
        private int               unreadChar = -1;
    }
}
