
/*
 * Janino - An embedded Java[TM] compiler
 * Copyright (C) 2001 Arno Unkrig
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Contact information:
 *   Arno Unkrig
 *   Ferdinand-Miller-Platz 10
 *   80335 Muenchen
 *   Germany
 *   http://www.janino.net
 *   maintainer@janino.net
 */

package net.janino.samples;

import java.util.*;
import java.lang.reflect.*;
import net.janino.*;

/**
 * A test program that allows you to play around with the
 * {@link net.janino.ExpressionEvaluator ExpressionEvaluator} class.
 */

public class ExpressionTest {
    final static String cvs_header = "$Header: /home/arno/cvsroot/JANINO/src/net/janino/samples/ExpressionTest.java,v 1.2 2001/10/27 21:44:49 arno Exp $";

    public static void main(String[] args) throws Exception {
        String   expression      = "total >= 100.0 ? 0.0 : 7.95";
        Class    expressionType  = Double.TYPE;
        String[] parameterNames  = { "total", };
        Class[]  parameterTypes  = { Double.TYPE, };

        int i;
        for (i = 0; i < args.length; ++i) {
            String arg = args[i];
            if (arg.charAt(0) != '-') break;
            if (arg.equals("-e")) {
                expression = args[++i];
            } else
            if (arg.equals("-et")) {
                expressionType = stringToType(args[++i]);
            } else
            if (arg.equals("-pn")) {
                StringTokenizer st = new StringTokenizer(args[++i], ",");
                Vector v = new Vector();
                while (st.hasMoreTokens()) {
                    v.add(st.nextToken());
                }
                parameterNames = new String[v.size()];
                v.toArray(parameterNames);
            } else
            if (arg.equals("-pt")) {
                StringTokenizer st = new StringTokenizer(args[++i], ",");
                Vector v = new Vector();
                while (st.hasMoreTokens()) {
                    v.add(stringToType(st.nextToken()));
                }
                parameterTypes = new Class[v.size()];
                v.toArray(parameterTypes);
            } else
            if (arg.equals("-help")) {
                usage();
                System.exit(0);
            } else
            {
                System.err.println("Invalid command line option \"" + arg + "\".");
                usage();
            }
        }

        if (parameterTypes.length != parameterNames.length) {
            System.err.println("Parameter type count and parameter name count do not match.");
            usage();
            System.exit(1);
        }

        // One command line argument for each parameter.
        if (args.length - i != parameterNames.length) {
            System.err.println("Argument and parameter count do not match.");
            usage();
            System.exit(1);
        }

        // Convert command line arguments to parameter values.
        Object[] parameterValues = new Object[parameterNames.length];
        for (int j = 0; j < parameterNames.length; ++j) {
            Class t = parameterTypes[j];
            // Wrap primitive parameters.
            if (t.isPrimitive()) {
                t = (
                    t == Boolean.TYPE   ? Boolean.class   :
                    t == Character.TYPE ? Character.class :
                    t == Byte.TYPE      ? Byte.class      :
                    t == Short.TYPE     ? Short.class     :
                    t == Integer.TYPE   ? Integer.class   :
                    t == Long.TYPE      ? Long.class      :
                    t == Float.TYPE     ? Float.class     :
                    t == Double.TYPE    ? Double.class    : t
                );
            }
            // Construct object, assuming it has a constructor with one single
            // "String" argument.
            Constructor c = t.getConstructor(new Class[] { String.class });
            parameterValues[j] = c.newInstance(new Object[] { args[i + j] });
        }

        // Create "ExpressionEvaluator" object.
        ExpressionEvaluator ee = new ExpressionEvaluator(
            expression,
            expressionType,
            parameterNames,
            parameterTypes
        );

        // Evaluate expression with actual parameter values.
        Object res = ee.evaluate(parameterValues);

        // Print expression result.
        System.out.println("Result = " + (res == null ? "(null)" : res.toString()));
    }

    private ExpressionTest() {}

    private static Class stringToType(String s) {
        if (s.equals("boolean")) return Boolean.TYPE;
        if (s.equals("char"   )) return Character.TYPE;
        if (s.equals("byte"   )) return Byte.TYPE;
        if (s.equals("short"  )) return Short.TYPE;
        if (s.equals("int"    )) return Integer.TYPE;
        if (s.equals("long"   )) return Long.TYPE;
        if (s.equals("float"  )) return Float.TYPE;
        if (s.equals("double" )) return Double.TYPE;
        try {
            return Class.forName(s);
        } catch (ClassNotFoundException ex) {
            ex.printStackTrace();
            System.exit(1);
            throw new RuntimeException(); // Never reached.
        }
    }

    private static void usage() {
        System.err.println("Usage:  ExpressionTest { <option> } { <parameter-value> }");
        System.err.println("Valid options are");
        System.err.println(" -e <expression>");
        System.err.println(" -et <expression-type>");
        System.err.println(" -pn <comma-separated-parameter-names>");
        System.err.println(" -pt <comma-separated-parameter-types>");
        System.err.println(" -help");
        System.err.println("The number of parameter names, types and values must be identical.");
    }
}
